%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:Mips-asm.sl
% Title:          Patterns and predicates for Mips PSL cmacro expansion
% Author:         Winfried Neun,
% Created:        September 15, 1989
% Modified:
% Status:         Experimental
% Mode:           Lisp
%
% Copyright (c) 1989, Konrad-Zuse-Zentrum Berin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn (load if!-system common)
	(off usermode)))

(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
	 outputsymfile!* commentformat!* labelformat!* 
	 externaldeclarationformat!* exporteddeclarationformat!* 
	 fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
	 reservedatablockformat!* definedfunctioncellformat!* 
	 undefinedfunctioncellinstructions!* mainentrypointname!* 
	 codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
	 numericregisternames!* pathin!* *usefastframe))
%%
%%      Mips assembler must declare global name before use.
%%
(fluid '(*declarebeforeuse))
(setq *declarebeforeuse t)

(setq codefilenameformat!* "%w.s")

(setq datafilenameformat!* "d%w.s")

(setq inputsymfile!* "mips.sym")

(setq outputsymfile!* "mips.sym")

(setq mainentrypointname!* (intern "psl_main"))

(setq commentformat!* " #  %p %n")

(setq labelformat!* "%w:")

(setq exporteddeclarationformat!* "     .globl %w%n")
(setq externaldeclarationformat!* "     .globl %w%n")

(setq fullwordformat!* "        .word %e%n")

% FullWord expects %e for parameter                                        

(setq doublefloatformat!* "     .double 0d%w%n")

(setq reservezeroblockformat!* "        .comm %w,%e%n")

(setq reservedatablockformat!* "        .space %e%n")

(setq definedfunctioncellformat!* " .word %w%n")

(setq undefinedfunctioncellinstructions!*
      '((!.word undefinedfunction)
	(!.word !%w)))
 
%% From lap-to-asm: 

(de initsymfnc1 (x)
  (prog (ep)
	(setq ep (get x 'entrypoint))
	(if (null ep)
	  (dataprintundefinedfunctioncell (id2int x))
	  (dataprintdefinedfunctioncell ep x))))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
	(setq oldout (wrs dataout*))
	(dataprintf " .word  undefinedfunction%n") 
	(wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
   (datadeclareexternal x)
     % dont disturb undefined function and interpreter
   (dataprintf definedfunctioncellformat* x (id2int idn)))

(fluid '(&links-present& locallabels*))

(de asmoutlap (u)
  (prog (locallabels* oldout *usefastframe)
	(setf &links-present& (or (look-for-links u)
	                          (assoc '*Id-apply u)
	                          (assoc '*call u)))

	(setq u (pass1lap (&fillframeholes u)))
	(setq u (lapopt u)) % lapopt pass
	% Expand cmacros, quoted expressions
	(codeblockheader)
	(setq oldout (wrs codeout*))
	(foreach x in u do (asmoutlap1 x))
	(wrs oldout)
	(codeblocktrailer)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq asmopenparen!* '!()

(setq asmcloseparen!* '!))

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
	 'binaryasmop)

(put 'lnot 'unaryasmop '!!)

(fluid '(*mkitemformat))   
(setq *mkitemformat "%e+(%e<<27)") 

(put 'mkitem 'asmexpressionfunction 'asm-mkitem)

(de asm-mkitem (mkitem-tag-inf)
  (printf
    *mkitemformat
    `(printexpression ,(caddr mkitem-tag-inf))
    `(printexpression ,(cadr mkitem-tag-inf))))

(global '(*writingasmfile))

(de codefileheader ()
  (setq *writingasmfile t)      %% Somebody must set this flag. but who?
  (codeprintf " .text%n")
 % (codeprintf " .set noreorder%n") % limit the intelligence of assembler
 % (codeprintf " .set nomacro%n")  
 % (codeprintf " .set noat%n"))
)

(de datafileheader ()
  (dataprintf " .data%n")
 %(dataprintf " .set noreorder%n")
 %(dataprintf " .set nomacro%n")  
 %(dataprintf " .set noat%n") )
)

(de codefiletrailer ()  nil)

(de datafiletrailer ()  nil)

(de codeblockheader ()  (codeprintf " .align 2%n"))

(de codeblocktrailer ()  nil)

(de dataalignfullword ()  (dataprintf " .align 2%n"))

(de printstring (s)
  (prog (n)
	(setq n (size s))
	(for (from i 0 n 1) (do (printbyte (indx s i))))
	(printbyte 0)
	(printf "       .align 2%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf "     .byte ")
  (printexpression x)
  (terpri))

(de printhalfwordlist (l)
  (foreach x in l do (printhalfword x)))

(de printhalfword (x)
  (printf "     .half ")
  (printexpression x)
  (terpri))

(de printopcode (x)
  (prog (!*lower)
	(setq !*lower t)
	(prin2 x)))

(de specialactionformainentrypoint ()
  (codeprintf " .globl psl_main%npsl_main:"))

(de asmsymbolp (x)
  (unixsymbolp (if (idp x) (id2string x)
	             x)))

(de unixsymbolp (x)
  (prog (n c i)
	(setq n (size x))
	(setq c (indx x 0))
	(unless (or (and (geq c (char a)) (leq c (char z))) 
	            (equal c (char _)) 
	            (and (geq c (char !a)) (leq c (char !z))))
	  (return nil))
	(setq i 1)
   loop
	(when (greaterp i n) (return t))
	(setq c (indx x i))
	(unless (or (and (geq c (char a)) (leq c (char z))) 
	            (equal c (char _)) 
	            (and (geq c (char !0)) (leq c (char !9))) 
	            (and (geq c (char !a)) (leq c (char !z))))
	  (return nil))
	(setq i (plus i 1))
	(go loop)))

(de printnumericoperand (x)  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (mips-register-p x)
    (progn (prin2 '0!()
	   (printoperand x)
	   (prin2 '!)))
    (error 0 (list "In Mips, deffered operand can't be used like this:" x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))             % (displacement (reg x) num)
  (printoperand (cadr x))
  (prin2 '!( )
  (printexpression (car x))
  (prin2 '!) ))

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintindexed (x)
  (setq x (cdr x))
  (printoperand (cadr x))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!) ))

(put 'indexed 'operandprintfunction 'operandprintindexed)

(de operandprintimmediate (x)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)
%%
%%      high and low is special keyword for mips assembler.
%%
(de operandprinthigh (x)
  (prin2 "%hi(")
  (printoperand (cadr x))
  (prin2 ")"))

(put 'high 'operandprintfunction 'operandprinthigh)
(put 'hi 'operandprintfunction 'operandprinthigh)

(de operandprintlow (x)
  (prin2 "%lo(")
  (printoperand (cadr x))
  (prin2 ")"))

(put 'low 'operandprintfunction 'operandprintlow)
(put 'lo 'operandprintfunction 'operandprintlow)

(de operandprintwarray (x)
  (prin2 (cadr x)))

(put 'WARRAY 'operandprintfunction 'operandprintwarray)

(de operandprintforeignentry (x)
  (prog (!*lower symnamx)
	(setq symnamx (symnam (id2int (cadr x))))
	(setq !*lower t)
	(prin2 (cadr x))))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf " # %p */%n" x) )

(de instructionprint (x)
  (codeprintf " #         %p */%n" x))

(de !*cerror (x)
  (prog (i)
	(setq i (wrs nil))
	(printf "%n *** CERROR: %r %n " x)
	(wrs i)
	(return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)
%%
%%      (word label)
%%
(put 'word 'asmpseudoop 'print-word)

(de print-word (x)
  (prin2 " .word ")
  (printoperand (cadr x))
  (terpri))

(put '*entry 'asmpseudoop 'asmprintentry)
(fluid '(*mainfound))
(de asmprintentry (x)
  (prog (y)
	(printcomment x)
	(setq x (cadr x))
	(setq y (findentrypoint x))
	(unless (flagp x 'internalfunction)
	  (findidnumber x))
	(if (eq x mainentrypointname*)
	  (progn (setq *mainfound t)
	         (specialactionformainentrypoint))
	  (codedeclareexporteduse y))))

(de codedeclareexporteduse (y)
    (codedeclareexported y)
    (printlabel y)
    (terpri))

(de findentrypoint (x)
  (cond ((get x 'entrypoint) (get x 'entrypoint))
	((and (asmsymbolp x)
              (not (get x 'symbol))
	      (not (flagp x 'foreignfunction)))
	 (put x 'entrypoint x) 
	 x)
	(t 
	 (let ((name (stringgensym)))
	   (put x 'entrypoint name)
	   name))
	))

(put 'define 'asmpseudoop 'printdefine)

(de print-register-or-immediate(name)
  (cond ((and (pairp name)  (or (eq (car name) 'low)
	       			(eq (car name) 'lo)))
	        (prin2 "%lo(") (printoperand (cadr name)) (prin2 ")"))
	((mips-register-p name) (printoperand name))
	(t       		 (printoperand name))))

(de print-immediate (val)
  (cond ((atom val)	         (printoperand val))
	((eq (car val) 'WARRAY)	 (prin2 (cadr val)))
	(t	 (error 0 (list "Unknown constant expression:" val)))))

(de print-register-name(name)
  (cond ((get name '*register-code)
		(prin2 (get name '*register-print-name)))
	((get name '*f-register-code) 
		(prin2 "$f") (prin2 (get name '*f-register-code)))
	(t      (printoperand name))))  %% if reg-name is not a register, it
	                                %% must be a constant.

%%
%%
%%      This function is originally defined in lap-to-asm.sl.
%%
(de printoperand (x)
  (setq x (to-standard-name x))
  (cond ((stringp x) 		    (prin2 x))
	((numberp x) 		    (printnumericoperand x))
	((mips-register-p x)	    (print-register-name x))
	((idp x)		    (prin2 (findlabel x)))
	((mips-register-p (car x)) (print-src-operand x))
	(t (prog (hd fn)
	         (setq hd (car x))
	         (cond ((setq fn (get hd 'operandprintfunction)) 
	                (apply fn (list x)))
	               ((and (setq fn (getd hd)) (equal (car fn) 'macro)) 
	                (printoperand (apply (cdr fn) (list x))))
	               ((setq fn (wconstevaluable x)) (printoperand fn))
	               (t (printexpression x)))))))
%
(de print-src-operand (src)
  (and (cdr src) (print-secondary-source (cadr src)))
  (prin2 "(")
  (printoperand (car src))
  (prin2 ")") )

(de print-secondary-source (src)
  (cond ((and (pairp src) (eq (car src) '-))
	 (cond ((numberp (cadr src))
	        (cond ((minusp (cadr src))
	               		      (prin2 "+") (prin2 (minus (cadr src))))
	              (t  (prin2 "-") (prin2 (cadr src)))))
	       (t      (prin2 "-") (printoperand (cadr src)))))

	((numberp src)
	 	(cond  ((minusp src)  (prin2 "-") (prin2 (minus src)))
	       		(t            (prin2 "+") (printoperand src))))
	(t	 (prin2 "+") (printoperand src))))

(deflist
  '((r0  "$0")  (r1  "$1")  (r2  "$2")  (r3 "$3")   (r4  "$4")  (r5  "$5")
    (r6  "$6")  (r7  "$7")  (r8  "$8")  (r9 "$9")   (r10 "$10") (r11 "$11")
    (r12 "$12") (r13 "$13") (r14 "$14") (r15 "$15") (r16 "$16") (r17 "$17")
    (r18 "$18") (r19 "$19") (r20 "$20") (r21 "$21") (r22 "$22") (r23 "$23")
    (r24 "$24") (r25 "$25") (r26 "$26") (r27 "$27") (r28 "$28") (r29 "$29")
    (r30 "$30") (r31 "$31"))
  '*register-print-name)

(ds kernelstring2string(str)
    (mkitem (tag str) (wplus2 16#4000000 (wshift (inf str) -2))))

