%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:spectrum-cmac.sl
% Title:          Patterns and predicates for HP Risc PSL cmacro expansion
% Author:         Winfried Neun
% Created:        September 19, 1991
% Mode:           Lisp
% Package:        Kernel
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(fluid '(AddressingUnitsPerItem CharactersPerWord StackDirection
	*usefastframe *usesafecondjumps
	*ImmediateQuote AddressingUnitsPerFunctionCell))

(loadtime (on ImmediateQuote))

(flag '(Immediate UnImmediate Indirect deferred label extrareg Indexed
	displacement) 'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart)
  (lor (lsh TagPart 27) (land InfPart 16#7ffffff)))

(de regp (r) (eqcar r 'reg))

(de short-5-immediate-p (x) (and (numberp x) (greaterp x -17) (lessp x 16)))

(de short-11-immediate-p (x) (and (numberp x)
                                (lessp x 16#400) (greaterp x (minus 16#401))))

(de short-immediate-p (x) (and (numberp x)
                                (lessp x 16#2000) (greaterp x (minus 16#2001))))

(de spectrum-indirect-p (r)
  (and (pairp r)
       (or (spectrum-register-p (car r))
           (eq (car r) 'deferred)
           (eq (car r) 'displacement))))

(de spectrum-huge-displacement-p (d)
   (and (numberp d) (geq d 16#2000) (lessp d (minus 16#2000))))

(de spectrum-relocatable-immediate-p (r)  (eqcar r 'IMMEDIATE))

(de spectrum-immediate-p(r)
  	(cond ((fixp r) t)
	      ((stringp r) t)
	      ((idp r)  (cond ((get r '*register-code) nil)
				((get r '*f-register-code) nil)
				((get r '*g-register-code) nil)
				(t             t)))  %% label
	      (t nil)))

(de spectrum-double-indirect-p (r)
   (and (eqcar r 'deferred) (spectrum-indirect-p (cadr r))))

(de spectrum-fluid-p (r)
          (and (pairp r) (member (car r)  '(!$fluid !$global fluid global))))

(de spectrum-absolute-p (r)
          (or (eqcar r 'quote) (stringp r)))

(de spectrum-foreignentry-p (r)
          (eqcar r 'foreignentry))

(de freep (asize)
	(if (and (wleq asize 16) (not &links-present&))
		      (setq *usefastframe t)
		      (setq *usefastframe nil)))

(de FiveP (x)  (eq x 5))

(de TwentySevenP (x)  (eq x 27))

(de evenp (x) (and (fixp x) (eq 0 (land x 1))))

(de indexedp (x) (eqcar x 'indexed))

(de DeferrableP (Expression)
  (and (PairP Expression)
       (not (memq (first Expression)
	      '(deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))

(de FakeRegisterNumberP (Expression)
  (and (IntP Expression) (GreaterP Expression 5)))

% NegativeImmediateP is for numbers whose negative is a one byte operand.

(de negp (value) (and (numberp value) (lessp value 0)))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de NegativeImmediateP (Expression)
  (and (FixP Expression) (minusp Expression) (GreaterP Expression -64)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR)  % ZIB

(de ZIBAnyregCAR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. caar
	       (member (caddr source) *nextreglist*))
	     (setq areg (caddr source))
	     (setq areg (nexttempreg)))

	  (push `(*move ,source , areg) Resultingcode*)
          (push `(depi 8 4 5 ,areg) Resultingcode*)
      (return `(displacement (reg sr0) ,areg 0))))

(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. cddar
	       (member (caddr source) *nextreglist*))
	     (setq areg (caddr source))
	     (setq areg (nexttempreg)))

	  (push `(*move ,source , areg) Resultingcode*)
	  (push `(depi 8 4 5 ,areg) Resultingcode*)
      (return `(displacement (reg sr0) ,areg 4))))

(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Four registers should be sufficient
      *nextreglist* '((reg t2) (reg t3) (reg t4) ))

(de nexttempreg (x)   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*))

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (setq aReg (if (and (displacementp source)
                    (member (setq cadrsource (caddr source))
                                 *nextreglist*))
               cadrsource      (nexttempreg)))
   (push `(*move ,source ,aReg) Resultingcode*)
   (unless (eq offset 0) (push `(*wplus2 ,aReg ,offset) Resultingcode*))
   (push `(*mkitem ,areg (quote 8)) Resultingcode*)
   (return `(displacement (reg sr0) ,areg 0))))

(DefAnyreg QUOTE
       AnyregQUOTE
       ((Null) (REG NIL))
       ((EqTP) (reg t))
       ((InumP) SOURCE)
       (       (QUOTE SOURCE)))

(DefAnyreg FRAME ZIBAnyregFrame)

(de ZIBAnyregFRAME (dum number)
     (if *usefastframe (nth '((reg ff1) (reg ff2) (reg ff3) (reg ff4)) number)
	`(displacement (reg sr0) (reg st) 
		,(minus (plus2 32 (times2 4 number))))))

(DefAnyreg xFRAME ZIBAnyregxFrame)

(de ZIBAnyregxFRAME (dum number)   % regardless of *usefastframe
    `(displacement (reg sr0) (reg st) ,(minus (plus2 32 (times2 4 number)))))

(DefAnyreg REG ZIBAnyregREG)

(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source) % argumentblock lies just below SYMVAL
	     `(displacement (reg sr0) (reg bndstkupperbound)
			 ,(wtimes2 source 4)))
	 (t   `(REG ,SOURCE))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

(de ZIBanyregFluid (reg source)
   (cond ((and (idp source) (get source 'registername))
		         `(reg ,(get source 'registername)))
	 (t       `($Fluid ,SOURCE))))

(defanyreg Global ZIBanyregglobal)

(defanyreg $Global ZIBanyregglobal)

(de ZIBanyregGlobal (reg source)
   (cond ((and (idp source) (get source 'registername))
		         `(reg ,(get source 'registername)))
	 (t       `($global ,SOURCE))))

%-----------------------------------------------------------------------------

(de asmp () *writingasmfile)

(DefCMacro *Call
       ((InternallyCallableP)  (bl (InternalEntry ArgOne) (reg r2))
				(NOP))
        (	 (LDO (displacement2 (reg 0) (plus -8000 (idloc ArgOne)))
				 (reg t2))
                  (ldwx (compl S)
			 (indexed (reg t2) (reg sr0) (reg symfncstart))
		         (reg t3)) 
		  (ble (displacement (reg sr5) (reg t3) 0))
                  (or  (reg r31) (reg r0) (reg r2))))

(DefCMacro *JCall
       ((InternallyCallableP) (bl (InternalEntry ArgOne) (reg r0))
			      (nop))
        (	 (LDO (displacement2 (reg 0) (plus -8000 (idloc ArgOne)))
				 (reg t2))
                  (ldwx (compl S)
                         (indexed (reg t2) (reg sr0) (reg symfncstart))
                         (reg t3))
                  (ble (displacement (reg sr5) (reg t3) 0))
                  (or  (reg r31) (reg r0) (reg r0))))

(de quote-fix-p (x)
    (and (eqcar x 'quote) (fixp (cadr x))))

(de displacement-p (x)
    (and (eqcar x 'displacement)
         (eq (length x) 4) 
         (equal (cadr x) '(REG SR0))
         (spectrum-register-p (caddr x))))

(de deferred-p (x)
    (and (pairp x) (or (spectrum-register-p (car x))
		   (eq (car x) 'deferred))))

(remprop '*move 'optfn)

(DefCMacro *Move
       (Equal)
       ((regp regp)                (or ArgOne (reg 0) ArgTwo))
       ((short-immediate-p regp)   (ldo (displacement2 (Reg 0) ArgOne) ArgTwo))
       ((indexedp regp)            (ldw ArgOne ArgTwo))
       ((regp indexedp)            (stw ArgOne ArgTwo))
       ((regp anyp)                (*movex ArgOne ArgTwo))
       ((anyp regp)                (*movex ArgOne ArgTwo))
       (                           (*move ArgOne (reg t5))
				   (*move (reg t5) ArgTwo)))

(de *ALLOC (framesize)
    (setq NAlloc!* framesize)
    (if (freep (times2 addressingunitsperitem framesize)) NIL
      (if (not (evenp framesize))
	(progn (setq framesize (iadd1 framesize))
	      `((stw (reg st) (displacement (reg sr0) 
				(reg st) ,(plus2 28 (times2 4 framesize))))
	       (stwm (reg NIL) (displacement (reg sr0) (reg st) 
			,(plus2 32 (times2 4 framesize))))
	       (stw (reg returnaddr) (displacement (reg sr0) (reg st) -20))))
	`((stw (reg st) (displacement (reg sr0)
				(reg st) ,(plus2 28 (times2 4 framesize))))
	  (addi ,(plus2 32 (times2 4 framesize)) (reg st) (reg st))
          (stw (reg returnaddr) (displacement (reg sr0) (reg st) -20))))))

(defcmacro *ALLOC)

(de *XALLOC (framesize)
     (if (not (evenp framesize))
	 (progn (setq framesize (iadd1 framesize))
	 `((stw (reg st) (displacement (reg sr0)
		    (reg st) ,(plus2 28 (times2 4 framesize))))
	  (stw (reg nil) (displacement (reg sr0) (reg st) 0))
	  (addi ,(plus2 32 (times2 4 framesize)) (reg st) (reg st))
	  (stw (reg returnaddr) (displacement (reg sr0) (reg st) -20))))
	`((stw (reg st) (displacement (reg sr0)
			(reg st) ,(plus2 28 (times2 4 framesize))))
	  (addi ,(plus2 32 (times2 4 framesize)) (reg st) (reg st))
	  (stw (reg returnaddr)(displacement (reg sr0) (reg st) -20)))))

(defcmacro *XALLOC)

(de *DeAlloc (Framesize)
 (if (freep (times2 addressingunitsperitem framesize)) NIL
  (when (not (evenp framesize)) (setq framesize (iadd1 framesize)))
  (Expand1OperandCMacro
   (times framesize (compiler-constant 'AddressingUnitsPerItem))
   '*DeAlloc)))

(defcmacro *XALLOC)

(DefCMacro *DeAlloc
	   ((FreeP))
	   (    (ldw (displacement (reg sr0) (reg st) -20) (reg returnaddr))
		(LDO (displacement2 (reg st) (minus (plus2 32 ArgOne)))
				 (reg st))))
 
(de *xDeAlloc (framesize)
     (when (not (evenp framesize)) (setq framesize (iadd1 framesize)))
	`((ldw (displacement (reg sr0) (reg st) -20) (reg returnaddr))
	  (LDO (displacement2 (reg st) (minus (plus2 32 (times2 4 ,framesize))))
			(reg st))))
 
(DefCMacro *xDeAlloc)

(de *Exit (N)
 (if (freep (times2 addressingunitsperitem N)) NIL
  (when (not (evenp N)) (setq N (iadd1 N))))
 (Expand1OperandCMacro
   (times N (compiler-constant 'AddressingUnitsPerItem)) '*Exit))

(DefCMacro *Exit     % leaf routine first
       ((Freep)  (bv (displacement2 (reg returnaddr) (reg r0)))
		 (nop))
       (         (ldw (displacement (reg sr0) (reg st) -20) (reg returnaddr))
		 (bv (displacement2 (reg returnaddr) (reg r0)))
		 (*Wplus2 (reg st) (minus (plus2 32 ArgOne)))))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de deferredp (x) (eqcar x 'deferred))

(DefCMacro *Jump
       ((deferredp)        (*move (undeferred ArgOne) (reg t5))
		           (bv (displacement2 (reg t5) (reg 0))) 
			   (or (reg 0) (reg 0) (reg 0))) % NO NOP!!!!!LAPOPT
       (                   (bl ArgOne (reg 0))
		           (nop)))


(DefCMacro *Lbl           (ArgOne))

(de *3op (a1 a2 instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	   (list (ResolveOperand '(Reg t1) A1)
		 (ResolveOperand '(Reg t2) A2)
		  Instruction)
		 (get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
       ((regp regp) (ArgThree ArgOne ArgTwo ArgOne))
       ((regp AnyP)  (*move ArgTwo (reg T5))
		      (*3op ArgOne (reg t5) ArgThree))
       ((Anyp regp)  (*move ArgOne (reg T5))
		      (ArgThree (reg t5) ArgTwo (reg t5))
		      (*move (reg t5) ArgOne))
       (              (*move ArgTwo (reg T5))
		      (*move ArgOne (reg T6))
		      (ArgThree (reg t5) (reg t6) (reg t5))
		      (*move (reg t5) ArgOne)))

(DefCMacro *WPlus2)

(de *wplus2 (arg1 arg2) (*wplus3 arg1 arg1 arg2))

(de *wplus3(arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wplus3))

(DefCMacro *WPlus3
	   ((regp regp regp) (add ArgTwo ArgThree ArgOne))
	   ((regp short-11-immediate-p regp)
		              (addi ArgTwo Argthree ArgOne))
	   ((regp regp short-11-immediate-p)
		              (addi  ArgThree ArgTwo ArgOne ))
           ((regp short-11-immediate-p anyp)
                              (*move ArgThree (reg t5))
                               (addi ArgTwo (reg t5) ArgOne))
	   ((regp anyp short-11-immediate-p)
		              (*move ArgTwo (reg t5))
		               (addi ArgThree (reg t5) ArgOne))
           ((regp regp anyp)  (*move ArgThree (reg t5))
                               (add ArgTwo (reg t5) ArgOne))
           ((anyp regp anyp)  (*move ArgThree (reg t5))
                               (add (Reg t5) ArgTwo (reg t5))
                               (*move (reg t5) ArgOne))
           ((regp anyp anyp)  (*move ArgTwo (reg t5))
                               (*move ArgThree (reg t6))
                               (add (reg t5) (reg t6) ArgOne))
	   (		      (*move ArgTwo (reg t5))
		               (*move ArgThree (reg t6))
		               (add (reg t5) (reg t6) (reg t5))
		               (*move (reg t5) ArgOne)))

(DefCMacro *WDifference
       ((*Wdifference3 ArgOne ArgOne ArgTwo)))

(de *wdifference3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wdifference3))

% format ArgOne = ArgTwo - ArgThree

(DefCMacro *Wdifference3
	   ((regp regp regp)  (sub ArgTwo ArgThree ArgOne))
	   ((regp short-11-immediate-p regp)
			      (subi ArgTwo ArgThree ArgOne))
	   ((regp anyp regp)  (*move ArgTwo (reg t5))
		               (sub (reg t5) ArgThree ArgOne))
	   ((regp anyp anyp)  (*move ArgTwo (reg t5))
		               (*move ArgThree (reg t6))
		               (sub (reg t5) (reg t6) ArgOne))
	   ((anyp regp anyp)  (*move ArgThree (reg t5))
		               (sub  ArgTwo (Reg t5) (reg t5))
		               (*move (reg t5) ArgOne))
	   (                  (*move ArgTwo (reg t5))
		               (*move ArgThree (reg t6))
		               (sub (reg t5) (reg t6) (Reg t5))
		               (*move (reg t5) ArgOne)))

(remprop '*wtimes2 'mc) (remprop 'wtimes2 'openfn)

(remprop '*wquotient 'mc) (remprop 'wquotient 'openfn)

(DefCMacro *WAnd
       ((*3op ArgOne ArgTwo and)))

(DefCMacro *WOr
       ((*3op ArgOne ArgTwo or)))

(DefCMacro *WXOr
       ((*3op ArgOne ArgTwo xor)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
    ((AnyP ZeroP))
    ((regp negp)     (*signedfield ArgOne ArgOne 0 (PLUS2 ArgTwo 32)))
    ((regp fixplusp) (*wshift ArgOne ArgTwo))
    ((regp regp)     (*wminus (reg t6) ArgTwo)
		      (comb (compl <) (reg t6) (reg r0) (label templabel))
		      (mtsar (reg t6))
		      (*SIGNEDFIELD (reg t6) ArgOne  0 1)
                      (VSHD (compl TR) (reg t6) ArgOne ARGONE)
                 templabel
                      (vshd  ArgOne (reg r0) ArgOne))
    ((regp anyp)     (*move ArgTwo (reg t5))
		      (*ashift ArgOne (reg t5)))
    ((Anyp fixp)      (*move ArgOne (reg t5))
		      (*ashift (reg t5) ArgTwo)
		      (*move (reg t5) ArgOne))
    (                 (*move ArgOne (reg t1))
		      (*ashift (reg t1) ArgTwo)
		      (*move (reg t1) ArgOne)))


(DefCMacro *WShift
    ((AnyP ZeroP))
    ((regp negp)     (shd (reg r0) ArgOne (minus ArgTwo) ArgOne))
    ((regp fixplusp) (shd ArgOne (reg r0) (difference 32 ArgTwo) ArgOne))
    ((regp regp)     (*wminus (reg t6) ArgTwo )
		      (comb (compl <) (reg t6) (reg r0) templabel)
		      (mtsar (reg t6))
		      (VSHD (compl TR) (reg r0) ArgOne ARGONE)
		 templabel
		      (vshd ArgOne (reg r0) ArgOne))
    ((regp anyp)     (*move ArgTwo (reg t5))
		      (*wshift ArgOne (reg t5)))
    ((Anyp fixp)      (*move ArgOne (reg t5))
		      (*wshift (reg t5) ArgTwo)
		      (*move (reg t5) ArgOne))
    (                 (*move ArgOne (reg t1))
		      (*wshift (reg t1) ArgTwo)
		      (*move (reg t1) ArgOne)))

(DefCMacro *WNot
       ((regp regp)  (*move -1 (reg t2))
		     (ANDCM  (reg t2) ArgTwo ArgOne))
       ((regp Anyp)  (*move ArgTwo ArgOne)
		      (*wnot  ArgOne ArgOne))
       (              (*move ArgTwo (reg t5))
		      (*wnot (reg t5) (reg t5))
		      (*move (reg t5) ArgOne)))

(DefCMacro *WMinus
	   ((regp regp) (sub (reg 0) ArgTwo ArgOne))
	   ((regp Anyp)  (*move ArgTwo ArgOne)
		          (sub (reg 0) ArgOne ArgOne))
	   (              (*move ArgTwo (reg t5))
		          (*wminus (reg t5) (reg t5))
		          (*move (reg t5) ArgOne)))

(DefCMacro *MkItem
       ((regp short-5-immediate-p)
			 (depi ArgTwo 4 5 ArgOne))
       ((regp regp)      (dep  ArgTwo 4 5  ArgOne))
       ((regp anyp)      (*move ArgTwo (reg t5))
		          (*Mkitem ArgOne (reg t5)))
       (                  (*move ArgOne (reg t1))
		          (*mkitem (reg t1) ArgTwo)
		          (*move (reg t1) ArgOne)))

(de smallinump (x) (short-immediate-p x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
      ((regp regp) (*field ArgOne ArgTwo 0 5))
      ((Anyp regp)  (*field (reg t1) ArgTwo 0 5 )
		    (*move (reg t1) ArgOne))
      ((regp anyp) (*move ArgTwo (reg t1))
		    (*field ArgOne (reg t1) 0 5))
      (             (*move ArgTwo (reg t1))
		    (*tag ArgOne (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
      (             (*Tag (Reg t6) ArgOne)
		    (*JumpEQ ArgThree ArgTwo (reg t6))))

(DefCMacro *JumpNotType
      (             (*Tag (Reg t6) ArgOne)
		    (*JumpNOTEQ ArgThree ArgTwo (reg t6))))

(DefCMacro *JumpInType
       ((regp Anyp) (*tag (reg t5) ArgOne)
		     (comib  (compl >=) ArgTwo (reg t5) ArgThree )
		     (*move 31 (reg t6))
		     (comb  (compl =) (reg t5) (reg t6) ArgThree )
		     (nop))
       (             (*move ArgOne (reg t5))
		     (*tag (reg t6) (reg t5))
                     (comib  (compl >=) ArgTwo (reg t6) ArgThree )
                     (*move 31 (reg t5))
                     (comb  (compl =) (reg t5) (reg t6) ArgThree )
		     (nop)))

(DefCMacro *JumpNotInType
       ((regp Anyp) (*tag (reg t5) ArgOne)
                     (comib  (compl >=) ArgTwo (reg t5) Templabel)
                     (*move 31 (reg t6))
                     (comb  (compl <>) (reg t5) (reg t6) ArgThree )
                     (nop)
               Templabel)
       (             (*move ArgOne (reg t5))
                     (*tag (reg t6) (reg t5))
                     (comib  (compl >=) ArgTwo (reg t6) Templabel)
                     (*move 31 (reg t5))
                     (comb  (compl <>) (reg t5) (reg t6) ArgThree )
                     (nop)
		Templabel))

%---------------------------------------------------------------------
%   General Purpose Jumper.
%   FORMAT: Operand1 Operand2 Address JumpOpCode

(DefCMacro *JumpIf
     ((regp regp)        (comb (compl argfour) ArgOne ArgTwo Argthree)
		         (nop))
     ((regp zerop)      (*jumpif ArgOne (reg r0) ArgThree ArgFour))
     ((regp anyp)       (*move ArgTwo (reg t5))
			 (comb (compl ArgFour) ArgOne (reg t5) Argthree)
		         (nop))
     ((short-5-immediate-p regp)
			 (comib (compl ArgFour) ArgOne ArgTwo Argthree)
                         (nop))
     ((zerop anyp)      (*jumpif (reg r0) ArgTwo ArgThree ArgFour))
     ((anyp  regp)      (*move ArgOne (reg t5))
		         (comb (compl argfour) (reg t5) ArgTwo Argthree)
		         (nop))
     (                   (*move ArgOne (reg t5))
		         (*move ArgTwo (reg t4))
		         (comb (compl argfour) (reg t5) (reg t4) Argthree)
		         (nop))
)

(DefCMacro *longJumpIf
     ((regp regp)        (comb (compl argfour) ArgOne ArgTwo TempLabel)
                         (nop)
			 (*jump Argthree)
			Templabel)
     ((regp zerop)      (*jumpif ArgOne (reg r0) ArgThree ArgFour))
     ((regp anyp)       (*move ArgTwo (reg t5))
                         (comb (compl ArgFour) ArgOne (reg t5) TempLabel)
                         (nop)
                         (*jump Argthree)
                        Templabel)
     ((short-5-immediate-p regp)
                        (comib (compl ArgFour) ArgOne ArgTwo TempLabel)
                         (nop)
                         (*jump Argthree)
                        Templabel)
     ((zerop anyp)      (*longjumpif (reg r0) ArgTwo ArgThree ArgFour))
     ((anyp  regp)      (*move ArgOne (reg t5))
                         (comb (compl argfour) (reg t5) ArgTwo TempLabel)
                         (nop)
                         (*jump Argthree)
                        Templabel)
     (                   (*move ArgOne (reg t5))
                         (*move ArgTwo (reg t4))
                         (comb (compl argfour) (reg t5) (reg t4) TempLabel)
                         (nop)
                         (*jump Argthree)
                        Templabel)
)


(de *JumpIF (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
	   (list (ResolveOperand '(Reg t1) ArgOne)
		 (ResolveOperand '(Reg t2) ArgTwo)
		 (ResolveOperand '(Reg Error) Label)
		  Instruction)
		 (get '*JumpIF 'CMacroPatternTable)))))

  % a variant for potential long jumps which do not fit 
  % into condjump size

(de *longJumpIF (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg Error) Label)
                  Instruction)
                 (get '*longJumpIF 'CMacroPatternTable)))))



(defcmacro *jumpeq)

(de *jumpeq (Lbl ArgOne ArgTwo)
       (if *usesafecondjumps
	(if (short-5-immediate-p (ResolveOperand '(Reg t1) ArgTwo))
                (*longjumpif ArgTwo ArgOne lbl '<>)
                (*longjumpif ArgOne ArgTwo lbl '<>))
        (if (short-5-immediate-p (ResolveOperand '(Reg t1) ArgTwo))
                (*jumpif ArgTwo ArgOne lbl '=)
                (*jumpif ArgOne ArgTwo lbl '=))))

(defcmacro *jumpnoteq)

(de *jumpnoteq(Lbl ArgOne ArgTwo)
       (if *usesafecondjumps
        (if (short-5-immediate-p (ResolveOperand '(Reg t1) ArgTwo))
                (*longjumpif ArgTwo ArgOne lbl '=)
                (*longjumpif ArgOne ArgTwo lbl '=))
        (if (short-5-immediate-p (ResolveOperand '(Reg t1) ArgTwo))
                (*jumpif ArgTwo ArgOne lbl '<>)
                (*jumpif ArgOne ArgTwo lbl '<>))))

(defcmacro *jumpwlessp)

(de *jumpwlessp (Lbl ArgOne ArgTwo)
       (if *usesafecondjumps
	(*longjumpif ArgOne ArgTwo lbl '>=)
	(*jumpif ArgOne ArgTwo lbl '<)))

(defcmacro *jumpwgreaterp)

(de *jumpwgreaterp (Lbl ArgOne ArgTwo)
       (if *usesafecondjumps
        (*longjumpif ArgOne ArgTwo lbl '<=)
	(*jumpif ArgOne ArgTwo lbl '>)))

(defcmacro *jumpwleq)

(de  *jumpwleq(Lbl ArgOne ArgTwo)
       (if *usesafecondjumps
        (*longjumpif ArgOne ArgTwo lbl '>)
	(*jumpif ArgOne ArgTwo lbl '<=)))

(defcmacro *jumpwgeq)

(de *jumpwgeq (Lbl ArgOne ArgTwo)
       (if *usesafecondjumps
        (*longjumpif ArgOne ArgTwo lbl '<)
	(*jumpif ArgOne ArgTwo lbl '>=)))

% a new type of cmacro: set dest to T or NIL under condition

(de *setcond (ArgOne ArgTwo ArgThree instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg t1) ArgThree)
                  instruction)
                 (get '*setcond 'CMacroPatternTable)))))

(defcmacro *setcond
     ((regp regp regp)  (comb (compl argfour) ArgTwo ArgThree templabel)
                         (*move (reg T) ArgOne)
                         (*move (reg NIL) ArgOne)
                         templabel)
     ((regp anyp zerop) (*setcond ArgOne ArgTwo (reg r0) ArgFour))
     ((regp zerop anyp) (*setcond ArgOne (reg r0) ArgThree ArgFour))
     ((regp short-5-immediate-p regp)
                        (comib (compl argfour) ArgTwo ArgThree templabel)
                         (*move (reg T) ArgOne)
                         (*move (reg NIL) ArgOne)
                         templabel)
     ((regp regp anyp)  (*move ArgThree (reg t5))
			 (*setcond ArgOne ArgTwo (reg t5) ArgFour))
     ((regp anyp regp)  (*move ArgTwo (reg t5))
                         (*setcond ArgOne (reg t5) ArgThree ArgFour))
     ((regp short-5-immediate-p anyp)
                        (*move ArgThree (reg t5))
                         (*setcond ArgOne ArgTwo (reg t5) ArgFour))
     ((regp anyp anyp)  (*move ArgTwo (reg t5))
                         (*move argThree ArgOne)
                         (*setcond ArgOne (reg t5) ArgOne ArgFour))
     (                  (*setcond (reg t5) ArgTwo ArgThree ArgFour)
                         (*move (reg t5) ArgOne))
)

(defcmacro *setcondeq)

(de *setcondeq (dest ArgOne ArgTwo)
	(if (short-5-immediate-p (ResolveOperand '(Reg t1) ArgTwo))
		(*setcond dest ArgTwo ArgOne '=)
		(*setcond dest ArgOne ArgTwo '=)))

(defcmacro *setcondnoteq)

(de *setcondnoteq(dest ArgOne ArgTwo)
        (if (short-5-immediate-p (ResolveOperand '(Reg t1) ArgTwo))
                (*setcond dest ArgTwo ArgOne '<>)
                (*setcond dest ArgOne ArgTwo '<>)))

(defcmacro *setcondwlessp)

(de *setcondwlessp (dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo '<))

(defcmacro *setcondwgreaterp)

(de *setcondwgreaterp (dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo '>))

(defcmacro *setcondwleq)

(de  *setcondwleq(dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo '<=))

(defcmacro *setcondwgeq)

(de *setcondwgeq (dest ArgOne ArgTwo)
        (*setcond dest ArgOne ArgTwo '>=))


(de *setcondtype (ArgOne ArgTwo ArgThree)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg t1) ArgThree))
                 (get '*setcondtype 'CMacroPatternTable)))))

(de *setcondNOTtype (ArgOne ArgTwo ArgThree)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                 (ResolveOperand '(Reg t1) ArgThree))
                 (get '*setcondNOTtype 'CMacroPatternTable)))))

(DefCMacro *setcondType
      (             (*Tag (Reg t6) ArgTwo)
                    (*setcondEQ ArgOne (Reg t6) ArgThree)))

(DefCMacro *setcondNotType
      (             (*Tag (Reg t6) ArgTwo)
                    (*setcondNOTEQ ArgOne (Reg t6) ArgThree)))

%---------------

(de fluid-arg-p (x)
    (and (pairp x)
     (member (car x) '(!$fluid !$global fluid global))))


(DefCMacro *Push
       ( (*cerror "NO push ")))

(DefCMacro *Pop
       ( (*cerror "NO pop ")))

(de *loc (ArgOne ArgTwo)
   (let ((resolvedarg (resolveoperand '(reg t2) ArgTwo)))
      (cond ((deferred-p resolvedarg)
	 `((*move (undeferred ,resolvedarg) ,ArgOne))) % get the address.
	 ((displacement-p resolvedarg)
            `((*wplus3 ,ArgOne ,(caddr resolvedarg) ,(cadddr resolvedarg))))
	 ((eqcar resolvedarg 'cdr)
	    `((add ,(cadr resolvedarg) ,ArgOne 4)))
         ((eqcar resolvedarg ' reg)
             (*cerror `("illegal reference" (*loc ,ArgOne ,ArgTwo))))
	 (t `((*movex ,resolvedarg ,ArgOne))))))

(DefCMacro *loc)

(DefCMacro *Field
  ((regp regp fixp fixp) (extru ArgTwo (sub1 (plus Argthree Argfour))
				ARGFOUR ArgOne))
  ((regp anyp fixp fixp) (*move ArgTwo (reg t5))
                          (*field ArgOne (reg t5) argthree argfour))
  ((anyp regp fixp fixp) (*field (reg t5) ArgTwo ArgThree ArgFour)
		          (*move (reg t5) ArgOne))
  (                      (*move ArgTwo (reg t6))
			  (*field (reg t5) (reg t6) ArgThree ArgFour)
			  (*move (reg t5) ArgOne)))

(DefCMacro *SignedField

  ((regp regp fixp fixp) (extrs ArgTwo (sub1 (plus Argthree Argfour))
				ARGFOUR ArgOne))
  ((regp anyp fixp fixp) (*move ArgTwo (reg t5))
			  (*signedfield ArgOne (reg t5) argthree argfour))
  ((anyp regp fixp fixp) (*Signedfield (reg t5) ArgTwo ArgThree ArgFour)
		          (*move (reg t5) ArgOne))
  (			 (*move ArgTwo (reg t6))
			  (*Signedfield (reg t5) (reg t6) ArgThree ArgFour)
			  (*move (reg t5) ArgOne)))

% following is used by garbage collector only. Not very useful to
% optimize other cases than put inf field or put tag field (with constant)


(DefCMacro *PutField
	((short-5-immediate-p regP)
	  (depi  ArgOne (SUB1 (PLUS2 ARGTHREE ARGFOUR)) ARGFOUR Argtwo))
	((short-5-immediate-p anyp)
	  (*move ArgTwo (reg t6))
          (*Putfield ArgOne (reg t6) ArgThree ArgFour)
          (*Move (Reg T6) ArgTwo))
	((regp regp)
	  (dep   ArgOne (SUB1 (PLUS2 ARGTHREE ARGFOUR)) ARGFOUR ArgTwo))
	((anyp regp)
         (*move   ArgOne (reg t5))
         (*Putfield (reg t5) ArgTwo ArgThree ArgFour))
	(         (*move ArgTwo (reg t6))
		  (*Putfield ArgOne (reg t6) ArgThree ArgFour)
		  (*Move (Reg T6) ArgTwo)))

(deflist '(
	(Byte        ((depi 8 4 5 (reg 1))
		      (ldbx (indexed (reg 2) (reg sr0) (reg 1)) (reg 1))))
	(PutByte     ((depi 8 4 5 (reg 1))
		      (*wplus3 (reg t1) (reg 2) (reg 1))
		      (stb (reg 3) (displacement (reg sr0) (reg t1) 0 ))))
	(HalfWord    ((depi 8 4 5 (reg 1))
			(ldhx (compl S) (indexed (reg 2) (reg sr0) (reg 1))
			(reg 1))))
	(PutHalfWord ((depi 8 4 5 (reg 1))
		      (*move (reg 2) (reg t1))
		      (*wshift (reg t1) 1)
		      (*wplus2 (reg t1) (reg 1))
		      (sth (reg 3) (displacement (reg sr0) (reg t1) 0)))))
  'OpenCode)

(&OneReg '(Byte PutByte HalfWord PutHalfWord))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
      (put function 'opencode
         `((extru (reg ,register) 31 27 (reg t2))
	   (*wdifference (reg t2) 8000)
           (ldwx (compl s) (indexed (reg t2) (reg sr0) (reg symfncstart)) (reg t1))
           (ble (displacement (reg sr5) (reg t1) 0))
           (or (reg r31) (reg r0) (reg r2))))
      (put function 'exitopencode
         `((extru (reg ,register) 31 27 (reg t2))
           (*wdifference (reg t2) 8000)
           (ldwx (compl s)
		 (indexed (reg t2) (reg sr0) (reg symfncstart)) (reg t1))
           (ble (displacement (reg sr5) (reg t1) 0))
           (or (reg r31) (reg r0) (reg r0))))
      ))

(for (in function '(addressapply0 addressapply1 addressapply2
		addressapply3 addressapply4))
     (from register 1)
     (do
      (put function 'opencode
	 `((ble (displacement (reg sr5) (reg ,register) 0))
                  (or  (reg r31) (reg r0) (reg r2))))

      (put function 'exitopencode
	 `((ble (displacement (reg sr5) (reg ,register) 0))
           (or  (reg r31) (reg r0) (reg r0))))
      ))

(defcmacro *fast-apply-load
   ( (*move ArgOne (reg t1)) ))

(put 'fast-idapply    'opencode
     '( (extru (reg t1) 31 27 (reg t2))
        (*wdifference (reg t2) 8000)
        (ldwx (compl s)
		  (indexed (reg t2) (reg sr0) (reg symfncstart)) (reg t1))
        (ble (displacement (reg sr5) (reg t1) 0))
        (or (reg r31) (reg r0) (reg r2))))

(put 'fast-codeapply    'opencode
     '( (depi 8 4 5 (reg t1))
                             %codepointer passed, so just strip off tag
        (ble (displacement (reg sr5) (reg t1) 0))
        (or (reg r31) (reg r0) (reg r2))))

(put 'fast-idapply    'exitopencode
     '( (extru (reg t1) 31 27 (reg t2))
        (*wdifference (reg t2) 8000)
        (ldwx  (compl s)
		 (indexed (reg t2) (reg sr0) (reg symfncstart)) (reg t1))
        (ble (displacement (reg sr5) (reg t1) 0))
        (or (reg r31) (reg r0) (reg r0))))

(put 'fast-codeapply    'exitopencode
     '( (depi 8 4 5 (reg t1))
                             %codepointer passed, so just strip off tag
        (ble (displacement (reg sr5) (reg t1) 0))
        (or (reg r31) (reg r0) (reg r0))))

(De *LamBind (Regs Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cregs cfluids cadrcfluids initload freeregs genlabel)
      
      (setq genlabel (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 8)) % two words per BndStk entry *4

   %looking for free register to preload values of fluid s

      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
		   (function (lambda (x)
		                (cond ((member x regs) nil)
		                       (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg T3) '(reg T4)) freeregs))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
		(nconc cadrcfluids (cons (car freeregs) nil)))
	(nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*move (reg BndStkPtr) (Reg t1))
                (*wplus3 (Reg t2) (Reg t1) ,lng)
                (comb (compl <=) (Reg t2) (reg BndstkUpperBound) ,genlabel)
                (*move (Reg t2) (reg BndstkPtr))
                (*call Bstackoverflow)  % is never come back
                ,genlabel   )) %start of code

      (setq list (append initload list))

  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	   (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
     `((*move ,(car freeregs) (displacement (reg sr0) (Reg t1) ,n))
       (*move (quote ,Cadrcfluids)
		   (displacement (reg sr0) (reg t1) ,(wplus2 n -4)))
       (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cfluids cadrcfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 8)) % two words per BndStk entry
		                         % * 4 addressingunits
      (setq freeregs '((reg t3)(reg T4)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
		(nconc cadrcfluids (cons (car freeregs) nil)))
	(nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist


      (setq list `((*move (reg BndStkPtr) (Reg t1))
                (*wplus3 (Reg t2) ,lng (Reg t1))
		(comb (compl <=) (Reg t2) (reg BndstkUpperBound) ,genlabel)
                (*move (Reg t2) (reg BndstkPtr))
                (*call Bstackoverflow)  % is never come back
                ,genlabel   )) %start of code

      (setq list (append initload list))

 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
	   (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
	     `((*move ,(car freeregs) (displacement (reg sr0) (Reg t1) ,n))
	       (*move (quote ,Cadrcfluids)
		           (displacement (reg sr0) (reg t1)  ,(wplus2 n -4)))
	       (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)

(De *FreeRstr (Fluids)

 (prog (n lng list cfluids listfluids genlabel initload freeregs)
      (setq genlabel (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t4)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 4 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 4)) % * addressingunitperitem
      (setq initload (list `(addi ,n (reg Bndstkptr) (reg t1))))
      (setq n 0)

preload  (setq initload
       (progn (setq listfluids
	       (if freeregs
		(nconc listfluids (cons (car freeregs) nil))
		(nconc listfluids (cons nil nil))) )
	(nconc initload
	 (if freeregs
     `((*move (displacement (reg sr0) (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 8))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))

       (cond (cfluids (go preload)))

       (setq freeregs listfluids ) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available

      (setq list `((addi ,(minus lng) (reg Bndstkptr) (reg t2))
		(comb (compl <=) (reg BndstkLowerBound) (reg t2) ,genlabel)
                (*move (Reg t2) (reg BndstkPtr))
                (*call Bstackunderflow)
                ,genlabel  )) %start of code

     (setq list (append initload list))
     (setq n 0)

 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 8))

  % insert reloaded register or memory reference

      (setq list (append list
	   (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
		`((*move (displacement (reg sr0) (Reg t2) ,n) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *freerstr)

(setq *unsafebinder t)   % has to save Registers across calls

(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
	 (join
	  `((*jumpeq ,(pop labellist) ,register ,x )))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if *writingasmfile
       (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (stringgensym))
       (SETQ X
	(if (and (weq lowerbound 0) (weq upperbound 31))
		  % jumpon on tags (most probably)
      `((move-relocatable-object ,ll2 (reg t1))
	(bv  (displacement2 (reg t1) ,register))
	(nop)
       ,ll2)
      `((*move ,upperbound (reg t5))
	(comb (compl =) ,register (reg t5) ,(lastcar labellist))
	(nop)
	(comb (compl <) (reg t5) ,register (label ,ll))
	(addi ,(minus lowerbound) ,register (reg t4))
        (*move ,lowerbound (reg t5))
	(comb (compl <) ,register (reg t5) (label ,ll))
        (move-relocatable-object ,ll2 (reg t1))
	(bv (displacement2 (reg t1) (reg t4)))
	(or (reg r31) (reg r0) (reg r0)) % no NOP here
       ,ll2) ) )
      Loop  (Setq x (nconc X 
		`((comb (compl tr) (reg r0) (reg r0) ,(car Labellist)) 
		,(list 'nop))) )
	    (setq Labellist (cdr Labellist))
	    (cond (Labellist (go loop)))

	    (setq x (nconc x `((*lbl (Label ,ll))) ))
	    (return x)
)  )  )

(defcmacro !*jumpon)


% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

  (if (wgreaterp numberofarguments 4)
      `((*alloc 5)
        (addi 128 (reg st) (reg st) )
	(*move (reg 5) (displacement (reg sr0) (reg st) -52))
	(*move (reg 6) (displacement (reg sr0) (reg st) -56))
	(*move (reg 7) (displacement (reg sr0) (reg st) -60))
	(*move (reg 8) (displacement (reg sr0) (reg st) -64))
	(*move (reg 9) (displacement (reg sr0) (reg st) -68))
	(*move (reg 10)(displacement (reg sr0) (reg st) -72))
	(*move (reg 11)(displacement (reg sr0) (reg st) -76))
	(*move (reg 12)(displacement (reg sr0) (reg st) -80))
	(*move (reg 13)(displacement (reg sr0) (reg st) -84))
	(*move (reg 14)(displacement (reg sr0) (reg st) -88))
	(*move (reg 15)(displacement (reg sr0) (reg st) -92))
        (*move (reg t) (fluid *kernelmode))
        (*move (reg nil) (frame 17))
        (!.import (foreignentry ,functionname) (foreignentry  CODE))
        (!.import (foreignentry psl_stub) (foreignentry CODE))
        (ldil (high (foreignentry ,functionname)) (reg r7))
        (ldo  (displacement2 (reg r7) (low (foreignentry ,functionname)))
                 (reg r8))
        (ldil (high (foreignentry psl_stub)) (reg t6))
        (ldo  (displacement2 (reg t6) (low (foreignentry psl_stub)))
                         (reg r31))
        (ldsid (displacement2 (reg r31) 0) (reg r1))
        (mtsp (reg r1) (reg sr0))
        (BLE  (displacement (reg sr0) (reg r31) 0))
        (or   (reg r31) (reg r0) (reg r2))

        (*move (reg r28) (reg 1))  % returnvalue
        (*move (frame 17) (reg nil))
        (*move (reg NIL) (fluid *kernelmode))
        (addi -128 (reg st) (reg st) )
        (*dealloc 5))

      `((*alloc 5)
	(addi 128 (reg st) (reg st) )
 	(*move (reg t) (fluid *kernelmode))
	(*move (reg nil) (frame 17))
	(!.import (foreignentry ,functionname) (foreignentry  CODE))
	(!.import (foreignentry psl_stub) (foreignentry CODE))
	(ldil (high (foreignentry ,functionname)) (reg r7))
	(ldo  (displacement2 (reg r7) (low (foreignentry ,functionname)))
		 (reg r8))
	(ldil (high (foreignentry psl_stub)) (reg t6))
	(ldo  (displacement2 (reg t6) (low (foreignentry psl_stub)))
			 (reg r31))
	(ldsid (displacement2 (reg r31) 0) (reg r1))
	(mtsp (reg r1) (reg sr0))
	(BLE  (displacement (reg sr0) (reg r31) 0))
	(or   (reg r31) (reg r0) (reg r2))
	
	(*move (reg r28) (reg 1))  % returnvalue
	(*move (frame 17) (reg nil))
        (*move (reg NIL) (fluid *kernelmode))
	(addi -128 (reg st) (reg st) )
        (*dealloc 5))))

(defcmacro !*foreignlink)

(defcmacro *movex)
   % Movex is of type cmacro

(de *movex (src dst)
     (let (tmp)
      (cond ((spectrum-register-p dst)
	  (cond ((spectrum-register-p src)  `((*move ,src ,dst)))
		((eqcar src 'idloc)     `((add ,dst (reg zero) ,src)))
		((eqcar src 'quote)
		     (if *writingasmfile           %% asmout and lap generate
		         `((ldil  (high ,src) ,dst    )  %% different code.
		           (ldo (displacement2 ,dst (low ,src))  ,dst))
		         `((move-relocatable-object ,src ,dst))))
		((or (spectrum-absolute-p src)(spectrum-fluid-p src))
		 (if *writingasmfile           %% asmout and lap generate
		     `((addil (high ,src) (reg 0)) %% different code.
		       (ldw (displacement (reg sr0) (reg r1) (low ,src)) ,dst))
		     `((move-relocatable-object-and-load ,src ,dst))))
		((short-immediate-p src) `((or ,dst (reg 0) ,src )))
		((or (spectrum-relocatable-immediate-p src)
		     (spectrum-foreignentry-p src))
		 (if *writingasmfile           %% asmout and lap generate
		     `((ldil  (high ,src) ,dst    )  %% different code.
		       (ldo (displacement2 ,dst (low ,src))  ,dst))
		     `((move-relocatable-object ,src ,dst))))
		((spectrum-immediate-p src)
		     `((LDIL (high ,src) ,dst)
		       (LDO (displacement2 ,dst (low ,src)) ,dst)))
		((spectrum-huge-displacement-p src)
		     `((LDIL  (high ,(cadddr src)) ,dst )
		       (LDO (displacement2 ,dst (low ,(cadddr src))) ,dst)
		       (ldw (indexed ,(cadr src) (reg sr0) ,dst) ,dst)))
		((spectrum-indirect-p src) `((ldw ,src ,dst )))
		(t  (error 0
		 (list "Unknown addressing mode for *movex:" src dst)))))
	 ((spectrum-register-p src)
	  (cond ((or (spectrum-absolute-p dst) (spectrum-foreignentry-p src))
		 (if *writingasmfile
		     `((addil (high ,dst) (reg 0))
		       (stw ,src (displacement (reg sr0) (reg r1) (lo16 ,dst))))
		     `((move-relocatable-object-and-store
		                     ,src (reg t6) ,dst))))
                ((and (spectrum-fluid-p dst) *writingasmfile)
                  `((addil (high ,dst) (reg 0))
		   (stw ,src (displacement (reg sr0) (reg r1) (low ,dst)))))
                ((spectrum-fluid-p dst)
			`((move-relocatable-object-and-store
                                     ,src (reg t6) ,dst)))
		((spectrum-huge-displacement-p dst)
		 `((LDIL  (high ,(cadddr dst)) (reg t6))
		   (LDO  (displacement2 (reg t6) (low ,(cadddr dst))) (reg t6))
		   (stw ,src (indexed ,(cadr dst) (reg sr0) (reg t6)))))
		((spectrum-indirect-p dst) `((stw ,src ,dst)))
		(t  (error 0
		        (list "Unknown addressing mode:" src dst)))))
	 (t (print (list " Unsaeglich: Move " src dst))
))))

% patch for pass-1-reform LOC

(de pa1r-loc(u)
  (setq u (cadr u))
  (when (eqcar u 'memory)
    `(mkitemrev (wplus2 ,(third u) ,(second u)) ',vector-tag )))
