%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           SPARC-ASM.SL
% Title:          SUN-4 specific information for LAP-TO-ASM
% Author:         Tsuyoshi Yamamoto
% Created:        5 May 1982
% Package:        Kernel
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% Dec 12 1988 (Winfried Neun)
%       adapted symfnc preset to new calling scheme
%       removing things that are no more needed
% Aug 22 1988 (padget and burdorf)
%       added function to handle the sun 4 os 4 assembler directive 
%       wierdness which requires the syntax to be 3+(3>>5) instead of
%       the standard (3>>5)+3
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn (load if!-system common)
        (off usermode)))

%(compiletime
%  (if_system vax nil 
%             (progn (dskin "vax-data-machine.sl")
%                    (dskin "pc:data-machine.sl")
%                    )))


(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
         outputsymfile!* commentformat!* labelformat!* 
         externaldeclarationformat!* exporteddeclarationformat!* 
         fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
         reservedatablockformat!* definedfunctioncellformat!* 
         undefinedfunctioncellinstructions!* mainentrypointname!* 
         codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
         numericregisternames!* pathin!*))
%%
%%      Sun-4 assembler must declare global name before use.
%%
(fluid '(*declarebeforeuse))
(setq *declarebeforeuse t)

(setq codefilenameformat!* "%w.s")

(setq datafilenameformat!* "d%w.s")

(setq inputsymfile!* "sparc.sym")

(setq outputsymfile!* "sparc.sym")

(setq mainentrypointname!* (intern "_psl_main"))

(setq commentformat!* "! /* %p */%n")

(setq labelformat!* "%w:")

(setq exporteddeclarationformat!* "     .global %w%n")
(setq externaldeclarationformat!* "     .global %w%n")

(setq fullwordformat!* "        .word %e%n")

% FullWord expects %e for parameter                                        

(setq doublefloatformat!* "     .double 0d%w%n")

(setq reservezeroblockformat!* "        .common %w,%e%n")

(setq reservedatablockformat!* "        .skip %e%n")

(setq definedfunctioncellformat!* " ba %w%n sethi %w,%%o0%n")

(setq undefinedfunctioncellinstructions!*
      '((ba (plus2 symfnc (times2 8 (idloc undefinedfunction))))
        (sethi !%w (reg o0))))
 
%% From lap-to-asm: 

(de initsymfnc1 (x)
  (prog (ep)
        (setq ep (get x 'entrypoint))
        (if (null ep)
          (dataprintundefinedfunctioncell (id2int x))
          (dataprintdefinedfunctioncell ep x))))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
        (setq oldout (wrs dataout*))
%        (foreach x in undefinedfunctioncellinstructions* do
%                 (asmoutlap1 x))
        (asmoutlap1 (car undefinedfunctioncellinstructions*))
        (asmoutlap1 (list 'sethi idn '(reg o0)))
        (wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
  (datadeclareexternal x)
     % dont disturb undefined function and interpreter
  (if (memq x '(COMPILEDCALLINGINTERPRETED undefinedfunction))
   (dataprintf " ba %w%n nop%n" x)
   (dataprintf definedfunctioncellformat* x (id2int idn))))

(fluid '(&links-present& locallabels*))

(de asmoutlap (u)
  (prog (locallabels* oldout)
        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
                                  (assoc '*call u)))

        (setq u (pass1lap (&fillframeholes u)))
        (setq u (lapoptreformjumps u)) % lapopt pass
        % Expand cmacros, quoted expressions
        (codeblockheader)
        (setq oldout (wrs codeout*))
        (foreach x in u do (asmoutlap1 x))
        (wrs oldout)
        (codeblocktrailer)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq asmopenparen!* '!()

(setq asmcloseparen!* '!))

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
         'binaryasmop)

(put 'lnot 'unaryasmop '!!)

%(put 'mkitem 'asmexpressionformat "(%e<<27)+%e")
(fluid '(*mkitemformat))           %modified for sun 4 os 4
(setq *mkitemformat "%e+(%e<<27)") % ZIB

(put 'mkitem 'asmexpressionfunction 'asm-mkitem)

(de asm-mkitem (mkitem-tag-inf)
  (printf
    *mkitemformat
    `(printexpression ,(caddr mkitem-tag-inf))
    `(printexpression ,(cadr mkitem-tag-inf))))

(global '(*writingasmfile))

(de codefileheader ()

  (setq *writingasmfile t)      %% Somebody must set this flag. but who?

  (codeprintf " .seg    ""text""%n"))

(de datafileheader ()
  (dataprintf " .seg    ""data""%n"))

(de codefiletrailer ()
  nil)

(de datafiletrailer ()
  nil)

(de codeblockheader ()
  (codeprintf " .align 4%n"))

(de codeblocktrailer ()
  nil)

(de dataalignfullword ()
  (dataprintf " .align 4%n"))

(de printstring (s)
  (prog (n)
        (setq n (size s))
        (for (from i 0 n 1) (do (printbyte (indx s i))))
        (printbyte 0)
        (printf "       .align 4%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf "     .byte ")
  (printexpression x)
  (terpri))

(de printhalfwordlist (l)
  (foreach x in l do (printhalfword x)))

(de printhalfword (x)
  (printf "     .half ")
  (printexpression x)
  (terpri))

(de printopcode (x)
  (prog (!*lower)
        (setq !*lower t)
        (prin2 x)))

(de specialactionformainentrypoint ()
  (codeprintf " .global _psl_main%n_psl_main:"))

(de asmsymbolp (x)
  (unixsymbolp (if (idp x)
                 (id2string x)
                 x)))

(de unixsymbolp (x)
  (prog (n c i)
        (setq n (size x))
        (setq c (indx x 0))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i 1)
   loop
        (when (greaterp i n)
          (return t))
        (setq c (indx x i))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !0)) (leq c (char !9))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i (plus i 1))
        (go loop)))

(de printnumericoperand (x)
  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (sparc-register-p x)
    (progn (prin2 '![)
           (printoperand x)
           (prin2 '!]))
    (error 0 (list "In sparc, differed operand can't be used like this:" x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))
  (prin2 '![ )
  (printoperand (car x))
  (prin2 '!+ )
  (printexpression (cadr x))
  (prin2 '!]))

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintindexed (x)
  (setq x (cdr x))
  (prin2 '![ )
  (printoperand (cadr x))
  (prin2 '!+ )
  (printoperand (car x))
  (prin2 '!] ))

(put 'indexed 'operandprintfunction 'operandprintindexed)

(de operandprintimmediate (x)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)
%%
%%      high and low is special keyword for sparc assembler.
%%
(de operandprinthigh (x)
  (prin2 "%hi(")
  (printoperand (cadr x))
  (prin2 ")"))

(put 'high 'operandprintfunction 'operandprinthigh)
(put 'hi 'operandprintfunction 'operandprinthigh)

(de operandprintlow (x)
  (prin2 "%lo(")
  (printoperand (cadr x))
  (prin2 ")"))

(put 'low 'operandprintfunction 'operandprintlow)
(put 'lo 'operandprintfunction 'operandprintlow)

(de operandprintwarray (x)
  (prin2 (cadr x)))

(put 'WARRAY 'operandprintfunction 'operandprintwarray)

(de operandprintforeignentry (x)
  (prog (!*lower symnamx)
        (setq symnamx (symnam (id2int (cadr x))))
        (setq !*lower t)
        (when (not (eq (id2int '!.)(strbyt (strinf symnamx) 0)))
           (prin2 "_")) % dont change .mul into _.mul
        (prin2 (cadr x))))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf "/* %p */%n" x)
  )

(de instructionprint (x)
  (codeprintf "/*         %p */%n" x)
  )

(de !*cerror (x)
  (prog (i)
        (setq i (wrs nil))
        (printf "%n *** CERROR: %r %n " x)
        (wrs i)
        (return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)
%%
%%      jmpl has something different syntax in sun4 `as'.
%%
(put 'jmpl 'asmpseudoop 'printjmpl)

(de printjmpl (x)
  (let ((src (cadr x))
        (dst (caddr x)))
       (prin2 " jmpl ")
       (cond ((eq (car src) 'deferred)
              (printoperand (cadr src)))
             ((eq (car src) 'displacement)
              (printoperand (cadr src))
              (print-secondary-source (caddr src)))
             (t
              (printoperand (car src))
              (cond ((cdr src)
                     (print-secondary-source (cadr src))))))
       (prin2 ",")
       (printoperand dst)
       (terpri)))
%%
%%      (word label)
%%
(put 'word 'asmpseudoop 'print-word)

(de print-word (x)
  (prin2 " .word ")
  (printoperand (cadr x))
  (terpri))

(put '*entry 'asmpseudoop 'asmprintentry)
(fluid '(*mainfound))
(de asmprintentry (x)
  (prog (y)
        (printcomment x)
        (setq x (cadr x))
        (setq y (findentrypoint x))
        (unless (flagp x 'internalfunction)
          (findidnumber x))
        (if (eq x mainentrypointname*)
          (progn (setq *mainfound t)
                 (specialactionformainentrypoint))
          (codedeclareexporteduse y))))

(de codedeclareexporteduse (y)
    (codedeclareexported y)
    (printlabel y)
    (terpri))

(de findentrypoint (x)
  (cond ((get x 'entrypoint) (get x 'entrypoint))
        ((and (asmsymbolp x)
              (not (get x 'symbol))
              (not (flagp x 'foreignfunction)))
         (put x 'entrypoint x) 
         x)
        (t 
         (let ((name (stringgensym)))
           (put x 'entrypoint name)
           name))
        ))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = "))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = ")
  (prin2 (caddr x))
  (terpri))

(put 'define 'asmpseudoop 'printdefine)

(de print-register-or-immediate(name)
  (cond ((and (pairp name)
              (or (eq (car name) 'low)
                  (eq (car name) 'lo)))
         (prin2 "%lo(")
         (printoperand (cadr name))
         (prin2 ")"))
        ((sparc-register-p name)
         (printoperand name))
        (t
         (printoperand name))))

(de print-immediate (val)
  (cond ((atom val)
         (printoperand val))
        ((eq (car val) 'WARRAY)
         (prin2 (cadr val)))
        (t
         (error 0 (list "Unknown constant expression:" val)))))

(de print-register-name(name)
  (cond ((get name '*register-code)
         (prin2 "%")
         (prin2 (get name '*register-print-name)))
        ((get name '*f-register-code)
         (prin2 "%f")
         (prin2 (get name '*f-register-code)))
        ((get name '*c-register-code)
         (prin2 "%c")
         (prin2 (get name '*c-register-code)))
        (t
         (printoperand name))))         %% if reg-name is not a register, it
                                        %% must be a constant.

%%
%%
%%      This function is originally defined in lap-to-asm.sl.
%%
(de printoperand (x)
  (setq x (to-standard-name x))
  (cond ((stringp x) (prin2 x))
        ((numberp x) (printnumericoperand x))
        ((sparc-register-p x)
         (print-register-name x))
        ((idp x) (prin2 (findlabel x)))
        ((sparc-register-p (car x))
         (print-src-operand x))
        (t (prog (hd fn)
                 (setq hd (car x))
                 (cond ((setq fn (get hd 'operandprintfunction)) 
                        (apply fn (list x)))
                       ((and (setq fn (getd hd)) (equal (car fn) 'macro)) 
                        (printoperand (apply (cdr fn) (list x))))
                       ((setq fn (wconstevaluable x)) (printoperand fn))
                       (t (printexpression x)))))))
%
%       print operand by [rs1+rs2] format.
%
(de print-src-operand (src)
  (prin2 "[")
  (printoperand (car src))
  (and (cdr src)
       (print-secondary-source (cadr src)))
  (prin2 "]")
  )

(de print-secondary-source (src)
  (cond ((and (pairp src)
              (eq (car src) '-))
         (cond ((numberp (cadr src))
                (cond ((minusp (cadr src))
                       (prin2 "+")
                       (prin2 (minus (cadr src))))
                      (t
                       (prin2 "-")
                       (prin2 (cadr src)))))
               (t
                (prin2 "-")
                (printoperand (cadr src)))))

        ((numberp src)
         (cond ((minusp src)
                (prin2 "-")
                (prin2 (minus src)))
               (t
                (prin2 "+")
                (printoperand src))))
        (t
         (prin2 "+")
         (printoperand src))))

(deflist
  '((r0  "g0")(r1  "g1")(r2  "g2")(r3  "g3")(r4  "g4")
    (r5  "g5")(r6  "g6")(r7  "g7")(r8  "o0")(r9  "o1")
    (r10 "o2")(r11 "o3")(r12 "o4")(r13 "o5")(r14 "o6")
    (r15 "o7")(r16 "l0")(r17 "l1")(r18 "l2")(r19 "l3")
    (r20 "l4")(r21 "l5")(r22 "l6")(r23 "l7")(r24 "i0")
    (r25 "i1")(r26 "i2")(r27 "i3")(r28 "i4")(r29 "i5")
    (r30 "i6")(r31 "i7")
    (fsr "fsr") % floating point status register
    (g0 "g0")(g1 "g1")(g2 "g2")(g3 "g3")(g4 "g4")(g5 "g5")(g6 "g6")(g7 "g7")
    (o0 "o0")(o1 "o1")(o2 "o2")(o3 "o3")(o4 "o4")(o5 "o5")(o6 "o6")(o7 "o7")
    (l0 "l0")(l1 "l1")(l2 "l2")(l3 "l3")(l4 "l4")(l5 "l5")(l6 "l6")(l7 "l7")
    (i0 "i0")(i1 "i1")(i2 "i2")(i3 "i3")(i4 "i4")(i5 "i5")(i6 "i6")(i7 "i7"))
  '*register-print-name)

(put 'fsr '*register-code 4711)

