%
% 370-LAP.RED - IBM 370 PSL assembler
%
% Author:      James McDonald
%              IMSSS
%              Stanford University
% Date:        7 November 1983
% Copyright (c) 1983 Stanford University
%
% imports '(370LTBLS 370LDEP 370OPER);
%

(compiletime (load wdeclare)) % WN

% new tag -- july 16, 1984 -- remove this line when totally integrated

%------------------------------------------------------------------------
%   Fluid declarations
%------------------------------------------------------------------------

% user settable switches

(fluid '(*writingfaslfile 
        *pgwd                           % shows cmacros
        *pwrds 
        *listlap                        % shows hex code cmacros turn into
        *depositlap                     % actually deposit instructions in memory
        *crosslap                       % if running on non-target machine
	forwardinternalreferences*      % forward refs (internalfunctions)
        *contentsafterconstants))       % T for IBM, NIL for everyone else
        

(setq *contentsafterconstants t)
%
%  On a machine without actual base registers, think of BaseRegister as
%  being virtually defined and pointing at some convenient location such
%  as the start of the code being loaded.
%
%  Offsets are all expressed in addressing units, unless otherwise noted.
%  On machines that support multiple addressing modes, an offset might
%  be a pair, e.g. (Word 12000), (Byte 36446), etc.
%
%  If FASL'ing then a block will contain all the fns in the file.
%  Otherwise, each function will have its own block.
%

(fluid 
   '(lapcode*                           % lap code being manipulated
   lapdata*                             % lap data being manipulated
   lapinstruction*                      % lap representation of instr to deposit
   lapdatum*                            % lap representation of datum to deposit

   faslptr*                             % ptr where fasl code is being created
   faslend*                             % last memory location available to FASL

   codeblocksize*                       % in address units, code and lits in fn
   codeblockbase*                       % point of addressibility
   codeblockoffsets*                    % .. (lbl . offset from base address) ..
   entries*                             % .. ((fn type ..) . offset from base) ..

   datablocksize*                       % in address units, lispish data for fn
   datablockbase*                       % used only by listing routines
   datablockoffsets*                    % .. (lbl . offset within data area) ..

   lapdepositptr*                       % place where code or data is deposited
   hwoffsetwithininstruction*           % offset in halfwords within instruction

   lapreturnvalue*                      % absolute address of first entry

   opcode*                              % ID mnemonic for opcode
   opcodenumber*                        % bit pattern, as small integer
                                % i.e., justified in low bits of item
   operands*                            % list of operands as received
   operand*1 operand*2                  % individual operands extracted for use
   operand*3 operand*4                  % in fields to be EVAL'ed
   operand*5 operand*6 
   listlapout*                          % output channel for Lap listing
                                % where should this be set???
   addressesknowntolap*                 % runtime locations where various regs pt
   faslcodepointerindex*))              % counter for references to anonymous fns
 % end of fluid declaration

(fluid '(literalslabel* 
        halfwordidoffsetslabel* 
        fullwordidoffsetslabel* 
        constantslabel* 
        contentsofconstantslabel*))

(progn 
(setq literalslabel* "* --- Literals ---") 
(setq halfwordidoffsetslabel* "* --- Halfword ID offsets ---") 
(setq fullwordidoffsetslabel* "* --- Fullword ID offsets ---") 
(setq constantslabel* "* --- Top level items of constants ---") 
(setq contentsofconstantslabel* "* --- Contents of constants ---") nil)


% misc

(fluid '(laphalfwords* numericregisternames*))

(setq laphalfwords* (mkvect 20))
(setq numericregisternames* '[0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15])

(setq *pwrds t)


(compiletime
 (Progn
	(define-constant minmachineregister 0)
	(define-constant maxmachineregister 15)
	(define-constant offsettoentrypoint 20)
	))
 (commentoutcode
  (progn (wdeclare internal wconst (offsettoentrypoint nil 20))
	 (wdeclare internal wconst 	% must be same in KNFASL RED
		(minmachineregister nil 0) 
                (maxmachineregister nil 15))
	 (wdeclare internal wconst 
		(donothing nil 0) 
                (installnow nil 1) 
                (executenow nil 2) 
                (deletenow nil 3)) 

))
(compiletime 
(ds labelp (x) (atom x)) nil)


(loadtime (progn 

 (cond (*crosslap 
      (progn (setq *listlap t) (setq *depositlap nil))) (t 
 (progn (setq *listlap nil) (setq *depositlap t)))) nil))

(de posintp (x) 
(and (fixp x) (wgreaterp x 0)))

(fluid '(*using*))

(de lap (u)
  (prog (labeloffsets* lapreturnvalue* entries* *usefastframe *using* uu prev)
        (unless *writingfaslfile
          (setq currentoffset* 0)
          (setq codebase* (gtbps 0)))

        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
                                  (assoc '*call u)))

	(setq u (append u '((DS  "0F"))))
        (setq u (pass1lap (&fillframeholes u)))

        % removing "dummy ..." instructions

        (setq uu u)
        (while uu
           (progn
              (cond ((atom (car uu)) nil)
		    ((eq (caar uu) 'dummy) (rplacd prev (cdr uu))))
	      (setq prev uu)
              (setq uu (cdr uu))))

        (when (wlessp (length u) 1000) % a heuristic
	    (setq u (delete '(la (reg r2) (displacement (reg r3) 2048))
		(delete '(la (reg r2) (displacement (reg r2) 2048)) u))))

        (when *pgwd
          (foreach x in u do (if (atom x)
                     (prin2 x)
                     (printf "          %p%n" x))))

        %
        %       findlabels calculates codesize and fixup label address
        %
        (findlabels u)

        (unless *writingfaslfile
                (unless (eq nextbps (wand nextbps (wnot 7))) (gtbps 1))
                (setq codebase* (gtbps (wquotient (wplus2 codesize!* 3) 4))))

        %
        %       generate machine code
        %
        (foreach x in u do (depositinstruction x))

        (defineentries)
        (when (and (not *writingfaslfile) *pwrds)
          (errorprintf "*** %p: base 16#%x, length %d bytes"
                       (foreach x in entries* collect (first (car x)))
                       codebase* codesize*))
%       (setq currentoffset* (times2 2 currentoffset*)) % halfwords
        (return (if lapreturnvalue* (mkitem code-tag lapreturnvalue*) NIL))))

(de look-for-links (u)
       (cond ((null u) nil)
             ((atom (car u)) (look-for-links (cdr u)))
             ((eq (caar u) '*linke)
                  (cond ((flagp (caddar u) 'foreignfunction) t)
                         (t (look-for-links (cdr u)))))
             ((eq (caar u) '*link)
                  (cond
                    ((eq  (cadar u) 'catchsetup) t) % is opencode
                         % but problems will arise in throw if no
                         % allocation takes place
                    ((memq (cadar u) '(fast-codeapply fast-idapply)) t)
                    ((get (cadar u) 'opencode) (look-for-links (cdr u)))
                    (t t)))
             (t (look-for-links (cdr u)))
)      )

(de saveentry (x)
  (prog (curoff)

   (setq curoff (wtimes2 currentoffset* 2)) % is a halfword counter
   (return 
   (cond ((equal (second x) '***code**pointer***)
         (setq lapreturnvalue*
                (if *writingfaslfile curoff (wplus2 codebase* curoff))))
        ((not *writingfaslfile)
          (setq entries* (cons (cons (rest x) curoff) entries*))
          (unless lapreturnvalue*
            (setq lapreturnvalue* (wplus2 codebase* curoff))))
        ((equal (second x) '**fasl**initcode**)
			(setq initoffset* curoff))
        ((flagp (second x) 'internalfunction)
          (put (second x) 'internalentryoffset curoff))
        (t (findidnumber (second x))
           (dfprintfasl (list 'putentry (mkquote (second x))
                         (mkquote (third x)) curoff)))))))

(de defineentries ()
  (foreach x in entries* do
           (putd (first (car x)) (second (car x))
                 (mkcode (wplus2 codebase* (cdr x))))))

(de systemfaslfixup ()
  (prog (x type label place fwdir )
        (setq currentoffset* (times2 2 currentoffset*)) % halfwords
        (while forwardinternalreferences*
          (setq fwdir (car forwardinternalreferences*))
          (setq type (car fwdir) label (cadr fwdir) place (caddr fwdir))
          (if (null (get label 'internalentryoffset))
            (errorprintf "*** %r not defined in this module, call is incorrect"
                         (cadr fwdir))
            (cond ((eq type 'CALL)
                   (putmem (wplus2 codebase* place)
				 (get label 'internalentryoffset)))
                  (t  (error 0 (list "Strange forward internal reference: "
                                  fwdir)))))
          (setq forwardinternalreferences* (cdr forwardinternalreferences*)))
        (mapobl (function (lambda (y) (remprop y 'internalentryoffset))))
))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                      DepositInstruction( X )                         %
%----------------------------------------------------------------------%
%                                                                      %
%  This converts one CMACRO for the target lisp machine into actual    %
%  binary code and inserts it in memory.                               %
%                                                                      %
%  See AsmOutOneInstrucion for analagous code to print that CMACRO     %
%   as assembly code.                                                  %
%                                                                      %
%  Both DepositInstruction and AsmOutOneInstruction are table driven.  %
%  They use the same opcode access functions (such as DisplacementOf), %
%  and they refer to same opcode formats (RR, RX, etc.).               %
%                                                                      %
%  DepositInstruction uses its own set of tables to insert the         %
%  components for a given opcode format into memory.                   %
%                                                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de depositinstruction (lapinstruction*) 
 (prog (bytelength ochan type fn format operands hwoffsetwithininstruction* 
        opcode* opcodenumber* operands* operand*1 operand*2 operand*3 
        operand*4 operand*5 operand*6  lng oldbto) 

  % special cases:  labels, other atoms (error), and entries

  (cond ((labelp lapinstruction*)
		      (cond (*listlap (listlapcodeoffset) 
            		  	      (lapprintf "%t%p%n" 18 lapinstruction*))) 
		       (return nil)) 
       ((eqcar lapinstruction* 'using) (return nil))
       ((atom lapinstruction*) 
            (stderror (bldmsg "DepositInstruction given non-label atom: %p" 
                        lapinstruction*)))) 

  % analyze opcode

  (setq opcode* (car lapinstruction*)) 
  (setq type (get opcode* 'opcodetype)) 
  (setq opcodenumber* (get opcode* 'opcodenumber)) 
  (setq hwoffsetwithininstruction* 0) 
 
  (when *writingfaslfile
        (setq oldbto bittableoffset*)
  	(setq lng (onelaplength lapinstruction*))
        (updatebittable lng 0)
        (setq bittableoffset* oldbto))


  % check for comments and pseudo-ops such as FullWord, String, etc.

  (when  (null type) 
      (setq fn (get opcode* 'lapdepositfunction)) 
      (cond ((null fn) 
         (stderror (bldmsg "%w: %n   %w: %p" 'depositinstruction 
                   	"No type or LapDepositFunction for" 
                   	lapinstruction*)))) 

      (when  *listlap 
	 (when (not (memq opcode* '(samelinecomment comment)))
			 (listlapcodeoffset)) 
          (when (weq opcode* '*entry) (listlapinstruction lapinstruction*)))
      (apply fn (list lapinstruction*)) 
      (when (and *listlap (wneq opcode* '*entry)) 
         (listlapinstruction lapinstruction*))
      (setq lapdepositptr* (wplus2 lapdepositptr* (wtimes2 2 
                           hwoffsetwithininstruction*))) 
      (when *writingfaslfile (setq bittableoffset* (plus2 oldbto lng )))
      (return nil)) 

  (setq format (get type 'lapinstructionformat)) 
  (when  (null format) 
     (stderror (bldmsg "DepositInstruction: no lap format for: %p" 
                      lapinstruction*))) 

  % fetch and prepare to manipulate operands

  (setq operands* (cdr lapinstruction*)) 

  (cond ((and (numberp (car format)) (wneq (car format) (length operands*))) 
     (stderror (bldmsg "DepositInstruction: wrong number of args: %p" 
                      lapinstruction*)))) 
  %
  %  Set fluid variables for evaluation of template forms
  %
  (setq operands operands*) 
  (cond ((numberp (car format)) 
     (cond ((not (null operands)) 
     (progn (setq operand*1 (car operands)) 
        (cond ((not (null (setq operands (cdr operands)))) 
        (progn (setq operand*2 (car operands)) 
           (cond ((not (null (setq operands (cdr operands)))) 
           (progn (setq operand*3 (car operands)) 
              (cond ((not (null (setq operands (cdr operands)))) 
              (progn (setq operand*4 (fourth operands)) 
                 (cond ((not (null (setq operands (cdr operands)))) 
                 (progn (setq operand*5 (fifth operands)) 
                    (cond ((not (null (setq operands (cdr operands)))) 
                    (progn (setq operand*6 (sixth operands)) 
                       (cond ((not (null (setq operands (cdr operands)))) 
                          (stderror (bldmsg 
                           "AsmOutInstruction: more than 6 operands: %p" 
                                    lapinstruction*)))))))))))))))))))))))) 

  (when  *listlap (listlapcodeoffset)) 

  (eval (get opcode* 'lapdepositopcode)) 
  (foreach fieldformat in (cdr format) do (eval fieldformat)) 
        %  The field formats use formatting routines defined below.

  (when *listlap (listlapinstruction lapinstruction*)) 

  (setq lapdepositptr*
	 (wplus2 lapdepositptr* (wtimes2 2 hwoffsetwithininstruction*)))

  (when *writingfaslfile (setq bittableoffset* (plus2 oldbto lng ))))) 
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de depositdatum (lapdatum*) 
 (prog (bytelength ochan 
              type fn format operands 
              hwoffsetwithininstruction* 
              opcode* 
              operands* 
              operand*1 operand*2 operand*3 
              operand*4 operand*5 operand*6) 

  % special cases:  labels and other atoms (error)

  (cond ((labelp lapdatum*) 
   (progn (cond (*listlap 
          (progn (listlapdataoffset) 
              (lapprintf "%t%p%n" 18 lapdatum*)))) 
       (return nil))) 
  ((atom lapdatum*) 
       (stderror (bldmsg "DepositDatum given non-label atom: %p" 
                        lapdatum*)))) 

  % analyze opcode

  (setq opcode* (car lapdatum*)) 
  % check for comments and pseudo-ops such as FullWord, String, etc.
  (setq fn (get opcode* 'lapdepositfunction)) 
  (cond ((null fn) 
     (stderror (bldmsg "%w: %n   %w: %p" 'depositdatum 
                      "No LapDepositFunction for" lapdatum*)))) 

  (setq hwoffsetwithininstruction* 0) 
  (cond ((and *listlap (not (memq opcode* '(samelinecomment comment)))) 
     (listlapdataoffset))) 
  (apply fn (list lapdatum*)) 
  (cond (*listlap 
     (listlapinstruction lapdatum*))) 
  (setq lapdepositptr* (wplus2 lapdepositptr* (wtimes2 2 hwoffsetwithininstruction*))) 
  (return nil)))
 

(de depositfaslinitcommands (u) 
 (prog (pr name command) 
  (setq pr (assoc '*entry u)) 
  (setq name (cadr pr)) 
  (cond ((weq name '**fasl**initcode**) 
   (progn (setf (getmem faslptr*) 3)    % me and 2 others
       (setq faslptr* (wplus2 faslptr* 4)) 
       (setf (getmem faslptr*) executenow) 
       (setq faslptr* (wplus2 faslptr* 4)) 
       (setf (getmem faslptr*) deletenow) 
       (setq faslptr* (wplus2 faslptr* 4)))) 
  ((flagp name 'internal) 
   (progn (setf (getmem faslptr*) 2)    % me and 1 other
       (setq faslptr* (wplus2 faslptr* 4)) 
       (setf (getmem faslptr*) donothing) 
       (setq faslptr* (wplus2 faslptr* 4)))) (t 
  % normal case
   (progn (setf (getmem faslptr*) 4)    % me and 3 others
       (setq faslptr* (wplus2 faslptr* 4)) 
       (setf (getmem faslptr*) installnow) 
       (setq faslptr* (wplus2 faslptr* 4)) 
       (setf (getmem faslptr*) (findidnumber name)) 
       (setq faslptr* (wplus2 faslptr* 4)) 
       (setf (getmem faslptr*) (findidnumber (caddr pr))) 
       (setq faslptr* (wplus2 faslptr* 4)))))))
 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  Print Routines used for LAP listing                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de listlapheader (u) 
 (prog (x) 
  (lapprintf "%f%n-------------------------------------------------%n") 
  (setq x (car u)) 
  (cond ((weq (car x) '*entry) 
       (lapprintf "%n  %w procedure %w of %w argument%w.%n%n" 
                  (cond ((wneq (caddr x) 'expr) (caddr x)) (t "")) 
                  (cadr x) (cadddr x) (cond ((wneq (cadddr x) 1) "s") (t "")))) (t 
  (errorlapprintf "Arg to LAP does not start with !*ENTRY: %p" u)))))
 

(de listlapcodeoffset nil 
 (cond ((numberp codeblockbase*) 
      (lapprintf "%f %w%t" 
                 (hexstr (wdifference lapdepositptr* codeblockbase*) 4 '! ) 8)) (t 
 (lapprintf "%f ****   "))))

(de listlapdataoffset nil 
 (lapprintf "%f %w%t" 
             (hexstr (wdifference lapdepositptr* datablockbase*) 4 '! ) 8))

(de listlaphalfword (hw) 
 (lapprintf "%w " (hexstr hw 4 '!0)))

(de listlapinstruction (x) 
 (prog (ochan) 
  (setq ochan (wrs listlapout*)) 
  (cond ((memq (car x) '(samelinecomment comment)) 
       (printsamelinecomment 28 (cdr x))) (t 
  
   (progn (printf "%t%p " 24 (car x)) 
       (cond ((wlessp (posn) 29) (printf "%t" 29))) 
       (printf "%p%n" (cdr x))))) 
  (wrs ochan)))
 

(de hexstr (n fieldsize filler) 
 (prog (outputbase* s) 
  (cond ((not (numberp n)) (return (bldmsg "%w" n)))) 
  (setq outputbase* 16) 
  (setq s (cddddr (explode (bldmsg "%w" n)))) 
  (while (wlessp (length s) (wplus2 fieldsize 1)) (setq s (cons filler s))) 
  (return (compress (cons '!" s)))))
 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    DepositEntry( E )                                  %
%                    SaveEntry   ( E )                                  %
%                    DefineEntriesForUser  ()                           %
%-----------------------------------------------------------------------%
%               code to processe entry declarations                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(put '*entry 'lapdepositfunction 'depositentry)

(de depositentry (e) 
 (prog (base pr) 
  (cond ((numberp codeblockbase*) 
       (printf "%f*** Second entry within one function: %p%n" (cadr e))) (t 
  (setq codeblockbase* lapdepositptr*))) 

  (saveentry e)                         % assumes CodeBlockBase!* has been set

  % second entries are very rare -- ok to repeat this work for them
  (setq pr (assoc '(reg baser) addressesknowntolap*)) 
  (cond ((pairp pr) (rplacd pr codeblockbase*)) (t 
  (setq addressesknowntolap* 
	(cons (cons '(reg baser) codeblockbase*) addressesknowntolap*)))) 

  (setq pr (assoc '(implicitreg baser) addressesknowntolap*)) 
  (cond ((pairp pr) (rplacd pr codeblockbase*)) (t 
  (setq addressesknowntolap*
		 (cons (cons '(implicitreg baser) codeblockbase*)
			 addressesknowntolap*)))) 
))

(de defineentriesforuser nil 
 (prog nil 
  (cond ((cdr entries*) 
      (ttyprintf "%fMultiple entries: %p%n" 
                 (mapcar entries* (function caar))))) 
  (cond (*depositlap 
     (foreach x in entries* do 
         (putd (car (car x)) (cadr (car x)) (mkcode (cdr x))))))))
 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%      ResolveBranchesAndLabels( CodeList )
%------------------------------------------------------------------------
%  Sets CodeBlockSize!*,    PseudoHeapSize!*,
%       CodeBlockOffsets!*, PseudoHeapOffsets!*
%
%  It may add spacers to get fullword alignment.
%  The resulting data list and code list will look like:
%
%                    ====================================================
%  PSEUDO-HEAP DATA: |     F6     | 0 0 |       number of items of data |
%                    |  (8 bits)  (2 bits)         (22 bits)            |
%                    |--------------------------------------------------|
%                    |  item (logically) inside some constant(s)        |
%                    |                     ...                          |
%                    |  item (logically) inside some constant(s)        |
%                    ====================================================
%
%  Note that the data may be anywhere in memory.
%  It is consistent to assume that it immediately preceeds the code.
%  It is also consistent to assume that it is in a noncompacting heap.
%
%
%                    ====================================================
%              CODE: |     F6     | 1 0 |       number of items of code |
%                    |  (8 bits)  (2 bits)         (22 bits)            |
%                    |--------------------------------------------------|
%  A, B, C, D are    |  A: MiscLiterals        |  B: HalfWordIdOffsets  |
%  offsets in bytes  |--------------------------------------------------|
%  from BASE.        |  C: FullWordIdOffsets   |  D: Constants          |
%                    |--------------------------------------------------|
%                    |      32-bit address of Contents of Constants     |
%                    |--------------------------------------------------|
%                    |     0 : reserved        |  NumberOfArguments     |
%                    |--------------------------------------------------|
%              BASE: |  ...code...                                      |
%                    |  .......... (ID offsets may appear in this code  |
%  Calls arrive at   |  ..........  as displacements, if they are small |
%  BASE or BASE+10   |  ..........  enough.  Larger ones are referenced |
%                    |  ..........  via the IDOffsets stored below.)    |
%                    |--------------------------------------------------|
%                 A: |  ...misc literals...                             |
%                    |--------------------------------------------------|
%                 B: |  HalfWordIdOffset       |  HalfWordIdOffset      |
%                    |        ...              |        ...             |
%                    |  HalfWordIdOffset       |  HalfWordIdOffset      |
%                    |--------------------------------------------------|
%                 C: |  FullWordIdOffset                                |
%                    |       ...                                        |
%                    |  FullWordIdOffset                                |
%                    |--------------------------------------------------|
%                 D: |  item at top level of a constant                 |
%                    |       ...                                        |
%                    |  item at top level of a constant                 |
%                    ====================================================
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de resolvebranchesandlabels (codelist) 
 (prog (sizes) 

  (setq codeblocksize* 16)              % account for 4 words appended below
  (resolvebranchesandlabelsincode codelist) 	% sets LapData!*

  (setq datablocksize* 4)               % account for word appended below
  (resolvelabelsindata lapdata*) 

  (cond ((and codeblockoffsets* (not (dispinump (cdar codeblockoffsets*)))) 
     (correctlonglap lapcode* lapdata*))) 

  %   DATA

  %  DataBlockSize!* is in bytes -- deposit size in items
  %  LapData!* := list( 'Fullword, DataBlockSize!* / 4 ) . LapData!*;
  (setq lapdata* 
                     
                (cons (list 'fullword (makemkitem 'hbpsitems (wquotient datablocksize* 4))) lapdata*)) 

  %   CODE

  (setq lapcode* 
  %   list( 'Fullword, CodeBlockSize!* / 4 )   % deposit size in items
      
            
    (cons (list 'fullword (makemkitem 'hbpsitems (wplus2 8388608 (wquotient codeblocksize* 4)))) 
             
             
             
             
    (cons (list 'halfword (cdr (assoc literalslabel* codeblockoffsets*)) (cdr (assoc halfwordidoffsetslabel* codeblockoffsets*)) (cdr (assoc fullwordidoffsetslabel* codeblockoffsets*)) (cdr (assoc constantslabel* codeblockoffsets*))) 
             
    (cons (list 'fullword (cdr (assoc contentsofconstantslabel* datablockoffsets*))) lapcode*))))))

 

(de resolvebranchesandlabelsincode (codelist) 
 (prog (offsettobase offset u l insertionpoint pendinglabels) 
  (setq offset codeblocksize*)          % start at current location
     % offset first holds offsets from start of block, then from base
  (while codelist 
  (progn (setq u (car codelist)) 
      (cond ((not (labelp u)) 
       (progn                           % insert dummy halfwords if necessary to maintain alignment
           (cond ((needsfullwordalignmentp u) 
              (cond ((not (zerop (wremainder offset 4))) 
              (progn (rplacd insertionpoint 
                          (cons '(halfword 0) (cdr insertionpoint))) 
                  (setq offset (wplus2 offset 2)) 
                  (cond ((needsdoublewordalignmentp u) 
                      (cond ((not (zerop (wremainder offset 4))) 
                      (progn (rplacd insertionpoint 
                                  (cons '(fullword 0) (cdr insertionpoint))) 
                          (setq offset (wplus2 offset 4)))))))))))) 
           (cond ((and pendinglabels (null offsettobase)) 
              (stderror "Labels preceed entry.  See LapCode!* "))) 
           (foreach pendinglabel in pendinglabels do 
               (setq codeblockoffsets* (cons 
                                     (cons pendinglabel offset) codeblockoffsets*))) 
           % might need to insert halfword after this instruction
           (setq insertionpoint codelist) 
           (setq pendinglabels nil) 
           (cond ((eqcar u '*entry) (setq offsettobase offset) (setq offset 0)))
           (setq l (onelaplength u)) 
           %  L better be even  (test for that??)
           (setq offset (wplus2 offset l))))
       ((weq u contentsofconstantslabel*) 
       (progn (setq lapdata* (cons (car codelist) (cdr codelist))) 
           (rplaca codelist nil) 
           (rplacd codelist nil))) (t 
      %  Don't assign location to label yet --
      %  the instruction being labelled might need alignment.
      (setq pendinglabels (cons u pendinglabels)))) 

      (setq codelist (cdr codelist)))) 

  %  Clean up at end -- pending labels for fullword id offsets and
  %  top level of constants should be fullword aligned, even if no
  %  such objects are present (to avoid confusing routines that
  %  analyze code).  It doesn't hurt to align other labels.

  (cond ((and insertionpoint (not (zerop (wremainder offset 4)))) 
   (progn (rplacd insertionpoint (cons '(halfword 0) (cdr insertionpoint))) 
       (setq offset (wplus2 offset 2))))) 
  (foreach pendinglabel in pendinglabels do 
      (setq codeblockoffsets* (cons 
                            (cons pendinglabel offset) codeblockoffsets*))) 

  %  Round declared size up to even fullword, since next fn will FW-align
  (setq codeblocksize* (wplus2 offsettobase offset)) 
  (setq codeblocksize* (wtimes2 4 (wquotient (wplus2 codeblocksize* 3) 4)))))
 

(de resolvelabelsindata (datalist) 
 (prog (l) 
  (foreach u in datalist do 
   (cond ((labelp u) 
       (setq datablockoffsets* (cons (cons u datablocksize*) datablockoffsets*))) (t 
   
   (progn (setq l (onelaplength u)) 
       (cond ((wneq l (wtimes2 4 (wquotient l 4))) 	% paranoia?
          (progn (errorprintf "Length of data %p = %p, not divisible by 4" 
                           u l) 
              (setq l (wtimes2 4 (wquotient (wplus2 l 3) 4)))))) 
       (setq datablocksize* (wplus2 datablocksize* l))))))))
 

% syntax:  (DC (F 33))      DC  F'33'
%          (DC (3 F 33))    DC  3F'33'
%          etc.

(de needsfullwordalignmentp (u) 
 
  (or (eqcar u 'fullword) (and (member (car u) '(ds dc)) 
       (pairp (cadr u)) 
       
          (or (member (car (cadr u)) '(d f)) (and (numberp (car (cadr u))) 
               (member (cadr (cadr u)) '(d f)))))))

(de needsdoublewordalignmentp (u) 
 (and (member (car u) '(ds dc)) 
  (or (eqcar (cadr u) 'd) 
    (and (numberp (car u)) (eqcar (cdr u) 'd)))))

(de correctlonglap (lapcodelist* lapdatalist*) 
 (stderror (bldmsg "LAP: lbl & offset = %p  See %l" 
                  (car codeblockoffsets*) 
                  '(codeblockoffsets* datablockoffsets* 
                    lapcodelist* lapdatalist*))))

(de findlabels (u)
     (setq codesize* 0)
     (setq label-alist* nil)
     (foreach x in u do (setq codesize* (plus2 (onelaplength x) codesize*))))

(de onelaplength (u) 
 (prog (x) 
  (return
  (cond ((labelp u) (push (cons u (plus2 (wtimes2 2 currentoffset*) codesize*))
				 label-alist*) 0) 
        ((atom u) (laperrorprintf "%f***** Wierd atom in LAP code: %p%n" u) 0)
	((eq (car u) '*entry) (setq u (second u))
                (put u 'internalentryoffset (wtimes2 2 currentoffset*))
                            0)
	((eqcar u 'using) 
		(unless *using* (setq *using* (cons (cddr u) 
				(wplus2 (wtimes2 2 currentoffset*) codesize*))))
		0)
        ((setq x (get (car u) 'opcodetype)) 
         (or (get x 'laplength) (computeonelaplength u))) 
        ((setq x (get (car u) 'laplength)) 
         (cond ((inump x) x)
                (t (apply x (list u)))))
        (t % minor klugde for long constants
            (progn (cond ((and (weq (length u) 3) (fixp (setq x (caddr u)))
		 (not (dispinump x))) 
            (prog (y) 
               (errorprintf "Doing kludge in OneLapLength on %p" u) 
               (rplaca (cddr u) (setq y (stringgensym))) 
               (nconc lapcode* (list y (list 'fullword x)))))) 4))))))
            
        % should be some sort of parameter
 

(de onelaplengthwithnosideeffects (u) 
 (prog (x) 
  (return (cond ((labelp u) 0) 
    ((atom u) 
     (progn (laperrorprintf "%f***** Wierd atom in LAP code: %p%n" u) 0)) 
         
    ((setq x (get (car u) 'opcodetype)) 
         (or (get x 'laplength) (computeonelaplength u))) 
    ((setq x (get (car u) 'laplength)) 
         (cond ((inump x) x) (t (apply x (list u))))) (t 4)))))
    % should be some sort of parameter
 

(de computeonelaplength (u) 
 (cond ((memq (car u) '(ds dc)) (dslaplength (car u) (cadr u)))
       (t (errorprintf "ComputeOneLapLength using 0 for %p%n" u) 0)))
          

 % NB: (DS "0F") and (DS "0D") could increase length via forced alignment
(de dslaplength (d u) 
 (cond  ((equal u "0F") 
		(cond ((eq currentoffset* 0)
			(if (eq 2 (wand codesize* 3)) 2 0))
		      (t (if (eq 2 (wand (wor (wshift currentoffset* 1)
					      codesize*) 3))  2 0))))
	((not (pairp u)) (errorprintf "DSLapLength using 0 for %w %p%n" d u) 0)
        ((numberp (car u)) (wtimes2 (car u) (dslaplength d (cdr u))))
       	((weq (car u) 'd) 8)
        ((weq (car u) 'f) 4)
        ((weq (car u) 'h) 2)
        ((weq (car u) 'al) (cadr u))
        ((weq (car u) 'cl) (cadr u))
        ((weq (car u) 'xl) (cadr u))
        (t (errorprintf "DSLapLength using 0 for %w %p%n" d u) 0)))
          

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Routines to calculate the length of specific (pseudo-)opcodes      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(deflist '((*entry 0)
          (string lapstringlength) 
          (ebcdic lapstringlength) 
          (ascii lapstringlength) 
          (fullword lapwordlength) 
          (halfword laphalfwordlength) 
          (byte lapbytelength) 
          (float 8) 
          (comment 0) 
          (samelinecomment 0)) 
          'laplength)


(de lapstringlength (u) (wtimes2 4 (wquotient (wplus2 
 (strlen (strinf (cadr u))) 5) 4)))
        % strlen yields number of chars - 1

(de lapwordlength (u) (wtimes2 4 
 (length (cdr u))))

(de laphalfwordlength (u) (wtimes2 2 
 (length (cdr u))))

(de lapbytelength (u) (wtimes2 4 (wquotient (wplus2 
 (length (cdr u)) 3) 4)))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                         Encode 4-bit Operands                        %
%----------------------------------------------------------------------%
%                  LapEncodeFourBitMask  ( Mask )                      %
%                  LapEncodeRegister     ( Reg )                       %
%                  LapEncodeLength       ( Operand, 16 )               %
%                                                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(deflist
  '((r0  0)(r1  1)(r2  2)(r3  3)(r4  4) (r5  5)(r6  6)
        (r7  7)(r8  8)(r9  9) (r10 10)(r11 11)(r12 12)
           (f0 0)                       % Floating point register 0
           (f2 2)                       % Floating point register 2
           (f4 4)                       % Floating point register 4
           (f6 6)                       % Floating point register 6
        (r13 13)(r14 14) (r15 15) (t1 10) (t2 11) (st 13)) 'registername)

(put 'nil 'registername 9)

(de lapencodefourbitmask (mask) 
  (cond ((stringp mask) (stderror "LapEncodeFourBitMask: %p" mask)) (t 
  (lapencodenumber (operandfourbitmaskof mask) "mask" 15))))

(de lapencoderegister (reg) 
 (prog (regnum regname) 
  (when  (null reg) (return 0) )
  (cond ((or (atom reg) (not (memq (car reg) '(reg implicitreg)))) 
     (stderror (bldmsg "%p is not a register" reg)))) 
  (setq reg (cadr reg)) 
  (when (numberp reg) 
	(when  (and (wgeq reg 1) (wleq reg 5)) (return (wplus2 3 reg)))
	(stderror (bldmsg "%p is an illegal register" reg)))
  (when  (not (idp reg)) (stderror (bldmsg "%p is an illegal register" reg)))

  (setq regnum (get reg 'registernumber)) 
  (cond ((numberp regnum) (return regnum))) 

  (setq regname (get reg 'registername)) 
  (cond ((numberp regname) (return regname))) 
  (cond ((not (null regname)) 
  (progn (setq regnum (get regname 'registernumber)) 
      (cond ((numberp regnum) (return regnum)))))) 
  (stderror (bldmsg "LAP: %p is an illegal register" reg))))
 

(de lapencodelength (operand limit) 
 (prog (expansion) 
  (cond ((atom operand) 
  (progn (setq expansion (expandatomicoperand operand)) 
      (cond ((atom expansion) 
         (stderror (bldmsg "LapEncodeLength: %p" operand)))) 
      (setq operand expansion)))) 
  (cond ((eqcar operand 'lengthbased) 
     (return (sub1 (lapencodenumber (cadr operand) 'length limit))))) 
  (stderror (bldmsg "LapEncodeLength: %p %w" operand limit))))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    Encode 8-bit Operands                             %
%----------------------------------------------------------------------%
%               LapEncodeImmediateByte ( Operand )                     %
%               LapEncodeLength        ( Operand, 256 )  [see above]   %
%                                                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapencodeimmediatebyte (operand) 
  (cond ((stringp operand) (stderror "LapEncodeImmediateByte: %p" operand)) (t 
  (lapencodenumber (operandimmediatebyteof operand) "immediate byte" 255))))
                        

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    Encode 16-bit Operands                            %
%----------------------------------------------------------------------%
%                  LapEncodeDispBase ( Operand )                       %
%                                                                      %
%     Given an object with an explicit or implicit base register and   %
%     displacement, resolve and encode those values in one halfword.   %
%                                                                      %
%     high order                                            low order  %
%     -0------------3-4-------------------------------------------15-  %
%     |   base reg   |                                displacement  |  %
%     ---------------------------------------------------------------  %
%                                                                      %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapencodedispbase (operand) 
 (prog (expansion disp breg) 
  (cond ((numberp operand) 
     (return (lapencodenumber operand "displacement" 4095)))) 
  (cond ((atom operand) 
  (progn (setq expansion (expandatomicoperand operand)) 
      (cond ((atom expansion) 
         (stderror (bldmsg "LapEncodeDispBase: %p" operand)))) 
      (setq operand expansion)))) 

  (setq disp (lapencodenumber (operanddisplacementof operand) "displacement" 4095)) 
                           
  (setq breg (lapencoderegister (operandbaseregisterof operand))) 
  (return (wor (wshift breg 12) disp))))
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapencodenumber (x type limit) 
 (prog (y)                              % ExpressionCount!*;
  (setq y (resolvewconstexpression x)) 
  (cond ((numberp y) (cond ((and (wgeq y 0) (wleq y limit)) (return y))))) 
  (errorprintf "***** LapEncodeNumber: 0 used for (%w) %p instead of %p" 
               type x y) 
  (return 0)))
 

(de lapencodeproblem (operands) 
 (prog (s) 
  (cond ((null operands) (stderror "Unspecified problem in LAP"))) 
  (setq s (bldmsg "%p" (car operands))) 
  (foreach operand in (cdr operands) do (setq s (bldmsg "%w,%p" s operand))) 
  (lapprintf s) 
  (stderror s)))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Listing routines for LAP -- must be compiled to handle args correctly %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de lapprintf (template a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13 a14) 
 (prog (ochan) 
  (setq ochan (wrs listlapout*)) 
  (printf template a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13 a14) 
  (wrs ochan)))
 

(de laperrorprintf (template a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13 a14) 
 (prog (ochan) 
  (errorprintf template a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13 a14) 
  (cond ((and *listlap listlapout*) 
  (progn (setq ochan (wrs listlapout*)) 
      (printf template a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11 a12 a13 a14) 
      (wrs ochan))))))
 

(de makerelocword (reloctag relocinf)
  (wplus2 (wshift reloctag 30) (field relocinf 2 30)))

(de makerelocinf (reloctag relocinf)
  (wplus2 (wshift reloctag 22) (field relocinf 10 22)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

