%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:XXX-CMAC.SL
% Description:  Patterns and predicates for M68K PSL cmacro expansion
% Author:       G. Q. Maguire Jr. and M. L. Griss
% Created:      16 January 1982
% Modified:
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% Based on VAX-CMAC.SL of 14 Jan 1982 by Eric Benson
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 1-Jun-1989 (Winfried Neun)
%  Put in some improvements for arithmetic using the 68020 instructions.
%  Changed *ALLOC not to clean up stack space, using shift for inf instead of
%  mask and 'and'
% 23-Oct-88 (Tony Hearn)
%  Removed definition of mkcode, since it is defined in datamachine.sl.
%  1-Sep-88 (Julian Padget)
%  Replaced explicit "15" by "code-tag" in mkcode.
% 12-Oct-87 (Harold Carr)
%  Moved foreignlink cmacro from xxx-asm.sl to here, where it belongs.
% 10-Jun-87 (Leigh Stoller)
%  Modifed *Alloc to push nils all the time, not to loop. This is much faster.
% 20-Oct-86 (Leigh Stoller)
%  Added version of *Alloc that initializes the stack space to nils.
% 18-Aug-86 (Leigh Stoller)
%  Changed the opencode for fast-codeapply so that it tag strips the address
%  before doing the jump.
% 09-Aug-86 (Leigh Stoller)
%  Added two new temp A registers called t5,t6. Needed so that we can devote
%  separate registers for cmacros, and others for functions like fastapply
%  and fastbind. Previously, the move cmacro was stepping the registers used
%  by thses functions. Also removed the patch made on 18-Jul that flagged
%  reg 14 as being real. All places where reg 14 was being used as a real reg
%  have been modified to use temp regs, so the patch is not needed anymore.
% 02-Aug-86 (Leigh Stoller)
%  Modified all references to fake reg 14 to be real reg t2.
% 29-Jul-86 (Leigh Stoller)
%  Added register number 14 back into Aregp util we can rewrite the code that
%  is dependent on this reg being real. This tied to the REG anyreg which is
%  also temporarily munged
% 24-jul-86 (Leigh Stoller)
%  Added a new version of the MEMORY anyreg that is designed to make most
%  of the decisions. Previously, it was highly dependent of the pattern in
%  ipc-comp. The new pattern does very little, while the new new anyreg is
%  based on the original Sun cmacros.
%  under all cases.
% 24-Jul-86 (Leigh Stoller)
%  Modified the REG anyreg so that it was not doing a displacement off the
%  Zero reg for the fakereg case.
% 21-Jul-86 (Leigh Stoller)
%  Modified Aregp and Dregp to reflect the change to maxrealregs being 6.
%   Registers 7 -> 14 are no longer real regs and should not be returned as
%   such. Added two new predicates TempAregp and TempDregp for use by the
%   compiler when looking for temp regs. See note below for an explanation
% 18-Jul-89 (Leigh Stoller)
%  Added another t reg called t0 to be a3. Also modified the reg function to
%  flag reg 14 as a real reg instead of fake.
% 09-Jul-86 (Leigh Stoller)
%  Added a second version of bitmask for use in cross compiliation from a
%   vax.
% 28-Jun-86 (Leigh Stoller)
%  Added versions of car and cdr that do tag stripping.
% 21-Jun-84 (Vicki O'Day)
%  Duplicated Bob Morain's changes for HP-UX
% 25-Apr-84 13:09:40 (Brian Beach)
%  Removed definitions of TAGNUMBER and INUMP. They are defined portably in
%  PASS-1-LAP.
% 25-Apr-84 08:20:06 (Brian Beach)
%  Removed CODEAPPLY 0-14.
% 27-Mar-84 12:58:38 (Jim Ambras/CRC)
%  Corrected (inump inump) case in !*jumpif cmacro pattern.  Used the
%  default case.  Cures problem of (eq 1 0) blowing up when compiled.
% 20-Mar-84 13:53:57 (Brian Beach)
%  Added FAST-APPLY-LOAD Cmacro.  Removed Wthings.
%  Added open code for FAST-IDAPPLY and FAST-CODEAPPLY.
% 27-Feb-84 14:49:43 (Brian Beach)
%  Included Mike's CODEAPPLY stuff.
% 15-Feb-84 (Steve Lowder)
%  Removed all usage of Wvar, WConst, WString, and Warray.
% Nov 83 (Sam Sands)
%  Revised for the D register model
% 30-Jul-83 Nancy Kendzierski
%  Corrected *WXOR bug which used an ori in one case instead of an eori.
% 26-Jul-83 Nancy Kendzierski
%  Added more RemProps to handle *WTIMES2 no longer being a cmacro.
%  (RemProp 'Wtimes2 'MemModFn) and (RemProp '*MpyMem 'UnMemMod)
% 21-Jul-83 Steve Lowder
%  Added line to IDapplys to mask the tag from the ID before multiplying
%  by 6.
% 1-Jul-83 Steve Lowder
%  Changed two addressing modes to reflect the Motorola syntax, changed
%  autodecrement to predecrement, and autoincrement to postincrement.
%  Most of the Cmacro Pattern Tables were changed to optimize the code
%  generation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This is needed to set up real 32 bit arithmetic.

(loadtime (progn
        (RemProp 'Wtimes2 'OpenFn)     % So need explicit code
        (RemProp 'Wtimes2 'MemModFn)   % Since no longer a cmacro
        (RemProp '*MpyMem 'UnMemMod)   %   "    "   "    "    "
 ))  %  (RemProp 'WQuotient 'OpenCode)
     %  (RemProp 'WRemainder 'OpenCode)))


(fluid '(AddressingUnitsPerItem       % Constants defined in data machine.
     *ImmediateQuote
     ))


% The following terminal operands try to follow the same meanings as
% those outlined in the Motorola manuals.

(flag '(Immediate                        % #xxx
        UnImmediate                      % used for (unimmediate (immediate x))
        Indirect                         % (An)
        displacement                     % d(An)
        predecrement                     % -(An)
        postincrement                    % (An)+
        Indexed                          % d(An,Dn)
        absolute                         % absolute.long
        reglist)
        'TerminalOperand)


(setf *ImmediateQuote NIL)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                    NECESSARY FUNCTIONS
% These are useful macros for defining Cmacro pattern tables.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The Second version of this function is need for cross compiliation from the
%  the vax since the vax lshift is weird. Once the compiler is running on the
%  target machine, this module must be recompiled with the second version
%  commented out. The first version is legal for 68K's
%
%(ds BitMask (StartingBit Length)
% (land (lshift 16#ffffffff (minus StartingBit))
%    (lshift 16#ffffffff (difference BitsPerWord
%                    (plus StartingBit Length)))))

(ds BitMask (StartingBit Length)
  (lor (lshift 16#1 (difference Bitsperword (plus startingbit 1)))
       (land (lshift 16#7fffffff (minus StartingBit))
         (lshift 16#ffffffff (difference BitsPerWord
                    (plus StartingBit Length))))))


(ds NegMask (Length) (lshift 16#ffffffff Length))

(ds ShiftAmt (StartingBit Length)
  (difference BitsPerWord
          (plus StartingBit Length) %always positive.
          ))

(ds MakeTag (tag) (lshift (land tag 16#ff) infbitlength))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% REGISTER HANDLING PREDICATES%
% The following predicates are for the 68000, they try to catagorize the
% constraints relating to legal operands in 68000 assembly language.
% There are a number of common predicates already available,
% they are defined the the file PC:Common-Predicates.SL

% The following predicates test for integers in the ranges outlined.
% All numbers are expressed in base 16.
%
% InumP        7FFFFFFF >= x >= -80000000          ( 32 bits )
% PosInumP     7FFFFFFF >= x >          0          ( 32 bits )
% NegInumP            0 >  x >= -80000000          ( 32 bits )
% Geq16P       7FFFFFFF >= x >=        10          ( 32 bits )
% LeqM16P           -10 >= x >= -80000000          ( 32 bits )
% Gt8P         7FFFFFFF >= x >          8          ( 32 bits )
% LtM8P              -8 >  x >= -80000000          ( 32 bits )
% PosHWordP       10000 >  x >          0          ( 16 bits, unsigned)
% NegHWordP           0 >  x >=    -10000          ( 16 bits, unsigned)
% DispInumP        7FFF >= x >=     -8000          ( 16 bits )
% SmallInumP         7F >= x >=       -80          (  8 bits )
% QInumP              8 >= x >=        -8          (  several bits )
% PosQInumP           8 >= x >          0          (  3 bits, unsigned)
% NegQInumP           0 >  x >=        -8          (  3 bits, unsigned)

(off R2I)            % This function should exist in the kernel!           scs
(de IntP(N) (IntP N))
(on  R2I)

(de PosInumP (Expression)  (and (InumP    Expression)
                                (GreaterP Expression 0)))

(de NegInumP (Expression)  (and (InumP  Expression)
                (minusp Expression)))

(de Geq16P (Expression)    (and (InumP  Expression)
                                (Geq    Expression 16)))

(de LeqM16P (Expression)   (and (InumP  Expression)
                (Leq    Expression -16)))

(de Gt8P (Expression)      (and (InumP    Expression)
                                (GreaterP Expression 8)))

(de LtM8P (Expression)     (and (InumP    Expression)
                (LessP    Expression -8)))

(de PosHWordP (Expression) (and (IntP Expression)
                                          (GreaterP Expression 0)
                                          (lessP    Expression 16#10000)))

(de NegHWordP (Expression) (and (IntP Expression)
                                          (minusp Expression)
                                          (geq    Expression (minus 16#10000))))

(de DispInumP (Expression) (and (IntP Expression)
                                          (leq Expression 16#7fff)
                                          (geq Expression (minus 16#8000))))

(de SmallInumP (Expression)(and (IntP Expression)
                          (leq Expression 16#7f)
                          (geq Expression (minus 16#80))))

(de QInumP    (Expression) (and (IntP Expression)
                          (leq Expression  8)
                          (geq Expression -8)))

(de PosQInumP (Expression) (and (IntP Expression)
                                          (GreaterP Expression 0)
                                          (leq      Expression 8)))

(de NegQInumP (Expression) (and (IntP Expression)
                                          (minusp Expression)
                                          (geq    Expression -8)))


%---------------------------------------------------------
% The following set of predicates describes certain classes of
% register classes. RegP tests if the operand is a valid 68000 register.%
%
%  RegP  any 68000 register
% ARegP  68000 Address Register
% DRegP  68000 Data Register
% ScratchARegP tests for an unused  A register
% FakeRegP tests for argument register numbers greater than LastActualReg


(de RegP (RegName) (EqCar RegName 'Reg))

(de ScratchARegP (RegName) NIL)                                     % scs
% (de ScratchARegP (RegName) (AND (RegP RegName)
%                                 (Equal (cadr RegName) 'a0)))
%
(de ZeroRegP (RegName) NIL)                                         % scs
% (de ZeroRegP (RegName) (AND (RegP RegName)
%                             (memq (cadr RegName) '(a6 zeroareg))))

(de NotARegP (RegName) (not (ARegP RegName)))

(de FakeRegP (Expression) (and (PosIntP  Expression)
                   (GreaterP Expression LastActualReg&)))  % scs

% ------------------------------------------------------------
% Symbolic names for registers                                       scs
% ------------------------------------------------------------

(commentoutcode                                         % A register model
(de DRegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)   %  scs
   '(nil 7  8  9 10 11 12 13
     d0 d1 d2 d3 d4 d5 d6 d7
        t1 t2 t3 t4          ))))

(de ARegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)   %  scs
   '( 6  1  2  3  4  5  nil sp
     a0 a1 a2 a3 a4 a5  a6  a7
                     db  sb  st))))
(DefList '(                                                % A register model
           (NIL 0) (d0 0) (a0 0)
       (t1  1) (d1 1) (a1 1)
       (t2  2) (d2 2) (a2 2)
           (t3  3) (d3 3) (a3 3)
       (t4  4) (d4 4) (a4 4)
               (d5 5) (a5 5) (db 5)
               (d6 6) (a6 6) (sb 6) (zeroareg 6)
               (d7 7) (a7 7) (sp 7) (st 7)
     ) 'RegisterNumber)

(de RegisterNumber (RegSymbol)
% registers numbered according to A register model                        scs
  (cond
   ((NumberP RegSymbol)
              (IGetV [1 2 3 4 5 0 1 2 3 4 5 6 7] (ISUB1 RegSymbol)))
   ( T (or (get RegSymbol 'RegisterNumber)
              (StdError (BldMsg "Unknown register %r" RegSymbol))))
   ))

)


% Modified Aregp and Dregp so that it no longer used the 14 maxrealreg
%  model. Registers 7 -> 14 are now fakeregs.

(de DRegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)    %  scs
   '( 1  2  3  4  5  6
     d0 d1 d2 d3 d4 d5 d6 d7
                       t3 t4))))

(de ARegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)    %  scs
   '(                   nil sp
     a0 a1 a2 a3 a4 a5  a6  a7
                    db  sb  st
        t0 t1 t2 t5 t6        ))))


% These two functions are used as predicates in pass-2-and-3 of the compiler
%  to find temporary registers. They need to be seperate from Aregp and Dregp
%  now that we no longer use 14 real regs. See the functions &GetAtemp and
%  &GetDtemp in comp-decls for a further explanation. Eventually we can get
%  rid of this if we decide to use patterns that ask for generic temp regs
%  instead of specific ones. /LBS
(de TempDRegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)    %  scs
   '(1 2 3 4 5 6 7 t3 t4))))

(de TempARegP (RegName) (AND (RegP RegName) (MemQ (cadr RegName)    %  scs
   '(8 9 10 11 12 13 14 t0 t1 t2))))


(loadtime
(DefList '(
           (d0 0)         (a0 0)                      % D register model
       (d1 1)  (t0 1) (a1 1)
       (d2 2)  (t1 2) (a2 2)
       (d3 3)  (t2 3) (a3 3)
       (d4 4)  (t5 4) (a4 4)
       (d5 5)  (t6 5) (a5 5)         (db 5)
       (d6 6) (t3 6)  (a6 6) (NIL 6) (sb 6)
       (d7 7) (t4 7)  (a7 7) (sp  7) (st 7)
     )  'RegisterNumber)
)

(de RegisterNumber (RegSymbol)
% registers numbered according to D register model                        scs
  (cond ((NumberP RegSymbol)
     (remainder (ISUB1 RegSymbol) 8))
    ( T (OR (GET REGSYMBOL 'REGISTERNUMBER)
        (StdError (BldMsg "Unknown register %r"  RegSymbol))))
    ))


%-------------
%  ImmediateP tests if an item is tagged IMMEDIATE. (immediate x)
%  WConsts and WArrays are tagged immediate when they are not
%  inside MEMORY. The tagging means that the following expression
%  is to be used as an immediate value. For example, if WArray
%  SYMFNC is the base of some table, the expression
%       (*WPLUS2 (Reg 1) (immediate (WArray SYMFNC)))
%  means to add the address of SYMFNC to (Reg 1) and not the contents
%  of the SYMFNC location. Another immediate expression example would be
%  (*MOVE (postincrement (Reg st)) (immediate (plus2 (WArray ArgumentBlock) 32))
%  which means move the popped value of the stack to the address resulting from
%  the plus computation.
%-------------

(de ImmediateP (x)
  (and  (EqCar x 'Immediate) (Null (fixp (cadr x)))))

(de TaggedLabel (X) (EqCar X 'Label))


% For powers of two we could implement (de poweroftwoP (x) (zerop (land x
%                                                           (sub1 x))))

(de Minus1P     (x) (equal x  -1))
(de OneP        (x) (equal x   1))
(de TwoP        (x) (equal x   2))
(de FourP       (x) (equal x   4))
(de SixP        (x) (equal x   6))
(de EightP      (x) (equal x   8))
(de TenP        (x) (equal x  10))
(de SixteenP    (x) (equal x  16))
(de TwentyFourP (x) (equal x  24))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                       ANYREG DEFINITIONS
% These are the patterns for anyreg definitions.  They are of the form:
%
% (DefAnyreg ANYREG_NAME RESOLUTION_FUNCTION_NAME
%    ((Register_Predicate1) (Returned value))
%    ((Register_Predicate2) (returned value))
%    ...
% )

% The last element in the defintion has no predicate, it is a default
% case.  "Anyregs" are flagged (on their property list) with 'ANYREG.
% They have property indicators ANYREGPATTERNTABLE.  The predicate
% value list, and the function  to call - ANYREGRESOLUTIONFUNCTION.
%
% The associated function of most ANYREG's is called with 2
% arguments, a temporary REGISTER, used in the computation if needed,
% and the single SOURCE.
% The associated function usually uses (OneOperandAnyreg Register  Source
% 'Anyregname) or (ExpandOneArgumentAnyReg Register Source
% 'AnyregName). The 'AnyregName is used to find the associated pattern.
% The MEMORY anyreg has an additional argument, called ARGTWO. The
% (OneOperandAnyreg Register Source Anyregname)%
% calls the lower level ExpandOneArgumentAnyreg, after replacing Source
% by (ResolveOperand Register Source).  This function processes Source
% to see if it is. The MEMORY anyreg means that the location is the
% address of the source or destination.
%It is important that the anyregs DO NOT assign any explicit
% registers because an instruction may have two similar anyregs which
% cause the explicit register to be used twice in the same instruction.
% The future may require a scheme to allocate temporary A and D regs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%% scs %%%%%%%%%%% scs %%%%%%%%%%%%%%%%% scs %%%%%%%%%%

(DefAnyreg CAR
           AnyregCAR   %Grab the source so caller can displace off it.
           (       (!*Field REGISTER SOURCE InfStartingBit InfBitLength)
                   (indirect REGISTER))
)

(DefAnyreg CDR
           AnyregCDR     %Same as CAR, except move to next word in pair.
           (       (!*Field REGISTER SOURCE InfStartingBit InfBitLength)
                   (Displacement REGISTER 4))
)


%% The following two have been taken out because they do not do tag stripping.
%%  Okay for the 68000 and 68010, but not the 68020.  LBS
%
%(DefAnyreg CAR
%  AnyregCAR   %Grab the source so caller can displace off it.
%           ((ARegP)  (indirect SOURCE))
%           (         (*MOVE SOURCE REGISTER) (indirect REGISTER )))
%
%(DefAnyreg CDR
%  AnyregCDR     %Same as CAR, except move to next word in pair.
%           ((ARegP)  (Displacement SOURCE 4))
%           (         (*MOVE SOURCE REGISTER) (Displacement REGISTER 4)))

%(DefAnyreg MEMORY
%           AnyregMEMORY
%           ((ARegP ZeroP)      (Indirect SOURCE))
%           ((ARegP DispInumP)  (Displacement SOURCE ARGTWO))
%           ((ARegP RegP)       (Indexed ARGTWO (Displacement SOURCE 0)))
%           ((AregP AnyP)       (*MOVE  ARGTWO REGISTER)
%                           (Indexed REGISTER (Displacement SOURCE 0)))
%           ((AnyP  ARegP)      (MEMORY ARGTWO SOURCE))
%           ((InumP InumP)      (IMMEDIATE (PLUS2 SOURCE ARGTWO)))
%           ((InumP  AnyP)      (MEMORY ARGTWO SOURCE))
%           ((AnyP ZeroP)       (*MOVE SOURCE REGISTER)
%                               (Indirect REGISTER))
%           ((AnyP DispInumP)   (*MOVE SOURCE REGISTER)
%                           (Displacement REGISTER ARGTWO))
%           ((AnyP RegP)        (*MOVE SOURCE REGISTER)
%                               (Indexed REGISTER (Displacement ARGTWO 0)))
%           ((RegP AnyP)        (*MOVE ARGTWO REGISTER)
%                               (Indexed REGISTER (Displacement SOURCE 0)))
%           (                   (*MOVE SOURCE REGISTER)
%                               (*WPLUS2 REGISTER ARGTWO)
%                               (Indirect REGISTER)))


% This new version is based on the old Sun 3.2 cmacro. It is designed to act
%  without the help of the memory pattern under most circumstances. Previously
%  the pattern, was doing most of work using the 14 real reg model.
%
(DefAnyreg MEMORY
           AnyregMEMORY
           ((ARegP ZeroP)      (Indirect SOURCE))
           ((Anyp  ZeroP)      (*MOVE SOURCE REGISTER)
                           (Indirect REGISTER))
           ((ARegP DispInumP)  (Displacement SOURCE ARGTWO))
           ((AnyP  DispInumP)  (*MOVE SOURCE REGISTER)
                           (Displacement REGISTER ARGTWO))
           ((RegP InumP)       (!*MOVE SOURCE REGISTER)
                           (!*WPLUS2 REGISTER ARGTWO)
                   (Indirect REGISTER))
           ((ARegP RegP)       (Indexed ARGTWO (Displacement SOURCE 0)))
           ((DRegP AnyP)       (!*MOVE ARGTWO REGISTER)
                               (Indexed SOURCE (Displacement REGISTER 0)))
           ((RegP  AnyP)       (!*MOVE SOURCE REGISTER)
                           (!*WPLUS2 REGISTER ARGTWO)
                               (Indirect REGISTER))
           ((AnyP DispInumP)   (!*MOVE SOURCE REGISTER)
                               (Indexed REGISTER (Displacement ARGTWO 0)))
           (                   (!*MOVE SOURCE REGISTER)
                               (!*WPLUS2 REGISTER ARGTWO)
                               (Indirect REGISTER)))


(DefAnyreg FRAME
  AnyregFRAME
           ((zerop)  (Indirect (reg st)))
           (         (Displacement (REG st) SOURCE)))

% The compiler will tag expressions immediate in the procedure ResolveWConst.
% Only expressions are tagged immediate, not numbers.


(De AnyRegImmediate(REGISTER SOURCE)
   (cond ((InumP  SOURCE)               SOURCE)
         ((Eqcar SOURCE 'Unimmediate)   SOURCE)
         (T                            (list 'immediate SOURCE) )
     ))

(DefAnyReg IMMEDIATE
           AnyRegImmediate)


(Defanyreg QUOTE
           AnyregQUOTE
           ((Null) (REG NIL))  %If NIL, return the NIL register (D6?)
           ((EqTP) (FLUID T))  %??? Is this defined?
           ((IntP)  SOURCE)    %?
           (       (QUOTE SOURCE)))   %? Recursivly expand??


% Took out the displacement from the Zero reg from the fakereg case. The
%  D reg model uses a6 as the nil reg, not the zero reg. /LBS
(DefAnyreg REG
           AnyregREG
           ((FakeRegP)          (*MOVE (fluid argumentblock) REGISTER)
                  (displacement REGISTER
                           (times addressingunitsperitem SOURCE)))
           (                  (REG SOURCE)))

% don't need this anymore. sal.
%(DefAnyreg WVAR
%           AnyregWVAR
%           ((RegisterNameP) (REG SOURCE))  %/
%           (                (WVAR SOURCE)))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                        CMACRO DEFINITIONS
% Remember that it IS legal to use Other CMACROS in expansions%
% or to define "fake" cmacros, ie procedures, for%
% Common cases. If a new C-macro is not standard,%
% e.g (*WFOO argone argtwo), then define its table and procedure.%
% You may have to define some more "anyregs" or 'TerminalOperands%
% Or define predicates to be CAREFUL.%
% (def Cmacro *WFOO%
%      ...... )%
% (de *WFOO (Arg1 Arg2)%
%   (Expand2OperandCmacro Arg1 Arg2 '*Wfoo))%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro *Loc                           % This needs more study    scs
          ((AregP ImmediateP)(lea (unimmediate ARGTWO) ARGONE))
          ((ARegP AnyP)      (lea              ARGTWO  ARGONE))
          ((AnyP  ImmediateP)(*MOVE           ARGTWO  ARGONE))
          (                  (lea              ARGTWO (Reg T2))
                             (*MOVE          (Reg T2) ARGONE)))



% Changed to use regs t0 and t1. Originally used regs t1 and t2, which caused
%  a register conflict in fastapply since it uses reg t2 to store the arg
%  list. The conflict occured when the number of args > maxnargs, since the
%  move from the argumentblock needed reg t2, which destroyed the argument
%  list.
(de *Move (Source Destination)            % redefined from COMMON-CMACROS
  (prog (resultingcode*)
    (return
     (CMacroPatternExpand
      (list (ResolveOperand '(REG t5) Source)
        (ResolveOperand '(REG t6) Destination))
      (get '*Move 'CMacroPatternTable)))))

(DefCMacro *Move                          %  (*Move Source Destination)
   ( Equal              )                  % if source=dest then do nothing
   ((ZeroP      ARegP) (suba!.l  ARGTWO ARGTWO))
   ((DispInumP  AregP) (movea!.w ARGONE ARGTWO))
   ((AnyP       AregP) (movea!.l ARGONE ARGTWO))
   ((SmallInumP DregP) (moveq    ARGONE ARGTWO))
   ((ZeroP      AnyP)  (clr!.l          ARGTWO))
   ((SmallInumP AnyP)  (moveq    ARGONE (Reg T4))
                       (move!.l (Reg T4) ARGTWO))
   (                   (move!.l  ARGONE ARGTWO))
   )

(de *MOVEUNSBYTE (ARGONE ARGTWO)
 (Expand2OperandCMacro ARGONE ARGTWO '*MOVEUNSBYTE))

(DefCMacro *MOVEUNSBYTE              %  MOVE BYTE  ARGONE --> ARGTWO     scs
 ((INUMP  REGP)      (*MOVE (LAND ARGONE 16#FF) ARGTWO))
 ((ANYP   AREGP)     (*MOVEUNSBYTE ARGONE (Reg T4))
                     (*MOVE  (Reg T4) ARGTWO))
 ((AREGP  ANYP)      (*MOVE  ARGONE (Reg T3))
                     (*MOVEUNSBYTE (Reg T3) ARGTWO))
 ((ANYP   DREGP)     ( clr!.l ARGTWO)
                     (move!.b ARGONE ARGTWO))
 (                   ( move!.b ARGONE ARGTWO))
 )

(de *MOVESGNBYTE  (ARGONE ARGTWO)
 (Expand2OperandCMacro ARGONE ARGTWO '*MOVESGNBYTE))

(DefCMacro *MOVESGNBYTE                %  MOVE BYTE  ARGONE --> ARGTWO     scs
 ((INUMP  REGP)      (*MOVE (SGNFIELD ARGONE 27 5) ARGTWO))
 ((DREGP  AREGP)     ( ext!.w ARGONE)
                     ( movea!.w ARGONE ARGTWO))
 ((AREGP  AREGP)     (*MOVE ARGONE   (Reg T4))
                     (*MOVESGNBYTE (Reg T4) ARGTWO))
 ((ANYP   AREGP)     ( move!.b ARGONE (Reg T4))
                     (*MOVESGNBYTE (Reg T4) ARGTWO))
 ((AREGP  DREGP)     (*MOVE   ARGONE ARGTWO)
                     ( ext!.w ARGTWO)
                     ( ext!.l ARGTWO))
 ((AREGP  ANYP)      (*MOVE   ARGONE (Reg T4))
                     (*MOVESGNBYTE (Reg T4) ARGTWO))
 ((ANYP   DREGP)     ( move!.b ARGONE ARGTWO)
                     ( ext!.w ARGTWO)
                     ( ext!.l ARGTWO))
 (                   ( move!.b ARGONE ARGTWO))
 )

(de *MOVESGNHWORD  (ARGONE ARGTWO)
 (Expand2OperandCMacro ARGONE ARGTWO '*MOVESGNHWORD))

(DefCMacro *MOVESGNHWORD        %  MOVE HALFWORD  ARGONE --> ARGTWO     scs
 ((INUMP  REGP)      (*MOVE (SGNFIELD ARGONE 16 16) ARGTWO))
 ((ANYP   AREGP)     ( movea!.w ARGONE ARGTWO))
 ((ANYP   DREGP)     ( move!.w  ARGONE ARGTWO)
                     ( ext!.l ARGTWO))
 (                   ( move!.w  ARGONE ARGTWO))
 )

(DefCMacro *Pop
 (                   (*MOVE (postincrement (Reg st)) ARGONE)))

(DefCMacro *Push     % check trap here
 (                   (*MOVE ARGONE (PreDecrement (reg st)))))

(DefCMacro *WPlus2                %  (*WPlus2 dest source)      %  scs
 ((AnyP  ZeroP)         )
 ((AnyP  PosQInumP)   (addq!.l ARGTWO ARGONE))
 ((AnyP  NegQInumP)   (subq!.l (minus ARGTWO) ARGONE))
 ((AregP DispInumP)   (adda!.w ARGTWO ARGONE))
 ((AregP AnyP)        (adda!.l ARGTWO ARGONE))
 ((AnyP  SmallInumP)  (moveq ARGTWO (Reg T4))
                      (add!.l (Reg T4) ARGONE))
 ((DRegP AnyP)        (add!.l  ARGTWO ARGONE))
 ((AnyP  DRegP)       (add!.l  ARGTWO ARGONE))
 ((AnyP  InumP)       (addi!.l ARGTWO ARGONE))
 (                    (*MOVE  ARGTWO (Reg t4))
                      (add!.l (Reg t4) ARGONE))
)

(DefCMacro *WDifference                                        %  scs
           ((AnyP  ZeroP)       )
           (equal             (*MOVE 0 ARGONE))
           ((AnyP  PosQInumP) (subq!.l ARGTWO ARGONE))
           ((AnyP  NegQInumP) (addq!.l (minus ARGTWO) ARGONE))
           ((AregP DispInumP) (suba!.w ARGTWO ARGONE))
           ((AregP AnyP)      (suba!.l ARGTWO ARGONE))
           ((AnyP  SmallInumP)(moveq ARGTWO (Reg T4))
                              (sub!.l (Reg T4) ARGONE))
           ((DregP AnyP )     (sub!.l  ARGTWO ARGONE))
           ((AnyP  DregP )    (sub!.l  ARGTWO ARGONE))
           ((AnyP  InumP)     (subi!.l ARGTWO ARGONE))
           (                  (*MOVE  ARGTWO (Reg t4))
                              (sub!.l (Reg t4) ARGONE))
)

(put 'wtimes2 'opencode '((muls!.l (reg 2) (reg 1))))

(put 'wquotient 'opencode '((divsl!.l (reg 2) (reg 3) (reg 1))))

(put 'wremainder 'opencode '((divsl!.l (reg 2) (reg 3) (reg 1))
                             (*move (reg 3) (reg 1))))

(put 'wdivide 'opencode '((divsl!.l (reg 2) (reg 3) (reg 1))
                          (*move (reg 3) ($fluid *second-value*))))

(de *WNegate(ARG1)
 (Expand1OperandCMacro ARG1 '*WNegate))

(DefCMacro *WNegate
           ( AregP            (*WMinus ARGONE (Reg T4))
                              (*MOVE   (Reg T4) ARGONE))
           (                  (neg!.l ARGONE))
 )

(DefCMacro *WMinus                                               %  scs
           ((AnyP  InumP)     (*MOVE (MINUS ARGTWO) ARGONE))
           ( Equal            (*WNegate ARGONE))
           ((DregP AnyP)      (*MOVE ARGTWO ARGONE)
                              (neg!.l ARGONE))
           ((AregP AnyP)      (*MOVE 0 ARGONE)
                              (suba!.l ARGTWO ARGONE))
           (                  (*WMinus ARGTWO (Reg T4))
                              (*MOVE (reg t4) ARGONE))
)

(de *WComplement(ARG1)
 (Expand1OperandCMacro ARG1 '*WComplement))

(DefCMacro *WComplement
           ( AregP            (*WNot ARGONE (Reg T4))
                              (*MOVE (Reg T4) ARGONE))
           (                  (not!.l ARGONE))
 )

(DefCMacro *WNot
           ((AnyP  InumP)     (*MOVE (LNOT ARGTWO) ARGONE))
           ( Equal            (*WComplement ARGONE))
           ((DregP AnyP)      (*MOVE ARGTWO ARGONE)
                              (not!.l ARGONE))
           (                  (*WNot ARGTWO (Reg T4))
                              (*MOVE (reg t4) ARGONE))
 )

(DefCMacro *WAnd                                               %    scs
           ( equal                )
           ((AnyP  Minus1P)       )
           ((AnyP  ZeroP)       (*MOVE  0 ARGONE))
           ((AnyP  AregP)       (*MOVE  ARGTWO (Reg T4))
                                (*WAND  ARGONE (Reg T4)))
           ((AregP AnyP)        (*MOVE  ARGONE (Reg T3))
                                (*WAND  (Reg T3) ARGTWO)
                                (*MOVE  (Reg T3) ARGONE))
           ((AnyP  NegHWordP)   (and!.w  ARGTWO ARGONE))
           ((AnyP  SmallInumP)  (moveq   ARGTWO (Reg T4))
                                (and!.l  (Reg T4) ARGONE))
           ((DRegP AnyP)        (and!.l  ARGTWO ARGONE))
           ((AnyP  DRegP)       (and!.l  ARGTWO ARGONE))
           ((AnyP  InumP)       (andi!.l ARGTWO ARGONE))
           (                    (*MOVE  ARGTWO (Reg t4))
                                (and!.l  (Reg t4) ARGONE))
)

(DefCMacro *WOr                                               %  scs
           ( equal                )
           ((AnyP  ZeroP)         )
           ((AnyP  Minus1P)     (*MOVE -1 ARGONE))
           ((AnyP  AregP)       (*MOVE ARGTWO (Reg T4))
                                (*WOR ARGONE (Reg T4)))
           ((AregP AnyP)        (*MOVE ARGONE (Reg T3))
                                (*WOR (Reg T3) ARGTWO)
                                (*MOVE (Reg T3) ARGONE))
           ((AnyP  PosHWordP)   (or!.w ARGTWO ARGONE))
           ((AnyP  SmallInumP)  (moveq ARGTWO (Reg T4))
                                (or!.l (Reg T4) ARGONE))
           ((DRegP AnyP)        (or!.l  ARGTWO ARGONE))
           ((AnyP  DRegP)       (or!.l  ARGTWO ARGONE))
           ((AnyP  InumP)       (ori!.l ARGTWO ARGONE))
           (                    (*MOVE  ARGTWO (Reg t4))
                                (or!.l (Reg t4) ARGONE))
)

(DefCMacro *WXOr                                                %  scs
           ((AnyP  ZeroP)         )
           ( equal              (*MOVE  0 ARGONE))
           ((AnyP  Minus1P)     (*WNOT    ARGONE))
           ((AnyP  AregP)       (*MOVE ARGTWO (Reg T4))
                                (*WXOR ARGONE (Reg T4)))
           ((AregP AnyP)        (*MOVE ARGONE (Reg T3))
                                (*WXOR (Reg T3) ARGTWO)
                                (*MOVE (Reg T3) ARGONE))
           ((AnyP  PosHWordP)   (eori!.w ARGTWO ARGONE))
           ((AnyP  SmallInumP)  (moveq ARGTWO (Reg T4))
                                (eor!.l (Reg T4) ARGONE))
           ((AnyP  InumP)       (eori!.l ARGTWO ARGONE))
           ((AnyP  DRegP)       (eor!.l  ARGTWO ARGONE))
           (                    (*MOVE  ARGTWO (Reg t4))
                                (eor!.l (Reg t4) ARGONE))
)


% ----------
% Ashift
% +index shifts left and shifts in zeroes.
% -index shifts right and sign extends.
%  ARGONE <- ARGONE shifted by ARGTWO
%------------------------------------------

(DefCMacro *AShift                                            %    scs
           ((AnyP  ZeroP)           )
           ((AnyP  PosInumP)    (*WShift ARGONE ARGTWO))
           ((DRegP NegQInumP)   (asr!.l (minus ARGTWO) ARGONE))
           ((DRegP LeqM16P)     (swap ARGONE)
                                (ext!.l ARGONE)
                                (*ASHIFT ARGONE (plus2 ARGTWO 16)))
           ((DRegP LtM8P)       (asr!.l 8 ARGONE)
                                (*ASHIFT ARGONE (plus2 ARGTWO  8)))
           ((DRegP DregP)       (tst!.l  ARGTWO)
                                (bge!.s TEMPLABEL)
                                (neg!.l ARGTWO)
                                (asr!.l ARGTWO ARGONE)
                                (bra!.s TEMPLABEL2)
                         (*LBL (label TEMPLABEL))
                                (asl!.l ARGTWO ARGONE)
                         (*LBL (label TEMPLABEL2)))
           ((DRegP AnyP)        (*MOVE ARGTWO (Reg T4))
                                (bge!.s TEMPLABEL)
                                (neg!.l (Reg T4))
                                (asr!.l (Reg T4) ARGONE)
                                (bra!.s TEMPLABEL2)
                         (*LBL (label TEMPLABEL))
                                (asl!.l (Reg T4) ARGONE)
                         (*LBL (label TEMPLABEL2)))
           (                    (*MOVE ARGONE (Reg t3))
                                (*ASHIFT (Reg t3) ARGTWO)
                                (*MOVE (Reg t3) ARGONE))
)


(DefCMacro *WShift                     %Logical shift. +index=left.
           ((AnyP  ZeroP)           )
           ((RegP  OneP)        (*WPLUS2 ARGONE ARGONE))
           ((ARegP TwoP)        (*WPLUS2 ARGONE ARGONE)
                                (*WPLUS2 ARGONE ARGONE))
           ((DRegP PosQInumP)   (lsl!.l ARGTWO ARGONE))
           ((DRegP Geq16P)      (*WSHIFT ARGONE (difference ARGTWO 16))
                                (swap ARGONE)
                                (clr!.w ARGONE))
           ((DRegP Gt8P)        (lsl!.l 8 ARGONE)
                                (*WSHIFT ARGONE (difference ARGTWO  8)))
           ((DRegP NegQInumP)   (lsr!.l (minus ARGTWO) ARGONE))
           ((DRegP LeqM16P)     (clr!.w ARGONE)
                                (swap ARGONE)
                                (*WSHIFT ARGONE (plus2 ARGTWO 16)))
           ((DRegP LtM8P)       (lsr!.l 8 ARGONE)
                                (*WSHIFT ARGONE (plus2 ARGTWO  8)))
           ((DRegP DregP)       (tst!.l  ARGTWO)
                                (bge!.s TEMPLABEL)
                                (neg!.l ARGTWO)
                                (lsr!.l ARGTWO ARGONE)
                                (bra!.s TEMPLABEL2)
                         (*LBL (label TEMPLABEL))
                                (lsl!.l ARGTWO ARGONE)
                         (*LBL (label TEMPLABEL2)))
           ((DRegP AnyP)        (*MOVE ARGTWO (Reg T4))
                                (bge!.s TEMPLABEL)
                                (neg!.l (Reg T4))
                                (lsr!.l (Reg T4) ARGONE)
                                (bra!.s TEMPLABEL2)
                         (*LBL (label TEMPLABEL))
                                (lsl!.l (Reg T4) ARGONE)
                         (*LBL (label TEMPLABEL2)))
           (                    (*MOVE ARGONE (Reg t3))
                                (*WSHIFT (Reg t3) ARGTWO)
                                (*MOVE (Reg t3) ARGONE))
)


% *JumpIfTag is an optimized form of *jumpif.It knows that we are doing word
% compares only.

(de *JumpIfTag (arg1 arg2 label instructions)
 (prog (resultingcode*)
       (return
        (cmacropatternexpand
         (list (resolveoperand '(reg t1) arg1)
               (resolveoperand '(reg t2) arg2)
               (resolveoperand '(reg error) label)
               (car instructions)
               (cdr instructions))
         (get '*JumpIfTag 'cmacropatterntable)))))



%  We could probably make the assumption here that we are comparing
%       a D register (16 bits only) to a constant of the form  TAGxxxxxxxxxx.

(DefCMacro *JumpIfTag                  %( JumpIfTag a b lbl jmp rev-jmp)
  ((AregP ZeroP    ) (move!.w ARGONE (Reg T4))(ARGFOUR ARGTHREE))
  ((AregP AnyP     ) (cmpa!.w ARGTWO ARGONE)  (ARGFOUR ARGTHREE))
  ((AnyP  AregP    ) (*JumpIfTag ARGTWO ARGONE ARGTHREE (ARGFIVE . ARGFOUR)))
%?((INumP INumP    )   should be caught by front end constant folding
  ((INumP AnyP     ) (*JumpIfTag ARGTWO ARGONE ARGTHREE (ARGFIVE . ARGFOUR)))
  ((AnyP  ZeroP    ) (tst!.w ARGONE)          (ARGFOUR ARGTHREE))
  ((DregP AnyP     ) (cmp!.w ARGTWO ARGONE)   (ARGFOUR ARGTHREE))
  ((AnyP  DregP    ) (cmp!.w ARGONE ARGTWO)   (ARGFIVE ARGTHREE))
  ((AnyP  InumP    ) (cmpi!.w ARGTWO ARGONE)  (ARGFOUR ARGTHREE))
  (                  (move!.w ARGONE (reg t4))
                     (cmp!.w  ARGTWO (reg t4)) (ARGFOUR ARGTHREE)))


(De *JumpEQTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Beq . Beq)))
(DefCmacro *JumpEqTag)

(De *JumpNotEQTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Bne . Bne)))
(DefCmacro *JumpNotEQTag)

(De *JumpWGEQTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Bge . Ble)))
(DefCmacro *JumpWGEQTag)

(De *JumpWGreaterPTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Bgt . Blt)))
(DefCmacro *JumpWGreaterPTag)

(De *JumpWLessPTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Blt . Bgt)))
(DefCmacro *JumpWLesspTag)


% The following always produce short jumps:

(De *SJumpEQTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Beq!.s . Beq!.s)))
(DefCmacro *SJumpEqTag)

(De *SJumpNotEQTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Bne!.s . Bne!.s)))
(DefCmacro *SJumpNotEQTag)

(De *SJumpWGEQTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Bge!.s . Ble!.s)))
(DefCmacro *SJumpWGEQTag)

(De *SJumpWGreaterPTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Bgt!.s . Blt!.s)))
(DefCmacro *SJumpWGreaterPTag)

(De *SJumpWLessPTag (Lbl Arg1 Arg2)
       (*JumpIfTag Arg1 Arg2 Lbl '(Blt!.s . Bgt!.s)))
(DefCmacro *SJumpWLesspTag)



(DefCMacro *JumpType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE)
                          (*JumpEQTag ARGTHREE ARGONE (lshift ARGTWO 11)))
     (                    (*MOVE ARGONE (Reg T3))
                          (*JumpType (Reg T3) ARGTWO ARGTHREE))
     )


(DefCMacro *JumpNotType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
                          (*JumpNotEQTag ARGTHREE ARGONE (lshift ARGTWO 11)))
     (                    (*MOVE ARGONE (Reg T3))
                          (*JumpNotType (Reg T3) ARGTWO ARGTHREE))
     )

(DefCMacro *JumpInType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
			  (*SJumpWGreaterPTag TEMPLABEL ARGONE (lshift ARGTWO 11))
                          (*JumpWGeqTag       ARGTHREE  ARGONE  16#F800)
                          (*LBL (label TEMPLABEL)))
     (                    (*MOVE ARGONE (Reg T3))
                          (*JumpInType (Reg T3) ARGTWO ARGTHREE))
     )

(DefCMacro *JumpNotInType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
                          (*JumpWGreaterPTag ARGTHREE ARGONE (lshift ARGTWO 11))
                          (*JumpWLessPTag    ARGTHREE ARGONE  16#F800))
     (                    (*MOVE ARGONE (Reg T3))
                          (*JumpInType (Reg T3) ARGTWO ARGTHREE))
     )


% The following always produce short jumps:

(de *SJumpType (Label Arg type-tag)
  (Expand2OperandAndLabelCMacro Arg
                    (compiler-constant type-tag)
                Label
                '*SJumpType))
(DefCMacro *SJumpType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
                          (*SJumpEQTag ARGTHREE ARGONE (lshift ARGTWO 11)))
     (                    (*MOVE ARGONE (Reg T3))
                          (*SJumpType (Reg T3) ARGTWO ARGTHREE))
     )


(de *SJumpNotType (Label Arg type-tag)
  (Expand2OperandAndLabelCMacro Arg
                    (compiler-constant type-tag)
                Label
                '*SJumpNotType))
(DefCMacro *SJumpNotType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
                          (*SJumpNotEQTag ARGTHREE ARGONE (lshift ARGTWO 11)))
     (                    (*MOVE ARGONE (Reg T3))
                          (*SJumpNotType (Reg T3) ARGTWO ARGTHREE))
     )

(de *SJumpInType (Label Arg type-tag)
  (Expand2OperandAndLabelCMacro Arg
                    (compiler-constant type-tag)
                Label
                '*SJumpInType))
(DefCMacro *SJumpInType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
			  (*SJumpWGreaterPTag TEMPLABEL ARGONE (lshift ARGTWO 11))
                          (*SJumpWGeqTag       ARGTHREE  ARGONE  16#F800)
                          (*LBL (label TEMPLABEL)))
     (                    (*MOVE ARGONE (Reg T3))
                          (*SJumpInType (Reg T3) ARGTWO ARGTHREE))
     )

(de *SJumpNotInType (Label Arg type-tag)
  (Expand2OperandAndLabelCMacro Arg
                    (compiler-constant type-tag)
                Label
                '*SJumpNotInType))
(DefCMacro *SJumpNotInType
     ((DregP AnyP AnyP)   (swap   ARGONE)
                          (andi!.w 16#f800 ARGONE) %(clr!.b ARGONE)
                          (*SJumpWGreaterPTag ARGTHREE ARGONE (lshift ARGTWO 11))
                          (*SJumpWLessPTag    ARGTHREE ARGONE  16#F800))
     (                    (*MOVE ARGONE (Reg T3))
                          (*SJumpInType (Reg T3) ARGTWO ARGTHREE))
     )







(DE *JUMPIF (ARG1 ARG2 LABEL INSTRUCTIONS)
  (PROG (RESULTINGCODE*)
    (RETURN
     (CMACROPATTERNEXPAND
      (LIST (RESOLVEOPERAND '(REG T1) ARG1)
        (RESOLVEOPERAND '(REG T2) ARG2)
        (RESOLVEOPERAND '(REG ERROR) LABEL)
               (CAR INSTRUCTIONS)
               (CDR INSTRUCTIONS))
         (GET '*JUMPIF 'CMACROPATTERNTABLE)))))

(DefCMacro *JumpIf                  %( JumpIF a b lbl jmp rev-jmp)
    ((AregP ZeroP    ) (move!.l ARGONE (Reg T4))(ARGFOUR ARGTHREE))
    ((AregP DispInumP) (cmpa!.w ARGTWO ARGONE)  (ARGFOUR ARGTHREE))
    ((AregP AnyP     ) (cmpa!.l ARGTWO ARGONE)  (ARGFOUR ARGTHREE))
    ((AnyP  AregP    ) (*JumpIf ARGTWO ARGONE ARGTHREE (ARGFIVE . ARGFOUR)))
    ((INumP INumP    ) (!*MOVE ARGONE (reg t4))
                       (cmp!.l ARGTWO (reg t4)) (ARGFOUR ARGTHREE))
    ((INumP AnyP     ) (*JumpIf ARGTWO ARGONE ARGTHREE (ARGFIVE . ARGFOUR)))
    ((AnyP  ZeroP    ) (tst!.l ARGONE)          (ARGFOUR ARGTHREE))
    ((AnyP  SmallInumP)(moveq ARGTWO  (Reg T4))
                   (cmp!.l ARGONE (Reg T4) )(ARGFIVE ARGTHREE))
    ((DregP AnyP     ) (cmp!.l ARGTWO ARGONE)   (ARGFOUR ARGTHREE))
    ((AnyP  DregP    ) (cmp!.l ARGONE ARGTWO)   (ARGFIVE ARGTHREE))
    ((AnyP  InumP    ) (cmpi!.l ARGTWO ARGONE)  (ARGFOUR ARGTHREE))
    (                  (*MOVE ARGONE (reg t4))
                       (cmp!.l ARGTWO (reg t4)) (ARGFOUR ARGTHREE)))

(De *JumpEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Beq . Beq)))
(DefCmacro *JumpEq)

(De *JumpNotEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Bne . Bne)))
(DefCmacro *JumpNotEQ)

(De *JumpWGEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Bge . Ble)))
(DefCmacro *JumpWGEQ)

(De *JumpWGreaterP (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Bgt . Blt)))
(DefCmacro *JumpWGreaterP)

(De *JumpWLEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Ble . Bge)))
(DefCmacro *JumpWLEQ)

(De *JumpWLessP (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Blt . Bgt)))
(DefCmacro *JumpWLessp)



%  The following always produce short jumps:

(De *SJumpEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Beq!.s . Beq!.s)))
(DefCmacro *SJumpEq)

(De *SJumpNotEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Bne!.s . Bne!.s)))
(DefCmacro *SJumpNotEQ)

(De *SJumpWGEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Bge!.s . Ble!.s)))
(DefCmacro *SJumpWGEQ)

(De *SJumpWGreaterP (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Bgt!.s . Blt!.s)))
(DefCmacro *SJumpWGreaterP)

(De *SJumpWLEQ (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Ble!.s . Bge!.s)))
(DefCmacro *SJumpWLEQ)

(De *SJumpWLessP (Lbl Arg1 Arg2)
       (*JumpIf Arg1 Arg2 Lbl '(Blt!.s . Bgt!.s)))                   %  scs
(DefCmacro *SJumpWLessp)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MkItem may be used when evaluating WConst expressions.%
% An item is formed on the apollo with the first eight bits as the Tag, %
%   ----------------------------%
%   | Tag |    Info field      |%
%   ----------------------------%
%  31     24                   0%
%%
% To create a constant ITEM, the TagPart must first be shifted 24 bits %
% to place it in the upper 8, then the InfoPart has its upper 8 bits%
% masked off.  The two are then ORed together.%
% Not really correct as the inf could end up overlapping the tag%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(DefCMacro *MkItem
     (              (*PUTFIELD ARGTWO ARGONE 0 5))
 )


% Note that the arguments to *PutField are source, destination, start, size.
% This is not what the CMACRO document indicates. MkItem above has been
% modified to reflect this ordering.  This is entirely backwards to the
% majority of the CMACRO's....

(DefCMacro *PutField
  ((InumP DregP ZeroP AnyP)
                     ( and!.l  (LNOT (BITMASK  ARGTHREE ARGFOUR)) ARGTWO)
                     ( or!.l   (LAND (BITMASK  ARGTHREE ARGFOUR)
                             (LSHIFT ARGONE (SHIFTAMT ARGTHREE ARGFOUR)))
                   ARGTWO))
  ((InumP DregP AnyP  AnyP)
                     (*WAND   ARGTWO (LNOT (BITMASK  ARGTHREE ARGFOUR)))
                     (*WOR    ARGTWO (LAND (BITMASK  ARGTHREE ARGFOUR)
                             (LSHIFT ARGONE (SHIFTAMT ARGTHREE ARGFOUR)))))

  ((INumP ARegP AnyP  AnyP)   (*MOVE   ARGTWO (Reg T4))
                              (*PUTFIELD ARGONE (Reg T4) ARGTHREE ARGFOUR)
                              (*MOVE (Reg T4) ARGTWO))
  ((DRegP RegP  ZeroP EightP) (*PUSH   ARGTWO)
                              (move!.b  ARGONE (indirect (reg st)))
                              (*POP    ARGTWO))
  ((InumP AnyP  ZeroP EightP) (move!.b  ARGTWO ARGONE))
  ((DregP AnyP  ZeroP EightP) (move!.b  ARGTWO ARGONE))

  ((DregP DregP ZeroP AnyP)
                     (*WSHIFT ARGONE       (SHIFTAMT ARGTHREE ARGFOUR))
                     ( and!.l  (LNOT (BITMASK  ARGTHREE ARGFOUR)) ARGTWO)
                     ( or!.l   ARGONE ARGTWO))
  ((DregP DregP AnyP AnyP)
                     (*WSHIFT ARGONE       (SHIFTAMT ARGTHREE ARGFOUR))
                     ( and!.l  (BITMASK  ARGTHREE ARGFOUR) ARGONE)
                     (*WAND   ARGTWO (LNOT (BITMASK  ARGTHREE ARGFOUR)))
                     (*WOR    ARGTWO ARGONE))
  ((AnyP  DregP AnyP AnyP)
                     (*MOVE ARGONE (reg t3))
             (*PUTFIELD    (Reg T3) ARGTWO ARGTHREE ARGFOUR))
  (                  (*MOVE ARGTWO     (reg t4))
                     (*PUTFIELD ARGONE (Reg T4) ARGTHREE ARGFOUR)
                     (*MOVE            (reg t4) ARGTWO))
 )

(DefCMacro *SignedField
  ((dregp anyp anyp anyp)(*MOVE   ARGTWO ARGONE)
                         (*ASHIFT ARGONE ARGTHREE)
                         (*ASHIFT ARGONE (DIFFERENCE ARGFOUR 32)))
  (                      (*SignedField (reg t3) ARGTWO ARGTHREE ARGFOUR)
                         (*Move        (reg t3) ARGONE))
)

% *Field and *SignedField could be improved by using ROL/ROR.

(DefCMacro *Field
  ((dregp anyp zerop anyp) (*MOVE   ARGTWO ARGONE)
                           (*WSHIFT ARGONE (DIFFERENCE ARGFOUR 32)))
  ((dregp anyp eightp twentyfourp)
                           (*move ARGTWO ARGONE)
                           (*wshift ARGONE 5)
                           (*wshift ARGONE -5))
  ((dregp anyp anyp anyp)(*MOVE   ARGTWO ARGONE)
                         ( and!.l  (bitmask ARGTHREE ARGFOUR) ARGONE)
                         (*WSHIFT ARGONE (minus (shiftamt  ARGTHREE ARGFOUR))))
  (                      (*Field (reg t3) ARGTWO ARGTHREE ARGFOUR)
             (*Move  (reg t3) ARGONE))
)

% ----------
% Alloc
% Allocates stack space upon procedure entry.
% ----------

%(DefCMacro !*Alloc
%  ((!*WDIFFERENCE (REG st) ARGONE)
%  (tst!.b (displacement (REG st) -16))))

% Redefine the *ALLOC Cmacro
% so that it fills in each stack location with NIL.

% (de *ALLOC (frame-size)
%   (setq NAlloc* frame-size)
%   (cond ((zerop frame-size) nil)
%         ((eqn frame-size 1)
%          '((*PUSH 'nil)))
%         ((eqn frame-size 2)
%          '((*PUSH 'nil)
%            (*PUSH 'nil)))
%         ((eqn frame-size 3)   %% ***** remove comments after testing...
%          '((*PUSH 'nil)
%            (*PUSH 'nil)
%            (*PUSH 'nil)))
%         ((eqn frame-size 4)
%          '((*PUSH 'nil)
%            (*PUSH 'nil)
%            (*PUSH 'nil)
%            (*PUSH 'nil)))
%         (t
%          % Push the appropriate number of NILs.
%      `((*MOVE (wconst ,frame-size) (reg T3))
%        (labelgen alloclabel)
%        (*PUSH 'nil)
%        (*WDIFFERENCE (reg T3) (wconst 1))
%        (*JUMPWGREATERP (labelref alloclabel) (reg T3) (wconst 0))))))

(de *ALLOC (framesize)
  (progn
    (setq NAlloc!* framesize)
    (cond
      ((ZeroP framesize)
        NIL)
      (T `((suba!.l ,(times2 framesize addressingunitsperitem) (reg st)))))))

  % a special pass in compiler will do the job

        % Otherwise, we could allocate the space and then the code to clear
        % it out.  Instead, we just push the appropriate number of NILs.

%        (for (from i 1 framesize)
%          (collect
%            '(*PUSH (REG NIL))))))))

% Declare *ALLOC to be a "cmacro".
% *ALLOC function handles its expansion.

(defcmacro *ALLOC)

(de fastapplyp (x) (memq x '(fast-idapply Fastapply)))

(DefCMacro *Call
   ((InternallyCallableP) (bsr (InternalEntry ARGONE)))
   ((Fastapplyp)          (*loc (reg t1) (entry ARGONE))
                          (*move (indirect  (reg t1)) (reg t1))
                          (JSR  (indirect (reg t1))))
           (              (*loc (reg t2) (entry ARGONE))
			  (*move (indirect  (reg t2)) (reg t1))
			  (JSR  (indirect (reg t1)))))

(DefCMacro *DeAlloc
   ((ZeroP))
   ((PosQInumP)         (addq!.l ARGONE (reg st)))
   ((DispInumP)         (adda!.w ARGONE (reg st)))
   (                    (adda!.l ARGONE (REG st))))

(DefCMacro *Exit
   ((ZeroP)             (rts))
   ((PosQInumP)         (addq!.l ARGONE (reg st))
                        (rts))
   ((DispInumP)         (adda!.w ARGONE (reg st))
                        (rts))
   (                    (adda!.l ARGONE (REG st))
                        (rts)))

(DefCMacro *FREERSTR
   ((*CALL fastunbind)(halfword ARGONE)))

(DefCMacro *JCall
   ((InternallyCallableP) (bra (InternalEntry ARGONE)))
   ((Fastapplyp)          (*loc (reg t1) (entry ARGONE))
                          (*move (indirect  (reg t1)) (reg t1))
                          (JMP  (indirect (reg t1))))
	   (              (*loc (reg t2) (entry ARGONE))
                          (*move (indirect  (reg t2)) (reg t1))
                          (JMP (indirect (reg t1)))))



(DefCMacro *Jump
   ((Atom)        (bra ARGONE))% internal labels before compile
   ((TaggedLabel) (bra ARGONE))% compiler generated labels
   ((ImmediateP)  (jmp  (unimmediate ARGONE)))
   (              (jmp ARGONE)))


(DefCMacro *Lbl
   (              ARGONE))

(de *Link (FunctionName FunctionType NumberOfArguments)
  (list
   (cond ((FlagP FunctionName 'ForeignFunction)
          (list '*ForeignLink FunctionName FunctionType NumberOfArguments))
         (t
          (list '*Call FunctionName)))))

(DefCMacro *Link)


(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
 (cons (list '*DeAlloc DeAllocCount)
       (cond ((FlagP FunctionName 'ForeignFunction)
              (list (list '*ForeignLink FunctionName
                           FunctionType NumberOfArguments)
                    '(*Exit 0)))
             (t
              (list (list '*JCall FunctionName))))))

(DefCMacro *LinkE)

%  lisp procedure *LamBind(Regs, FLst);
%  begin scalar X, Y;
%      FLst := reverse cdr FLst;
%      Regs := reverse cdr Regs;
%      while FLst do
%      <<  if null Regs then
%      X := 0
%  else
%  <<  X := cadr car Regs;
%      Regs := cdr Regs >>;
%  Y := list('byte, 16#FF, X) . list('halfword, list('IDLoc, cadar FLst)) . Y;
%  FLst := cdr FLst >>;
%      return '(*CALL FastBind) . Y;
%  end;

(de *lambind (regs flst)
  (let (x y)
    (setq flst (reverse (cdr flst)))
    (setq regs (reverse (cdr regs)))
    (while flst
      (cond ((null regs) (setq x 0))
        (t (setq x (cadr (car regs)))
           (setq regs (cdr regs))))
      (setq y (cons (list (quote byte) 255 x)
            (cons (list (quote halfword)
                (list (quote idloc) (cadar flst))) y)))
      (setq flst (cdr flst)))
    (cons (quote (*call fastbind)) y)))

(DefCMacro *Lambind)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Since all code is moved into data space, the relative jmp used by the
% JUMPON cmacro no longer works.  It is changed here to first do an
% adda.l #l0028,a0
%jmp 2(a0)
% Note this adda.l must be in-line coded because the immediate won't get
% relocated otherwise (i.e., must be and address const)
% Ultimatly this belongs in HP9836-CMAC, but since it's in RLISP>???
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% FOR NOW, THIS ASSUMES 'REGISTER' IS A D REGISTER

(de *JumpOn (Register LowerBound UpperBound LabelList)
(prog (BaseLbl EndLbl Result)
    (setq BaseLbl (Gensym))
    (setq EndLbl (Gensym))
    (setq Result (cons nil nil))
 (cond
   ((ZeroP  LowerBound) (TConc Result (list 'tst!.l       Register)))
   ((QInumP LowerBound) (TConc Result (list '*WDifference Register LowerBound)))
   (T                   (TConc Result (list 'sub!.l       LowerBound Register)))
   )
    (TConc Result (list 'blt EndLbl))
    (TConc Result (list 'cmp!.l (difference UpperBound LowerBound) Register))
    (TConc Result (list 'bgt EndLbl))
    (TConc Result (list 'lsl!.l 2 Register))
    (TConc Result (list 'HALFWORD 16#4EFB))      % JMP BaseLbl(D0.L)
    (TConc Result (list 'HALFWORD (iplus2 16#0802
                      (ilsh (registernumber
                         (cadr register)) 12))))
    (TConc Result BaseLbl)
    (foreach x in LabelList do
       (TConc Result (list 'bra!.w (cadr x))))
    (TConc Result EndLbl)
    (return (car Result))
    ))

(DefCMacro *JumpOn)

(defcmacro *fast-apply-load
   (       (*move argone (reg t2)))       % Really reg 14. But now is fake./LBS
   )

(put 'fast-idapply
     'opencode
     '((*push (reg 1))                  % save first arg.
       (*move  ($fluid SYMFNC) (reg 1))
       (*wshift (reg t2) (wconst 2))    % 4 *  ID number (ignore tag for now)
       (*wplus2 (reg 1) (reg t2))       % add base address to 6 times ID.
       (*wand (reg 1) (wconst 16#7ffffff)) % remove what's left of the tag
       (*move (reg 1) (reg t1))
       (*move (indirect (reg t1))(reg t1)) % Move the address out of the way.
       (*pop (reg 1))                     % retrieve first arg
       (jsr  (indirect (reg t1)))         % jump indirect.
       ))

(put 'fast-idapply
     'exitopencode
     '((*push (reg 1))                  % save first arg.
       (*move  ($fluid SYMFNC) (reg 1))
       (*wshift (reg t2) (wconst 2))    % 4 *  ID number (ignore tag for now)
       (*wplus2 (reg 1) (reg t2))       % add base address to 6 times ID.
       (*wand (reg 1) (wconst 16#7ffffff)) % remove what's left of the tag
       (*move (reg 1) (reg t1))
       (*move (indirect (reg t1))(reg t1))  % Move the address out of the way.
       (*pop (reg 1))                     % retrieve first arg
       (jmp  (indirect (reg t1)))         % jump indirect.
       ))


% Need to do tag stripping before doing the jsr. /LBS
%
(put 'fast-codeapply
     'opencode
     '((*field (reg t2) (reg t2) (wconst infstartingbit)
           (wconst infbitlength))
       (jsr (indirect (reg t2))))
     )

(put 'fast-codeapply
     'exitopencode
     '((*field (reg t2) (reg t2) (wconst infstartingbit)
           (wconst infbitlength))
       (jmp (indirect (reg t2))))

     )

(DE  !*ForeignLink (FunctionName  FunctionType NumberOfArguments)
     (append (PNTH '((!*PUSH (REG 14)) (!*PUSH (REG 13))
             (!*PUSH (REG 12)) (!*PUSH (REG 11))
             (!*PUSH (REG 10)) (!*PUSH (REG  9))
             (!*PUSH (REG  8)) (!*PUSH (REG  7))
             (!*PUSH (REG  6)) (!*PUSH (REG  5))
             (!*PUSH (REG  4)) (!*PUSH (REG  3))
             (!*PUSH (REG  2)) (!*PUSH (REG  1)))
           (difference 15  NumberOfArguments))
         (append
          (append
           (list (list 'jsr (list 'ForeignEntry FunctionName)))
           (cond
        ((eq NumberOfArguments 0) nil)
        ((lessp NumberOfArguments 3)
         (list (list 'addq!.l (times 4 NumberOfArguments) '(reg st))))
        (t
         (list (list 'adda!.l (times 4 NumberOfArguments) '(reg st))))
        ))
          (list
           '(!*move (fluid nil) (reg nil))
           ))))

(DefCMacro *ForeignLink)

% End of file.
