%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PX:SYS-SUPPORT.SL
% Description:  System specific kernel support functions.
% Author:       Brian Beach, Hewlett-Packard`CRC
% Created:      22-Feb-84
% Modified:     29-Jun-84 13:56:27 (RAM)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 08-Sep-86 (Leigh Stoller)
%  Changed kernal-fatal-error to call exit-with-status.
% 29-Jul-86 (Leigh Stoller)
%  Changed all references of fake reg 14 to real reg t2.
% 29-Jun-84 13:54:43 (RAM)
%  Removed stackcheck and other PWS specific functions.
%  Removed PWS register restoration from the end of undefinedfunction.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(lap '((!*entry PlantUnbound expr 1)
       (*push (reg t2))                         % Save temp reg in case
       % The following lines multiply the ID times 6 by using adds. 
       (add!.l   (reg 1)  (reg 1))              % ID*2
       (movea!.l (reg 1)  (reg t2))             % Move to an a reg
       (adda!.l  (reg t2) (reg t2))             % ID*4
       (adda!.l  (reg 1)  (reg t2))             % ID*4+ID*2 = ID*6 

       (!*WPLUS2 (reg t2) (fluid SYMFNC))       
       (move!.w UndefinedFunctionInstruction (indirect (reg t2)))
       (move!.l (plus UndefinedFunctionInstruction 2) 
		(displacement (reg t2) 2))
       (*pop (reg t2))
       (!*EXIT 0)
       UndefinedFunctionInstruction
       (!*CALL UndefinedFunction)
       ))


(lap '((!*entry PlantCodePointer expr 2)
       (*push (reg t2))                         % Save temp reg in case
       % The following lines multiply the ID times 6 by using adds. 
       (add!.l   (reg 1)  (reg 1))              % ID*2
       (movea!.l (reg 1)  (reg t2))             % Move to an a reg
       (adda!.l  (reg t2) (reg t2))             % ID*4
       (adda!.l  (reg 1)  (reg t2))             % ID*4+ID*2 = ID*6 

       (!*WPLUS2 (reg t2) (fluid SYMFNC))
       (move!.w CodePointerInstruction (indirect (reg t2)))
       (move!.l (reg 2) (displacement (reg t2) 2))      
       (*pop (reg t2))
       (!*EXIT 0)
       CodePointerInstruction
       (!*JUMP (fluid SYMFNC))              %SYMFNC used as a dummy to generate
                                            %correct long jump instruction.
       ))


(lap '((!*entry PlantLambdaLink expr 1)
       (*push (reg t2))                         % Save temp reg in case
       % The following lines multiply the ID times 6 by using adds. 
       (add!.l   (reg 1)  (reg 1))              % ID*2
       (movea!.l (reg 1)  (reg t2))             % Move to an a reg
       (adda!.l  (reg t2) (reg t2))             % ID*4
       (adda!.l  (reg 1)  (reg t2))             % ID*4+ID*2 = ID*6 

       (!*WPLUS2 (reg t2) (fluid SYMFNC))                        
       (move!.w LambdaLinkInstruction (indirect (reg t2)))
       (move!.l (plus LambdaLinkInstruction 2) (displacement (reg t2) 2))
       (*pop (reg t2))
       (!*EXIT 0)
       LambdaLinkInstruction
       (!*CALL CompiledCallingInterpreted)
       ))


%(lap '((*entry stackcheck expr 0)
%       (cmpa!.l (fluid stackupperbound) (reg st))
%       (ble overflowedstack)    % 2 cycles less if branch not taken
%       (rts)
%       overflowedstack
%       (adda!.l 100 (reg st))   % roll back the stack
%       (*move '"Stack overflow" (reg 1))
%       (*jcall stderror)))


(lap '((*entry addressapply0 expr 1)
       (*move (reg 1) (reg a0))
       (*jump (indirect (reg a0)))
       ))
  
(de bittable (baseaddress bitoffset)
  (field (ilsh (byte baseaddress (ilsh bitoffset -2))
               (idifference (itimes2 (field bitoffset 30 2) 2) 6))
         30 2))


(lap '((*entry undefinedfunction expr 0)
       (*move (quote "Undefined function called: ") (reg 1))
       (*call console-print-string)
       (*pop (reg d0))
       (subq!.l 6 (reg d0))
       (sub!.l (fluid symfnc) (reg d0))
       (divu 6 (reg d0))
       (lsl!.l 2 (reg d0))
       (add!.l (fluid symnam) (reg d0))
       (*move (reg d0) (reg a0))
       (*move (indirect (reg a0)) (reg 1))
       (*call console-print-string)
       (*call console-newline)
%       (movea!.l (fluid theirregs) (reg a0))
%       (movem!.l (indirect (reg a0)) (reglist (reg A5) (reg a6) (reg A7)))
%       (unlk (reg a6)) 
       (*exit 0)
       ))

(de kernel-fatal-error (string)
  (console-print-string "FATAL ERROR: ")
  (console-print-string string)
  (console-newline)
  (exit-with-status -1)
  )


%(de hpstring (sstring)
%  % May have to recopy into fixed WARRAY ?
%  (strbase (strinf sstring))
%  )

%(de hplength (sstring)
%  (+ (strlen (strinf sstring)) 1)
%  )

%(de popstate ()
%  (when throwsignal*
%    (syserrors_popstate)))

%(de sys_error ()
%  (unwind-protect (systemerror) (popstate)))

%(de key_int_stop ()
%  (keyboard-stop-interrupt)
%  )

%(de key_int_break ()
%  (keyboard-break-interrupt)
%  )

(off fast-integers)
