%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXK:KERNELOUT
% Description:  Generation of binary mini kernel          
% Author:       H.Melenk  ZIB BERLIN             
% Created:      16-Oct-89
% Modified:                         
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Support for kernel declarations
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load fasl-decls))
(loadtime (load compiler))

(dskin "foreignlink.sl")

(fluid '(heaplowerbound heapupperbound bpslowerbound bpsupperbound))

(fluid '(*foreign-functions* *foreign-references* *foreign-fnbr* SYMFOREIGN
         *foreign-variables* *foreign-vnbr* SYMFOREIGNV NEW-CODE-SIZE*
         CODE-SIZE* NEW-BIT-TABLE* NEW-BIT-TABLE-SIZE* ID-HIGH* ID-TABLE* 
         old-id-table* NEW-SYMVAL* NEW-SYMNAM* NEW-SYMFNC* NEW-SYMPRP*
	 new-symget* symget-size*
         SYMNAM-SIZE* SYMVAL-SIZE* SYMFNC-SIZE* SYMPRP-SIZE* ID-ENTRIES* 
         LOCAL-ID-NUMBER* CURRENT-BYTE-NUMBER* heapsize* bpssize*
         kernel-globals*))

(fluid '(NUMBER-OF-segmentS segmentlist* segment-CB-LENGTH segment-CB-TABLE* 
	SYMFOREIGNV-LENGTH* SYMPRP segment-number segment-table))

(fluid '(unixnames-map*))

(setf  *foreign-functions* nil 
       *foreign-references* nil 
       *foreign-fnbr* 0
       SYMFOREIGN (if (null SYMFOREIGN) (gtwarray 200) SYMFOREIGN)
       *foreign-variables* nil
       *foreign-vnbr* 0
       % SYMFOREIGNV (if (null SYMFOREIGNV) (gtwarray 100) SYMFOREIGNV)
	segment-CB-table* (if (null segment-CB-table*) (gtwarray 200) segment-CB-table*)
        segment-table nil
	kernel-globals* nil
        )
 
(setq unixnames-map* '(
	(fopen 		. "p_fopen")
	(fgets 		. "p_fgets")
	(clearerr  	. "p_clearerr")
	(ieee_handler 	. "fpt_exceptions")
        (sigrelse       . "sun3_sigrelse")))


%
%  provide for declarations: they should not appear in the init code
%  so we use special version for fluid and global calls here
%
(put 'fluid 'faslpreeval                              
     (function (lambda (u loadtime?)
                       (cond ((eqcar (second u) 'quote)
                     (foreach x in (second (second u)) do
                          (put x 'vartype 'fluid)))))))
        
(put 'global 'faslpreeval
     (function (lambda (u loadtime?) 
                       (cond ((eqcar (second u) 'quote) 
                     (foreach x in (second (second u)) do
                          (global1 x))))))) 

(defmacro external-function (name args)
     (setf *foreign-functions* (cons name *foreign-functions*))
     (flag1 name 'foreignfunction)
     (setf (get name 'foreign-fnbr) *foreign-fnbr*)
     (setf *foreign-fnbr* (plus *foreign-fnbr* 1))  % adressing units per item?
     (setf *foreign-references* (cons (cons name args) *foreign-references*))
     `(de ,name ,args (,name ,@args)))
 
(defmacro exported-global (list)
        (make-external-global (second list)) nil)

(defmacro exported-fluid (list)
        (make-external-global (second list))
        nil)

(defmacro internal-global (list)
   (for (in id (second list))   % list is quoted
        (do
         (flag1 id 'internalsymbol)
	 (findidnumber id)       
         (put id 'symbol id)
         (put id 'vartype 'global)
         ))
   nil)
 

(defmacro internal-fluid (list)
   (for (in id (second list)) 
        (do
	 (findidnumber id)
         (flag1 id 'internalsymbol)
         (put id 'vartype 'fluid)
         ))
   nil)

(defmacro system-global (list)
   (for (in id (second list))   % list is quoted
        (do
	 (findidnumber id)
         (put id 'vartype 'global)
         )))

(defmacro kernel-global (list)
   (for (in id (second list))
	(do
	 (put id 'vartype 'global)
	 (push id kernel-globals*)
	 (findidnumber id)
	)))

(de make-external-global (u)
   (for (in id u)   
      (do 
        (when (not (flagp id 'exportedsymbol))
         (flag1 id 'exportedsymbol) 
         (flag1 id 'external)
	 (findidnumber id)		% get it in the kernel anyway
	 (push id *foreign-variables*)
         (put id 'foreign-vnbr *foreign-vnbr*) 
         (setq *foreign-vnbr*(plus *foreign-vnbr* addressingunitsperitem)) 
         (put id 'vartype 'global) 
        )))) 

(dm declare-kernel-structure (u)
  (list 'declare-aux-1 (list 'quote (cdr u))))

(flag '(declare-aux-1) 'ignore)

(df declare-aux-1 (u)
  (foreach x in (cadar u) do
    (declare-aux-2 (car x) (cadr x) )))

(flag '(declare-aux-1) 'ignore)

(de declare-aux-2 (name upperbound)
  (prog (n)
	(setq n (times addressingunitsperitem upperbound))
	(findidnumber name)             % generate an ID it doesn't exist.
	(put name 'type 'fluid)         % flag as a fluid variable.
	(flag1 name 'kernel-structure)
	(put name 'segment (setq segment-number (add1 segment-number)))
	(put name 'segmentlength  n)
	(setq segmentlist*(append segmentlist* (list name)))
	(when (unboundp name) (set name (gtwarray upperbound)))
	(set-val name (eval name))
))

(defmacro update-pointers-macro(u)
   (cons 'progn
	(for (in x kernel-globals*)
	  (collect
		`(setf ,x (relocfromload ,x))))))
%
% patch to the compiler: generate indirect load and store for external variables
(commentoutcode
(de pass-1-atom (u vbls)
  % Process an atom.  Anything other than an ID is a constant.
  % IDs are checked for non-local first, since fluid
  % variables are always accessed through the value cell.
  (cond ((not (idp u))                          (mkquote u))
    ((and (constant? u)
          (not (memq u vbls)))              (mkquote (constant-value u)))

%	% access to external variables: (getmem (inf (fluid xxx)))
%   ((flagp u 'external)                    (list 'memory 
%					(list 'field (list '$fluid u) 
%						(mkquote INFSTARTINGBIT)
%						(mkquote INFBITLENGTH))
%					'(wconst 0)))
 
		% access to kernel structure:  (inf xxx)
    ((flagp u 'kernel-structure)            (list 'field (list '$fluid u) 
							(mkquote INFSTARTINGBIT)
							(mkquote INFBITLENGTH)))
    ((non-local u)                          (list (non-local u) u))
    ((memq u vbls)                          (list '$local u))
    (t                                      (make-non-local u))
    ))

)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Initialize Kernel generation
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(*quiet-faslout kernel-magic*))
 
(setq kernel-magic* (plus 80386 (hash-function (date-and-time)))))

(de kernelout(fil)
   (prog (*quiet-faslout faslout-magic)
	(setq faslout-magic kernel-magic*)
	(prin2t "KENRLOUT: (DSKIN files) or type in expressions")
        (prin2t "When all done execute (KERNELEND)")
	(setq *quiet-faslout t)
	(faslout fil)))

(de codefileheader ()
  (allocatefaslspaces))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  postprocessing
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de kernelend ()
  (when (not *writingfaslfile) (stderror "KERNELEND not within KERNELOUT"))
  (kernel-compute-idsizes)
  (kernel-make-ids) 
  (compileuncompiledexpressions)
  (kernel-compute-codesizes)
  (make-new-bit-table)
  (kernel-link)
  (make-C-from-externals)
  (remforeign)
  (make-kernel-file)
  (faslabort)
  )
 
(flag '(kernelend) 'ignore)

%
% compute all relevant sizes
%
 
(de kernel-compute-idsizes()
    (prog (n)
	(setq local-id-number* (difference nextidnumber* first-local-id-number))
	(setq id-high* (plus 128 local-id-number*))
	(setq n (add1 id-high*))
	(set-val 'nextsymbol n)
	(put 'T 'kernel-value 'T)
	(setq symnam-size* (times n addressingunitsperitem))
        (setq symval-size* (times n addressingunitsperitem))
        (setq symnam-size* (times n addressingunitsperitem))
        (setq symprp-size* (times n addressingunitsperitem))
	(setq symget-size* (times n addressingunitsperitem))
        (setq symfnc-size* (round-to-word (times n addressingunitsperfunctioncell)))))

(de kernel-compute-codesizes()
	(when (null bpssize*)
		(setq bpssize* (difference lastbps bpslowerbound)))
	(when (null heapsize*)
		(setq heapsize* (difference heapupperbound heaplowerbound)))
	(setq code-size* (round-to-word  currentoffset*))
	(set-val 'nextbps (plus codebase* code-size*))
	(setq new-code-size* (quotient code-size*  addressingunitsperitem))
	(setq new-bit-table-size* (bit-tablesize new-code-size*))
	(set-val 'bpssize bpssize*)
	(set-val 'heapsize heapsize*)
  )

(de round-to-word (u)
  % u is a (byte-) number minus 1; round u upwards to next word boundary
	(times (quotient (plus u (sub1 addressingunitsperitem)) 
			addressingunitsperitem)
		addressingunitsperitem))

(de kernel-make-ids()
  % process the id's which were generated during kernel compilation.
  % strings are compiled to the bps, their reference is placed to new-symnam*
  % an id-table in faslin-style is constructed
  (prog(u a n)
     (setq n local-id-number*)
     (setq id-table* (mkvect n))
     (setq id-entries* (extract-putentries (car uncompiledexpressions*)))
     (setq old-id-table* (mkvect n))
     (setq new-symval*(mkwrds (gtwrds id-high*)))
     (setq new-symnam*(mkwrds (gtwrds id-high*)))
     (setq new-symprp*(mkwrds (gtwrds id-high*)))
     (setq new-symget*(mkwrds (gtwrds id-high*)))
		% symfnc is allocated as string because it is non word aligned
     (setq new-symfnc* (mkstr (gtstr symfnc-size*)))
     (for (from i 0 127)
       (do (make-kernel-id i (int2id i))))
     (make-kernel-id 128 NIL)
     (for (in x (car orderedidlist*))
      (do
	(let* ((n (new-id-position x))
	       (m (new-id-nbr x)))
		(make-kernel-id m x)
		(setf (words-fetch id-table* n) m)
                (setf (words-fetch old-id-table* n) x))))))
 
(de id-table()
   (for (in x (car orderedidlist*))
        (do  (printf "%w  -> %w %n" x (new-id-nbr x)))))

(de make-kernel-id(nbr name)
     % build the id structure 
   (let ((a (plus2 codebase* currentoffset*)))
	(setf (words-fetch new-symval* nbr) 	
	   (cond%  ((memq name '(SYMFOREIGNV))
		%   (mkitem forward-tag (eval name)))
		((get name 'kernel-value)
		 (get name 'kernel-value))
		%  ((flagp name 'external)
		%  (mkitem forward-tag (plus (get name 'foreign-vnbr) SYMFOREIGNV)))
		(t nil)))
	(setf (words-fetch new-symprp* nbr) nil)
	(setf (words-fetch new-symget* nbr) nil)
	(make-kernel-id-function nbr name)
        (depositword (strlen (strinf (id2string name))))
	(depositstring (list 'STRING (id2string name)))
	(setf (words-fetch new-symnam* nbr) (mkstr a))
         nil))

(de set-val(id val)
	% assign a value to a future id;
	% the new symval might be constructed already, so insert
	% the value there too.
  (prog (n)
	(findidnumber id)
	(setq n (new-id-nbr id))
	(put id 'kernel-value val)
	(when new-symval* (setf (words-fetch new-symval* n) val))))

(de new-id-position (x)
   % get the position of x in the id tables
     (let ((u (get x fasl-idnumber-property*)))
	(when (null u) (stderror (bldmsg "id unknown for kernel: %w" x)))
	(difference u first-local-id-number)))

(de new-id-nbr (x) 
   % get the future id number for x 
	(plus 129 (new-id-position x)))

(fluid '(code-base-hack))

(if_system i386
    % version, where symfnc is a list of addresses 

(de make-kernel-id-function (nbr name)
     % the funtion cell: it is constructed in the function reference of "PATCH"
     % and copied bytewise into the target field
   (let*(*redefmsg
	 (code-base-hack codebase*)    % for safe usage of putentry
	 (ref (assoc name id-entries*))
	 (from (wplus2 symfnc  
	   (wtimes2 addressingunitsperfunctioncell
              (id2int 'patch))))
	 (from1 (wplus2 symfnc  
           (wtimes2 addressingunitsperfunctioncell
              (id2int 'patch1))))
	 (to (plus (inf new-symfnc*) addressingunitsperitem
			(times nbr addressingunitsperfunctioncell)))
	 (n1 (difference addressingunitsperfunctioncell addressingunitsperitem)))

       (if ref (printf "function %w at %x+%x %n" name "BPS" (cdr ref)))
       (putmem to
        (wplus2 codebase*
 	 (if ref (cdr ref) 
                (cdr (assoc 'undefinedfunction id-entries*)))))
        ))
)   % if_system
	

(if_system mc68000
     % version, where symfnc is a list of jump instructions

(de make-kernel-id-function (nbr name)
     % the funtion cell: it is constructed in the function reference of "PATCH"     % and copied bytewise into the target field
   (let*(*redefmsg
         (code-base-hack codebase*)    % for safe usage of putentry
         (ref (assoc name id-entries*))
         (from (wplus2 symfnc
           (wtimes2 addressingunitsperfunctioncell
              (id2int 'patch))))
         (from1 (wplus2 symfnc
           (wtimes2 addressingunitsperfunctioncell
              (id2int 'patch1))))
         (to (plus (inf new-symfnc*) addressingunitsperitem
                        (times nbr addressingunitsperfunctioncell)))
         (n1 (difference addressingunitsperfunctioncell addressingunitsperitem)))
 
        (if ref
            (progn
                (printf "function %w at %x+%x %n" name "BPS" (cdr ref))
                (putentry 'patch1 'expr (cdr ref))
                (putentry 'patch  'expr (cdr ref)))
            (progn
                (setq code-base-hack symfnc)
                (plantunbound (id2int 'patch1))
                (putentry 'patch  'expr
                        (times(new-id-nbr 'undefinedfunction)
                                addressingunitsperfunctioncell))
                ))
        
        (for (from i 0 (sub1 n1))   % instruction part
             (do (setf (byte to i) (byte from1 i))))
                                    % address part
        (for (from i n1 (sub1 addressingunitsperfunctioncell))
             (do (setf (byte to i) (byte from i))))))
        
) % if_system

(de extract-putentries(u)
  (cond	((null u) nil)
	((eqcar (car u) 'putentry)
	  (setf (car uncompiledexpressions*) (delete (car u) (car uncompiledexpressions*)))
	  (cons (cons   (cadr (cadr (car u)))  % doing unquote for the id
			(cadddr(car u)))       % ignoring non EXPRs here (we are in kernel)
		(extract-putentries (cdr u)))) 
        (t (extract-putentries (cdr u)))))

%
% perform an absolute relocation of the kernel code
% This code uses parts or the FASLIN code
%
 

(de kernel-link ()
   (kernel-relocation codebase* new-code-size* bittablebase* id-table*))

(de kernel-relocation (code-base code-size bit-table id-table)
  % CODE-AU-SIZE is the size of the code measured in addressing
  % units, rather than words.
 
  (let ((code-au-size  (wtimes2 code-size addressingunitsperitem)))
    (setq code-base (int2sys code-base))
    (for (from i 0 (wdifference code-au-size 1))
         (do
	  (setf current-byte-number* i)
          (let ((bit-table-entry  (bittable  bit-table i)) 
                (code-location    (wplus2 code-base i)))
            (case bit-table-entry
              ((reloc-word)
               (relocate-word code-location code-base id-table))
              ((reloc-inf)
               (relocate-inf code-location code-base id-table))
              ((reloc-right-half)
               (relocate-right-half code-location code-base id-table))
              )
)))))
        

(de compute-relocation (reloc-tag reloc-inf code-base id-table)
	(prog(u)
		(console-print-string "reloc ")
                (unixputn reloc-tag)(console-newline) 
		(unixputn reloc-inf)(console-newline)
		(unixputn code-base)(console-newline)
		(setq u (compute-relocation1 reloc-tag reloc-inf code-base id-table))
		(console-print-string "-->")
		(unixputn u)(console-newline)
		(return u)))

(de compute-relocation(reloc-tag reloc-inf code-base id-table)
 (when (not (eq reloc-tag reloc-id-number)) (put-new-bit current-byte-number*))
  (case reloc-tag
    ((reloc-code-offset)
     (wplus2 code-base reloc-inf))
    ((reloc-value-cell)
     (cond ((extraargumentp reloc-inf)
            (loc (wgetv argumentblock (makeextraargument reloc-inf))))
           ((local-id-number? reloc-inf)
            (let* ((n (local-to-global-id reloc-inf id-table))
                   (a (local-to-global-id reloc-inf old-id-table*)))
		%(if (flagp a 'external)
		%    (int2sys (plus SYMFOREIGNV (get a 'foreign-vnbr)))
		%    (loc (symval n)))
		(loc (symval n))
		))
           (t
            (loc (symval reloc-inf)))))
    ((reloc-function-cell)
      (if (local-id-number? reloc-inf) 
	(let* ((n (local-to-global-id reloc-inf id-table))) 
		(wplus2 symfnc (wtimes2 n addressingunitsperfunctioncell)))
	(wplus2 SYMFOREIGN (wtimes2 addressingunitsperitem reloc-inf))))
    ((reloc-id-number)
     (if (local-id-number? reloc-inf)
       (local-to-global-id reloc-inf id-table)
       reloc-inf
       ))
    ))

(de make-C-from-externals()
  % generate a C source program for the linkage to the C support functions
    (prog(fil oldhandle *lower n u x)
	(setq number-of-segments (plus 1 segment-number))

	(setq *lower t)
	(setq fil (open "psll.h" 'output))
	(setq oldhandle (wrs fil))
        (prin2t "/* external tables for PSL loader */") (terpri)
          
	     % part 0: definitions
	(printf "#define MAGIC        %w %n" kernel-magic*)
	(printf "#define SEGMENT_MAX     %w %n" number-of-segments)
	(printf "#define INFBITLENGTH %w %n" infbitlength)
	(printf "#define WORD_SIZE    %w %n" addressingunitsperitem)
	(printf "#define JUMP_SIZE    %w %n" addressingunitsperfunctioncell)
	(printf "#define JUMP_OFFSET  %w %n"
			(difference addressingunitsperfunctioncell addressingunitsperitem))
        (cond ((eq infbitlength 24) 
		(printf "#define INF_MASK      0x00ffffff%n")
		(printf "#define TAG_MASK      0xff000000%n"))
              ((eq infbitlength 27)  
                (printf "#define INF_MASK      0x03ffffff%n") 
                (printf "#define TAG_MASK      0xf4000000%n"))
	      (t (stderr "no mask definitions available")))
	(printf "#define ID_TAG       %w%n" id-tag)
	(printf "#define NEGINT_TAG   %w%n" negint-tag)
	(printf "#define POSINT_TAG   %w%n" posint-tag)
	(printf "#define FORWARD_TAG  %w %n" forward-tag) 
	(terpri)
	(printf "#define    SEGMENT_CB      0%n")
	(prog (*raise *lower)
		(setq *raise t)
		(for (in x (append '(symforeign start bps bittable 
					symfnc symnam symval symprp symget heap) 
                                    segmentlist*))
			(do (printf "#define   SEGMENT_%w    %w%n" 
					x (get x 'segment)))))
	(terpri)

             % part 1: external function references
	(setq *foreign-references* (reverse *foreign-references*))
	(for (in u *foreign-references*)
	     (do
		(let*((name (car u))(pl (cdr u))(pl1 pl))
		     (prin2 "extern long ")
		     (prin2 (fgntranslate name))
		     (prin2 "(")
		%    (while pl1
	        %	(prin2 (pop pl1))
		%	(when pl1 (prin2 ",")))
		     (prin2 ");")
		     (terpri))))
        (terpri)
        (prin2t	 "typedef long (*PFI)();")
	(prin2t  "PFI frfntab [] = { ")
	(while *foreign-references*
		(prin2 "   ")
                (prin2 (fgntranslate(car (pop *foreign-references*))))
		(if *foreign-references* (prin2t ",") (prin2t "};")))
	(terpri)

             % part 2: external variables
       (when *foreign-variables*
	(setq n (length *foreign-variables*))

	(for (in x *foreign-variables*)
	     (do (printf "   long * %w;%n" x)))

	(terpri)
        (prin2t "insert_fvars (sval) ")
        (prin2t "      long * sval;   ")
        (prin2t "  { ")
	(setq u *foreign-variables*)
	(while u 
		(setq x (pop u))
		(setq n (get x fasl-idnumber-property* ))
		(setq n (plus 129 (difference n first-local-id-number)))
                (printf "    %w = &sval [%w];%n" x n) 
                )
	(prin2t "  }")

	(terpri)
       )

	(wrs oldhandle)
	(close fil)))   

 

(de fgntranslate (u)
	(let ((v (assoc u unixnames-map*)))
		(if v (cdr v) u)))
%
% convert FASLIN bit-table to MINLOAD bit-table
%

(de make-new-bit-table()
	(setq new-bit-table* (mkwrds (gtwrds new-bit-table-size*)))
	(for (from i 0 new-bit-table-size*)
		(do (setf (words-fetch new-bit-table* i) 0))))
 
(de put-new-bit(i)
   % set bit number i to 1
   (let* ((w (quotient i bitsperword))
	  (b (remainder i bitsperword)))
	(setf (words-fetch new-bit-table* w) 
		(wor (wshift 1 (wdifference 31 b)) (words-fetch new-bit-table* w)))))

(de put-new-bit(i)
   % set bit number i to 1
   (let* ((w (quotient i 8))
          (b (remainder i 8))
	  (a (wplus2 (inf new-bit-table*) addressingunitsperitem)))
        (setf (byte a w)
                (wor (wshift 1 (wdifference 7 b)) (byte a w)))))


(de remforeign()
 (for (in name *foreign-functions*)
      (do (remflag1 name 'foreignfunction))
      ))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  write segments to file
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(setf segment-CB-length (times addressingunitsperitem 6))

(de make-kernel-file()
	(make-file-header)
	(make-segment-description-segment)
	(make-code-block)
	(make-symfnc-block)
	(make-symnam-block)
	(make-symval-block)
	(make-symprp-block)
	(make-symget-block)
	(make-start-address)
	(make-trailer))

(de make-file-header()
        (binarywrite codeout* kernel-magic*)
        (binarywrite codeout* 1111)       % the date
        (binarywrite codeout* (plus 2 1)) % type kernel & relocatable
        (binarywrite codeout* 0)          % sbrk address
)

(setq segment-number -1)
 
(for (in x '(
	   segment-description-segment
           symforeign
       %   symforeignv
           start
           bps
           bittable
           symfnc
           symnam
           symval
           symprp
	   symget
	   heap))
	(do (put x 'segment (setq segment-number (add1 segment-number)))))

(de make-segment-description-segment()
    (prog(n)
	(terpri)
	(setq number-of-segments (plus 1 segment-number))
	(make-block-head 0 0 (times number-of-segments segment-CB-length) 0)
	(make-segment-CB 'segment-description-segment
				 0 (times number-of-segments segment-CB-length) 
				segment-CB-table* 
				(plus segment-CB-table* (times number-of-segments 
							    segment-CB-length)))
	(make-segment-CB 'symforeign
			 0 (times *foreign-fnbr* addressingunitsperitem) 
				SYMFOREIGN 
				(plus SYMFOREIGN 
					(times *foreign-fnbr* 
						addressingunitsperitem)))
        %(make-segment-CB 'symforeignv
	%		 0 *foreign-vnbr* SYMFOREIGNV 
	%				(plus SYMFOREIGNV *foreign-vnbr*))
	(setq n (gtwarray 1))
	(make-segment-CB 'start 0 addressingunitsperitem n 
				(plus n addressingunitsperitem))
	(make-segment-CB 'bps 1 bpssize* codebase*(plus codebase* currentoffset*))
	(make-segment-CB 'bittable 0 (bit-tablesize bpssize*)
				 new-bit-table* new-bit-table*)
	(make-segment-CB 'symfnc 0 
		(times addressingunitsperitem (compiler-constant 'maxsymbols))
			symfnc (plus symfnc symfnc-size*))
	(make-segment-CB 'symnam
			 0 (times addressingunitsperitem 
				(compiler-constant 'maxsymbols))
			symnam (plus symnam symnam-size*))
	(make-segment-CB 'symval
			 0 (times addressingunitsperitem 
				(compiler-constant 'maxsymbols))
			symval (plus symval symval-size*))
	(make-segment-CB 'symprp
			 0(times addressingunitsperitem 
				(compiler-constant 'maxsymbols))
			symprp (plus symprp symprp-size*))
        (make-segment-CB 'symget
                         0(times addressingunitsperitem 
				(compiler-constant 'maxsymbols))
                        symget (plus symget symget-size*))
	(make-segment-CB 'heap 0 (plus heapsize* heapsize*)
		 heaplowerbound heaplowerbound)
        (for(in x segmentlist*)
	  (do 
		(make-segment-CB x 0 (get x 'segmentlength) (eval x)
			(sys2int (plus (eval x) (get x 'segmentlength))))))))
 
(de make-code-block ()
	(make-block-head 'bps 0 (byte2wordboundary code-size*) 1)
	(binarywriteblock codeout* codebase* (byte2words code-size*))
	(make-block-head -1 0 (times new-bit-table-size* addressingunitsperitem))
	(binarywriteblock codeout* (loc (words-fetch new-bit-table* 0)) 
			new-bit-table-size*))

(de make-symfnc-block ()
	(make-block-head 'symfnc 0 (byte2wordboundary symfnc-size*) -2)
	(binarywriteblock codeout* (loc (words-fetch  new-symfnc* 0))  % should be byte!!
				(byte2words symfnc-size*)))

(de make-symnam-block ()
	(make-block-head 'symnam 0 (byte2wordboundary symnam-size*) -1)
	(binarywriteblock codeout* (loc (words-fetch new-symnam* 0))
				(byte2words symnam-size*)))
(de make-symval-block () 
        (make-block-head 'symval 0 (byte2wordboundary symval-size*) -1) 
        (binarywriteblock codeout* (loc (words-fetch new-symval* 0)) 
                                (byte2words symval-size*)))
(de make-symprp-block ()  
        (make-block-head 'symprp 0 (byte2wordboundary symprp-size*) -1)  
        (binarywriteblock codeout* (loc (words-fetch new-symprp* 0))  
                                (byte2words symprp-size*)))
(de make-symget-block ()
        (make-block-head 'symget 0 (byte2wordboundary symget-size*) -1)
        (binarywriteblock codeout* (loc (words-fetch new-symget* 0))
                                (byte2words symget-size*)))
(de make-start-address ()
	(make-block-head 'start 0 addressingunitsperitem -3)
	(binarywrite codeout* (plus codebase* initoffset*)))

(de make-trailer() (make-block-head 0 0 0 0))

(de make-block-head (name rel len rtype)
	(binarywrite codeout* (if (numberp name) name (get name 'segment)))
	(binarywrite codeout* rel)
	(binarywrite codeout* len)
	(binarywrite codeout* rtype))
 
(de make-segment-CB (name type size low high)
	(printf " segment %w=%w type %w size %x low %x high %x %n"
		(if (numberp name) name (get name 'segment)) name type size low high)
	(binarywrite codeout* (if (numberp name) name (get name 'segment)))
	(binarywrite codeout* (if (idp name) (new-id-nbr name) name))
        (binarywrite codeout* type) 
        (binarywrite codeout* size) 
        (binarywrite codeout* low)
	(binarywrite codeout* (isub1 high)))
 
(de byte2wordboundary(u)
   % round byte number to next word boundary
	(times addressingunitsperitem (byte2words u)))

(de byte2words(u) 
   % convert a byte number (counting from 0) into a word number (counting from 1)
	(add1 (quotient u addressingunitsperitem)))

(de bit-tablesize (u)
   % u is the number of bytes in bps; result is the number of bytes for the
   % corresponding bit-table rounded to word boudnary
   (setq u (quotient (plus u 7) 8))  % the number of bytes needed
   (times (quotient (plus u 
                          (sub1 addressingunitsperitem )) 
                    addressingunitsperitem) 
          addressingunitsperitem))
