%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:LAP.SL
% Description:  Miscellaneous LAP routines for the D-register model
% Author:       Sam Sands, Tim Tillson
% Created:      12-Dec-83
% Modified:     23-Apr-84 10:21:43
% Mode:         Lisp
% Package:      Kernel
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%  Added header
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(global '(bndstkptr bndstkupperbound))

(lap '(              % This file contains miscellaneous LAP routines    scs


% syslsp procedure CodeApply(CodePtr, ArgList);
%
% if this could be written in Syslisp, it would look something like this:
% The arguments, up to 14, are loaded into the registers and execution jumps
% to the code pointer.
%
% begin scalar N;
%     N := 0;
%     while PairP ArgList do
%     <<  N := N + 1;
%	  ArgumentRegister[N] := car ArgList;
%	  ArgList := cdr ArgList >>;
%     (jump to address of code pointer)
% end;

% Note that this code assumes that (reg t2) is an A register            scs
%    and that an (EXIT 0) will do an RTS
%    and that the tagged code pointer is a valid address

  (!*entry CodeApply expr 2)           %  CodeApply(CodePointer, ArgList)
     %  code pointer, ready for RTS, but first it is tag stripped, then pushed
  (*inf (reg 1) (reg 1))
  (*move (reg 1) (reg s7))             % callers address

  (!*jumpeq (Label L0) (reg  2)  (reg NIL))    %  done if no arguments

  (!*move (reg 2) (reg s6))            %  move argument list to safe place

  (!*move (car (reg s6)) (reg  1))     %  get actual argument from list
  (!*move (cdr (reg s6)) (reg s6))     %  advance pointer
  (!*jumpeq (Label L0) (reg s6)  (reg NIL))    %  test for end of list  

  (!*move (car (reg s6)) (reg  2))     %  repeat for each argument
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg  3))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg  4))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg  5))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg  6))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg  7))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg  8))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 
  
  (!*move (car (reg s6)) (reg  9))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg 10))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg 11))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg 12))  
  (!*move (cdr (reg s6)) (reg s6))  
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*move (car (reg s6)) (reg 13))  
  (!*move (cdr (reg s6)) (reg s6))
  (!*jumpeq (Label L0) (reg s6)  (reg NIL)) 

  (!*jumpnoteq (Label TooMany) (cdr (reg s6)) (reg NIL))  
                                       %  check for more than 14

  (!*move (car (reg s6)) (reg 14))     % the 14th and last argument 

L0
  (!*move (reg s7) (reg a1))
  (!*move (idloc codeapply) (reg s7))
  (calls (indexed (reg a1) 0))
  (*exit 0)

TooMany
	(!*MOVE (QUOTE "Too many arguments to function") (reg 1))
	(!*JCALL StdError)
))

% procedure wtimes2
% procedure wquotient
% procedure wremainder

% these routines are written so that in the normal case they use
% only registers (reg 1) (reg 2) and (reg 9)

%
% TAG( CompiledCallingInterpreted )
%


(lap '((!*entry CompiledCallingInterpreted-1 expr 0)	%. link for lambda
       %
       % Called by JMP in the function cell. Stores the ID of the interpreted
       % function in CodeForm!* without disturbing its argument registers
       %
       (*mkitem (reg s7) (wconst id-tag))
       (*move (reg S7) (reg a1))     % the id-nbr
       (!*MOVE (reg a1) (Fluid CodeForm!*))
       (!*JCALL CompiledCallingInterpretedAux)
       ))
	

%
% TAG( FastApply )
%

(lap '((!*entry FastApply expr 0)	%. Apply with arguments loaded
%
% Called with arguments in the registers and functional form in reg t2 (14)
% Note: this routine is called from the compiler
%       
        (*move (reg t2)(reg a3))
        (*tag (reg a2) (reg a3))
	
        (*move (reg a3)(reg a1))
        (!*inf (reg a1) (reg a1))
	
        (!*JUMPNOTEQ (Label NotAnID) (reg a2) (wconst id-tag))
        
        (*move (reg a1) (reg a2))
        (*wtimes2 (reg a1) 4)
	(*wplus2  (reg a1) (fluid symfnc))
        (*move (indexed (reg a1) 0) (reg a1))
        (*move (reg a2) (reg s7))

        (calls (indexed (reg a1) 0))
        (*exit 0)              

NotAnID
	(!*JUMPNOTEQ (Label NotACodePointer) (wconst code-tag)(reg a2) )

        (calls (indexed (reg a1) 0)) 
        (*exit 0)

NotACodePointer
	(!*JUMPNOTEQ (Label IllegalFunctionalForm) (wconst pair-tag) (reg a2))
        
        (!*MOVE (indexed (reg a1) 0) (reg a4))  % CAR with pair already untagged
        
        (!*move (QUOTE LAMBDA) (reg a5))
	(!*JUMPNOTEQ (Label IllegalFunctionalForm) (reg a4) (reg a5))

        (!*MOVE (reg a3) (Fluid Codeform!*))     %put lambda in codeform!* 
	(!*JCALL FastLambdaApply)
        (*exit 0)

IllegalFunctionalForm
	(!*MOVE (QUOTE "Illegal functional form in Apply") (reg 1))
	(!*MOVE (reg a3) (reg 2))
	(!*CALL List5)
	(!*JCALL StdError)
))

%
% TAG( UndefinedFunction )
%

(de undefinedfunctionaux(id)
     (StdError (BldMsg "Undefined function %r called from compiled code" id)))
