%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:CHAR-IO.SL
% Title:        Bottom level character IO primitives
% Author:       Eric Benson
% Created:      27 August 1981
% Modified:     31-May-84 11:04:45 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:IO-DECLS.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 22-oct-86 W. Neun , ZIB Berlin
%  changed channel.....char to be allocfree
% 31-May-84 10:13:23 (Brian Beach)
%  Changed calls on IDAPPLY1, IDAPPLY2 to use IDAPPLY.
%  Assorted cleanup.
% 01-Dec-83 14:23:04 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(in* out*))
 
(compiletime (load io-decls))
 
(on fast-integers)
 
 
(de channelreadchar (channel)
 
  % Read one character from CHANNEL.
  %
  % All channel input must pass through this function.  When a channel is
  % open, its read function must be set up.
  %
 
  (if (and (wleq 0 channel) (wleq channel maxchannels))
 
   (let ((charr (wgetv unreadbuffer channel)))
     (if (neq charr (char null))
       (progn (setf (wgetv unreadbuffer channel) (char null))
              charr)
      (idapplye channel (wgetv readfunction channel))
    ))
    (noniochannelerror channel "ChannelReadChar")
   )
 )
 
 
(de readchar ()
 
  % Read single char from current input
 
  (channelreadchar in*)
  )
 
 
(de channelwritechar (channel charr)
 
  % Write one characater to CHANNEL.
  %
  % All channel output must pass through this function.  When a channel is
  % open, its write function must be set up, and line position set to zero.
  %
  % preload for CRAY, and some hack
 
(prog (linepo writefu x)
  (setq linepo (wgetv lineposition channel))
  (setq writefu (wgetv writefunction channel))
  (setq x (wrshift (wdifference charr (char space)) 63))
     % is 1 when charr < space i.e control char ASCII
     % is 0 else
  (when (eq 0 (wor x (wxor channel (wand channel maxchannels))))
             (setf (wgetv lineposition channel) (wplus2 linepo 1))
             (return (idapplye2 channel charr writefu)))
    % that's a hack, isnt it?
  (unless (and (wgeq channel 0) (wleq channel maxchannels))
    (return (noniochannelerror channel "ChannelWriteChar")))
 
  (cond
   ((weq charr (char eol))
    (setf (wgetv lineposition channel) 0)
    (setf (wgetv pageposition channel) (wplus2 (wgetv pageposition channel) 1))
    )
   ((weq charr (char tab))
    (setf (wgetv lineposition channel)
      (wand (wplus2 linepo 8) (wnot 7)))
    )
   ((weq charr (char ff))
    (setf (wgetv pageposition channel) 0)
    (setf (wgetv lineposition channel) 0)
    )
   (t
    (setf (wgetv lineposition channel) (wplus2 linepo 1))
    ))
 (return 
  (idapplye2 channel charr writefu)
)))
 
(de writechar (ch)
 
  % Write single char to current output
 
  (channelwritechar out* ch)
  )
 
 
(de channelunreadchar (channel ch)
 
  % Input backup function
  %
  % Any channel input backup must pass through this function.  The following
  % restrictions are made on input backup:
  %     1. Backing up without first doing input should cause an error, but
  %       will probably cause unpredictable results.
  %     2. Only one character backup is supported.
  %
 
  (setf (wgetv unreadbuffer channel) ch)
  )
 
 
(de unreadchar (ch)
 
  % Backup on current input channel
 
  (channelunreadchar in* ch)
  )
 
 
(off fast-integers)
 
