%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         LOAD-PSL-internal.SL  build NON EXPORT versions
% Description:  Load in the files comprising BARE-PSL
% Author:       Brian Beach, Hewlett-Packard CRC
% Created:      11-May-84
% Modified:     20-Jun-84 13:29:01
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, for CRAY version
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 3 May 86 W. Neun (ZIB Berlin)
%  CRAY COS version
% Tue Oct  8 05:56:55 1985 (Russ Fish)
%  Removed hardwired path to $pnkl from filenames, since the make-bare-psl
%  script cd's to $pnkl before starting bpsl.
% 16-Dec-84 08:30 (Brian Beach)
%  Added VAX-ENVIRONMENT. Removed CHAR-MACRO.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
(on fast-integers)
 
(fluid '(bootstrap-files pnk-files pvnk-files error-files
         *noreprieve show-new-ids optionaddresses* ))
 
(compiletime (flag '(loader-main suck-in-files suck-in trace-message)
                     'internalfunction))
 
(de loader-main ()
  (suck-in-files bootstrap-files)
  (suck-in-files pnk-files)
  (suck-in-files pvnk-files)
  (suck-in-files error-files)
  (suck-in "putd-getd.b")
  (suck-in "cray-specifics.b")
  (setf $eof$ (mkid 26))     % Must find the correct place to do this.
  (console-print-string "NonKernel Library loaded")
  (libclose1 'NonKernelLib)
  (cos-msgio nil)          % no more io-messages
  (setq optionaddresses* (list (cons 'kernel (cons 0 bps))
                               (cons 'nonkernel (cons bps nextbps))
  )                      )
  (setf *noreprieve nil)         %enable reprieve
  (remob 'bootstrap-files)
  (remob 'pnk-files)
  (remob 'pvnk-files)
  (remob 'error-files)
  (pre-main)
  )
 
(de suck-in-files (file-list)
  (for (in f file-list)
       (do
      (suck-in f)
      )) 
  )
 
(de suck-in (file-name)
  (cond
   (show-new-ids
    (progn
      (console-print-string "loading ")
      (console-print-string file-name)
      (console-print-string "heap-level:")
      (putoct heaplast)
      (console-newline)
      (console-newline)
  )))
  (faslin file-name)
  )
 
(de trace-message (string)
  (console-print-string "TRACE: ")
  (console-print-string string)
  (console-newline)
  )
 
(setf bootstrap-files
  '(
    "constants.b"
    "comp-support.b"
    "cons-mkvect.b"
    "property-list.b"
    "symbol-values.b"
    "type-conversions.b"
    "easy-sl.b"
    "function-primitives.b"
    "fluid-global.b"
    ))
 
(setf pnk-files
  '(
    "allocators.b"
    "apply-lap.b"
    "newarithmetic.b"
    "autoload.b"
    "autoload-trace.b"
    "backtrace.b"
    "base-io.b"
    "binding.b"
    "break.b"
    "carcdr.b"
    "catch-throw.b"
    "char-io.b"
    "newcompacti.b"
    "relink.b"
    "cont-error.b"
    "copiers.b"
%   "copying-gc.b"
    "defconst.b"
    "define-smacro.b"
    "dskin.b"
    "easy-non-sl.b"
    "equal.b"
    "eval-apply.b"
    "eval-when.b"
    "explode-compress.b"
    "io-extensions.b"
    "known-to-comp-sl.b"
    "lisp-macros.b"
    "load.b"
    "loop-macros.b"
    "mini-editor.b"
    "oblist.b"
    "onoff.b"
    "open-close.b"
    "other-io.b"
    "others-sl.b"
    "printers.b"
    "printf.b"
    "prog-and-friends.b"
    "rds-wrs.b"
    "read.b"
    "sequence.b"
    "sets.b"
    "string-gensym.b"
    "token-scanner.b"
    "top-loop.b"
    "vectors.b"
    ))
 
(setf pvnk-files
  '(
    "io-data.b"
    "scan-table.b"
    "system-io.b"
    "memory-sizes.b"
    "cos-extensions.b"
    "floats.b"
    ))
 
 
(setf error-files
  '(
    "error-errorset.b"
    "error-handlers.b"
    "io-errors.b"
    "type-errors.b"
    ))
 
(off fast-integers)
 
(loader-main)
 
