%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:SETS.SL 
% Title:        Functions acting on lists as sets 
% Author:       Eric Benson 
% Created:      12 December 1981 
% Modified:     23-Mar-84 14:58:07 (Brian Beach) 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 01-Dec-83 15:01:54 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(de list2set (l)
  % Remove redundant elements from L
  (cond ((not (pairp l)) nil)
        ((member (car l) (cdr l)) (list2set (cdr l)))
        (t (cons (car l) (list2set (cdr l))))))

(de list2setq (l)
  % EQ version of List2Set
  (cond ((not (pairp l)) nil)
        % Don't confuse it with SetQ!
        ((memq (car l) (cdr l)) (list2set (cdr l)))
        (t (cons (car l) (list2set (cdr l))))))

(de adjoin (element aset)
  % Add Element to Set
  (if (member element aset)
    aset
    (cons element aset)))

(de adjoinq (element aset)
  % EQ version of Adjoin
  (if (memq element aset)
    aset
    (cons element aset)))

(de union (x y)
  % Set union
  (if (not (pairp x))
    y
    (union (cdr x) (if (member (car x) y)
             y
             (cons (car x) y)))))

(de unionq (x y)
  % EQ version of UNION
  (if (not (pairp x))
    y
    (unionq (cdr x) (if (memq (car x) y)
              y
              (cons (car x) y)))))

(de xn (u v)
  % Set intersection
  (cond ((not (pairp u)) nil)
        ((member (car u) v) (cons (car u) (xn (cdr u) (delete (car u) v))))
        (t (xn (cdr u) v))))

(de xnq (u v)
  % EQ version of XN
  (cond ((null (pairp u)) nil)
        ((memq (car u) v) (cons (car u) (xn (cdr u) (delq (car u) v))))
        (t (xn (cdr u) v))))

(de intersection (u v)
  (xn u v))

(de intersectionq (u v)
  (xnq u v))


