%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PVC:FC-INDIRECT.SL
% Description:  Builds an indirect call to a function.
% Author:       Brian Beach, Hewlett-Packard CRC
% Created:      24-Apr-84
% Modified:     01-May-84 10:16:11 (Brian Beach)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% version for MIPS, Winfried Neun, ZIB ,28. August 1990

(compiletime (setq *syslisp t))

(imports '(sys-consts))

(on fast-integers)

(fluid '(symfnc))

(de define-indirect-function (calling-function called-function)

  % Build a  calling function named (calling-function) that simply calls
  % another function (called-function).
  % No knowledge of the number of arguments is needed.
  % Note that the interpreter will do no error checking on number of
  % arguments for a function defined by this routine.

  (if (member 'compiler options*)
    (lap `((*entry ,calling-function expr -1)
	   (*alloc 0)
	   (*linke 0 ,called-function expr -1)
	   ))
    (*hacked-define-indirect-function calling-function
				      called-function)
    )
  calling-function   % Return the name of the defined function
  )

(de *hacked-define-indirect-function (calling-function called-function)
  
  % Builds the function linkage by hand, without use of the compiler.
  % First, some BPS is allocated, and code is placed in it to jump to
  % the called function.
  % CODE-BASE points to the start of the segment allocated by GTBPS, where
  % the argument count resides.
  % CODE-ADDRESS points to the beginning of the actual code.

  (let* ((code-base     (gtbps 4))
	 (code-address  (loc (wgetv code-base 1)))
         (sixteen 16)
         (eighteec 16#8EEC)
	 (called-address 
               (wand 16#FFFF (wplus2 -32000 (* 4 (idinf called-function))))))

    (wputv code-base 0 -1)      % Number of arguments = undefined

    (wputv code-base 1 (wor (wshift eighteec 16) called-address))
                    % a BA and a nop  
    (wputv code-base 2 0) % a NOP
    (wputv code-base 3 (int2sys 16#0180f808)) % jr instr
    (wputv code-base 4 (wor (idinf called-function) (wshift 16#340b sixteen)))

    % Now the code is built.  Next we make a code pointer that points to
    % it and plant it in the function cell of the newly defined function:

    (flushcache code-base (wplus2 code-base 5))
    (putd calling-function 'expr (mkcode code-address))
    ))

(off fast-integers)

(compiletime (setq *syslisp NIL))
