%
% faults.sl - run-time fault handler for apollo PSL.
%
% Author:      John.W. Peterson
%              Computer Science Dept.
%              University of Utah
% Date:        1986
% Copyright (c) 1986 J.W. Peterson
%
(compiletime (load useful))

(bothtimes (load compiler))

% dumplisp to get addressoffunction and syscalls package.

% System call to install the fault handler(s)

(de pfm_establish_fault_handler( targ_stat funcptr )
  (let* (status)
	(putarg 1 targ_stat '(integer32))
	(putarg 2 0 '(integer16))       % No options
	(putarg 3 funcptr '(integer32))
	(cleararg 4)
	(syscall 'PFM_$ESTABLISH_FAULT_HANDLER 4)
        (setf status (getarg 4 '(integer32) '(fixn)))
        (cond ((equal status 0)
	       (apolloResetStatus)
	       (getarg 4 '(integer16) '(Inum)) % Returned value
	      )
	      (T (apolloError status)))))

% PSL fault handler, to be called from Aegis.

(lap `((!*entry aegis_fault_handler expr 0)
       
       % This code simply mimics the code generated by the Pascal compiler.
       % The idea is just to keep track of the stack so we can fetch the args
       
       (pea (indirect (reg a5)))	% Pea (db)  st -4
       (clr!.l (predecrement (reg st))) %          st -8
       (pea (indirect (reg a0)))        %          st -12
       (movea!.l (displacement (reg a0) 6) (reg a5))
       (fullword 16#4e56FFFC)		% LINK SB,#-4  % st -20
       
       % Fetch the arguments (status code into reg 1)

       (movea!.l (displacement (reg a6) 16#14) (reg a2))  % Rec base ptr
       (movea!.l (displacement (reg a2) 2) (reg a1))      % actual arg

       % Restore PSL registers Pascal may have stepped on.

       (move!.l (fluid nil) (reg nil))
%      (!*move 0 (reg a6))

       (movea!.l SystemErrorMsg (reg a2))
       (move!.l (reg nil) (reg a3))     % Errorform!* for ContAegisError

%%     (cmpa!.l (immediate 16#120010) (reg a1))  % is error process quit?

%%     (beq (label quit_fault_error))

       % Call error with the "system fault" message.       
       (jsr (entry aegisError))

%%%% Continuable error doesn't work - stack gets trashed.
%%%quit_fault_error
%%%       (jsr (entry contAegisError))

       (rts)			% Should never return...

SystemErrorMsgBytes
       (fullword 11)
       (string "system fault")
SystemErrorMsg
       (fullword (mkitem 4 SystemErrorMsgBytes))))

%
% Version of Syscall that does absolutly no consing
%
(lap '(
(!*entry no_cons_XSysCall0 expr 1)
        (!*MOVE (reg 1) (reg a0))
        (jsr (displacement (reg a0) 0))         % jump to Aegis routine
        (!*Move   (reg d0)    (reg 1))
        (!*MOVE   (Fluid NIL) (Reg NIL))  % fix up the LISP regs
%       (!*MOVE   0           (REG a6))
        (!*EXIT 0)
))

% Set up handlers for the common fault times (Note - you must call this
% again if you start up from a dumped image.

(fluid '(psl_fault_call_addr *inhibit_entry*))

(de install_fault_handlers()
    % System faults
    (PFM_ESTABLISH_FAULT_HANDLER 16#120000 psl_fault_call_addr)
    % Memory addressing errors
    (PFM_ESTABLISH_FAULT_HANDLER 16#40000 psl_fault_call_addr)

    % Prevent asynch faults during GC

    (setf *inhibit_entry* (LookUpAegisEntry 'PFM_$INHIBIT))

    % We call this by hand, because it is VERY IMPORTANT that a
    % cons doesn't occur when BeforeGCSystemHook is called (obviously,
    % we're already out of heap at that point!)

    (de BeforeGCSystemHook()
        (no_cons_Xsyscall0 *inhibit_entry*))
    (de AfterGCSystemHook()
        (syscall 'PFM_$ENABLE 0))

    (printf "*** Fault handlers installed.%n")
)

% Load time initialization:

    % Build dummy Aegis ECB (Entry Control Block).  This is a pointer to
    % the function and it's stack frame, but we just use the PC address, since
    % PSL doesn't care about maintaining an Aegis stack frame.

    (setf psl_fault_call_addr (gtbps 8))
    (putword psl_fault_call_addr 0 (addressoffunction 'aegis_fault_handler)) 
    (putword psl_fault_call_addr 1 (addressoffunction 'aegis_fault_handler))

    (install_fault_handlers)    



