%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           symflags.sl
% Description:    fast flags stored in a vector (up to 64 with CRAY)
%                 used for reducing the number of propertylist-requests
% Author:         Winfried Neun   ZIB Berlin
% Created:        16-October-1986
% Mode:           Lisp
% Package:
% Status:         Experimental (Do Not Distribute)
%
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% symflags are used in two different ways. For flags all property list
% activities are done via fast vector access. For get calls on symflags
% the NIL case is done first. If no entry is present in symflags vector
% no get call will take place.
% The property to be a SYMFLAG must be set at the very early beginning
% of compilation
% dont compile this source with symflags loaded !!
 
(compiletime
   (ds string-concat(a b)
        (compress (append (explode2 a)(explode2 b)))
)  )
 
(compiletime
 (defmacro def-pass-1-macro (name args . body)
   (let ((fcn-name (intern (string-concat "PA1M-" (id2string name)))))
     `(progn
       (put (quote ,name) (quote pass-1-macro) (quote ,fcn-name))
       (defmacro ,fcn-name ,args ,@body)
       ))))
 
(def-pass-1-macro get (a b)
  (when (and (pairp b)
           (get (cadr b) 'symflag?))
              `((lambda (x)
                    (cond ((and (idp x) (symflagp x ,b)) (!%get x ,b))
                           (t nil))) ,a )))
 
(def-pass-1-macro flagp (a b)
  (when (and (pairp b)
           (get (cadr b) 'symflag?))
              `((lambda (x)
                    (cond ((and (idp x) (symflagp x ,b)) t)
                           (t nil))) ,a )))
 
(de install-symflag (x val)
       (put x 'symflag? t)
       (put x 'symflagval val))
 
(put 'symflag 'openfn  '(subpat *symflag))
 
(de *symflag (ArgOne ArgTwo)
    (expand2operandcmacro ArgOne ArgTwo '*symflag))
 
(defcmacro *symflag  ((*move argone (reg a6))
                      (*move (reg t15) (reg a7))
                      (*wplus2 (reg a6) (reg a7))
                      (*move (indexed (reg a6) 1) (reg s6))
                      (*move 1 (reg s7))
                      (*wrshift (reg s6) argtwo)
                      (*wand (reg s6) (reg s7))
                      (*move (reg s6) argone))
                      )
(compiletime (progn
(put 'putsymflag 'opencode '((*move (reg 1) (reg a6))
                             (*move (reg t15) (reg a7))
                             (*wplus2 (reg a6) (reg a7))
                             (*move (indexed (reg a6) 1) (reg s6))
                             (*move (reg 2) (reg a5))
                             (*move 1 (reg s7))
                             (*wshift(reg s7) (reg a5))
                             (*wor  (reg s6) (reg s7))
                             (*move (reg s6) (indexed (reg a6) 1))
                             ))
 
(put 'delsymflag 'opencode '((*move (reg 1) (reg a6))
                             (*move (reg t15) (reg a7))
                             (*wplus2 (reg a6) (reg a7))
                             (*move (indexed (reg a6) 1) (reg s6))
                             (*move (reg 2) (reg a5))
                             (*move 1 (reg s7))
                             (*wshift(reg s7) (reg a5))
                             (*wnot (reg s7) (reg s7))
                             (*wand (reg s6) (reg s7))
                             (*move (reg s6) (indexed (reg a6) 1))
                             ))
))%end compiletime
 
(dm symflagp (u)
  (cond ((not (get (cadr (caddr u)) 'symflag?)) nil)
      (t `(weq (symflag ,(cadr u) ,(get (cadr(caddr u)) 'symflagval))1)
      )))
 
(fluid '(&symflag&))
 
(compiletime (flag '(sett15) 'internalfunction))
 
(lap '((*entry sett15 expr 1)
       (*move (reg 1) (reg t15))
       (*end 0)
)     )
 
(de !%put (a b c)
         (when (get b 'symflag?) (putsymflag a (get b 'symflagval)))
         (old-put a b c))
 
(when (funboundp 'old-put) (copyd 'old-put 'put)
                           (copyd 'put '!%put))
 
(de !%remprop (a b)
         (when (get b 'symflag?) (delsymflag a (get b 'symflagval)))
         (old-remprop a b))
 
(when (funboundp 'old-remprop) (copyd 'old-remprop 'remprop)
                               (copyd 'remprop '!%remprop))
 
(de !%flag1 (a b)
         (when (get b 'symflag?) (putsymflag a (get b 'symflagval)))
         (old-flag1 a b))
 
(when (funboundp 'old-flag1) (copyd 'old-flag1 'flag1)
                             (copyd 'flag1 '!%flag1))
 
(de !%remflag1 (a b)
         (when (get b 'symflag?) (delsymflag a (get b 'symflagval)))
         (old-remflag1 a b))
 
(when (funboundp 'old-remflag1) (copyd 'old-remflag1 'remflag1)
                                (copyd 'remflag1 '!%remflag1))
 
(null (if (memq 'double-symbol-space options*)
      (setq &symflag& (sett15 (make-words (wtimes2 2 maxsymbols) 0)))
      (setq &symflag& (sett15 (make-words maxsymbols 0)))))
 
(remob 'sett15)
(remob '!%put)
(remob '!%remprop)
(remob '!%flag1)
(remob '!%remflag1)
