%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PU:READ1MACROS.SL 
% Description:  First part of some specilized reader macros 
% Author:       Don Morrison, Hewlett-Packard CRC 
% Created:      Wednesday, 12 May 1982 
% Modified:     6 May 1984 1558-PDT 
% Package:      Utilities 
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 6 May 1984 1538-PDT (Nancy Kendzierski)
%  Split read-macros.sl into two parts to allow them to be separately
%  compilable without requiring itself.  This part defines the
%  ` (backquote) read-macro necessary to compile the second part.
%  Since char-macro is being moved out of the kernel into useful, and
%  char-macro.sl needs this backquote stuff, modified this file to
%  not use CHAR.
% 19 Jan 1984 1449-PST (Brian Beach)
%   Added standard header.
% 1 Feb 1983 1400-PST  (Cris Perdue)
%   Dochar moved into "nonkernel", "C" for "CONTROL", etc. commented out.
%   Many miscellaneous symbolic names for characters removed.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load backquote))

((lambda (o-table)
   (setq LispScanTable* (TotalCopy o-table)) % in case it's in pure space
   (cond ((eq CurrentScanTable* o-table)
	   (setq CurrentScanTable* LispScanTable*))))
  LispScanTable*)

% plug backquote and friends into the lisp reader via read macros
% ` for backquote, , for unquote, ,@ for unquotel, and ,. for unquoted

(de backquote-read-macro (channel qt)
  (list 'backquote (ChannelReadTokenWithHooks channel)))

(de unquote-read-macro (channel qt)
  (list 'unquote (ChannelReadTokenWithHooks channel)))

(de unquotel-read-macro (channel qt)
  (list 'unquotel (ChannelReadTokenWithHooks channel)))

(de unquoted-read-macro (channel qt)
  (list 'unquoted (ChannelReadTokenWithHooks channel)))

(putv LispScanTable* 96 11)   % CHAR not yet defined;  96 = (char !`)

(putv LispScanTable* 44 13)   % CHAR not yet defined;  44 = (char !,)

(put '!, (getv LispScanTable* 128) '((!@ . !,!@)(!. . !,!.)))

(deflist
  '((!` backquote-read-macro)
    (!, unquote-read-macro)
    (!,!@ unquotel-read-macro)
    (!,!. unquoted-read-macro))
  'LispReadMacro)

% End of file.

